local Apollo = require "Apollo"


local NLLC = Apollo.GetAddon("NextLabLootClient")
local JSON = Apollo.GetPackage("Lib:dkJSON-2.5").tPackage


local Coder = {}
NLLC.Coder = Coder;

--[[ Table of Contents:
		B64			- a modified version of Base64 to encode numbers (integers)
		EncTbl		- an Encoded table, consisting of only Tables, Boolean, Strings and Numbers (no metatables)
]]

--[[ ====== B64 ======
 THIS IS NOT THE DEFAULT Base 64, but a number-less version of it.
example:
	decimal: -12345
	octal: -30071 or -030071
	this B64: *DA=
]]

local tToB64 = { --this turns two octal characters into a single B64:
 ["00"] = "A",	["10"] = "I",	["20"] = "Q",	["30"] = "Y",	["40"] = "g",	["50"] = "o",	["60"] = "w",	["70"] = "$",
 ["01"] = "B",	["11"] = "J",	["21"] = "R",	["31"] = "Z",	["41"] = "h",	["51"] = "p",	["61"] = "x",	["71"] = "=",
 ["02"] = "C",	["12"] = "K",	["22"] = "S",	["32"] = "a",	["42"] = "i",	["52"] = "q",	["62"] = "y",	["72"] = "/",
 ["03"] = "D",	["13"] = "L",	["23"] = "T",	["33"] = "b",	["43"] = "j",	["53"] = "r",	["63"] = "z",	["73"] = "+",
 ["04"] = "E",	["14"] = "M",	["24"] = "U",	["34"] = "c",	["44"] = "k",	["54"] = "s",	["64"] = "!",	["74"] = "-",
 ["05"] = "F",	["15"] = "N",	["25"] = "V",	["35"] = "d",	["45"] = "l",	["55"] = "t",	["65"] = "?",	["75"] = "_",
 ["06"] = "G",	["16"] = "O",	["26"] = "W",	["36"] = "e",	["46"] = "m",	["56"] = "u",	["66"] = "&",	["76"] = ".",
 ["07"] = "H",	["17"] = "P",	["27"] = "X",	["37"] = "f",	["47"] = "n",	["57"] = "v",	["67"] = "#",	["77"] = ",",
}

local tFromB64 = setmetatable({}, {__index = function(_, k)
    error("[B64] invalid character for decoding:"..tostring(k))
  end})
for i, v in pairs(tToB64) do
	tFromB64[v] = i
end

Coder.toB64 = function(number)
	number = math.floor(number) --can only encode positive and negative integers
	local prefix = ""
	if number < 0 then
		number = math.abs(number)
		prefix = "*" --represents negative values.
	end

	local octal = ("%o"):format(number)

	if octal:len()%2 == 1 then
		octal = "0"..octal --doesnt change the value or encoding-legth
	end

	return prefix..octal:gsub("[0-7][0-7]", tToB64)
end

Coder.fromB64 = function(str)
	local negator = 1
	if str:sub(1,1) == "*" then
		negator = -1
		str = str:sub(2,-1) --remove the first character (the '*')
	end

	local octal = tonumber(str:gsub(".", tFromB64), 8)
	if octal then
		return negator*octal
	end
	return nil
end

--[[ ===== EncTbl ======
	EncTbl is used to Encode Tables into a single String. These can have values(and keys) of the type table, boolean,
  string and number. This Method is reference-keeping (example: local a = {}; b = {[1] = a, [2] = a}; after encoding and
  decoding b: b[1] == b[2]) Its using some parts of JSON-encoding. (non-integer-numbers, strings)
]]

local function search(str, key)
	return str:match("^(.-)"..key.."(.*)$")
end

--[[The following functions all have the same structure:
		'to???': number 	-> id, enc
        where id is an identifier and enc is the encoded value. Returning id==true expresses,
        that the value is fully defined by the id in enc (see: boolean).
		'from???: id, str	-> val, str
        where id is the given identifier, str is the left unprocessed string and val the resulting value
]]

local encTbl = {} --[type(v)] = func
local decTbl = {} --[id] = func


--encoding numbers in two different ways: integers to B64, others just to text (through JSON)
local floor = math.floor
local function toNumber(n)
	if floor(n) == n then --integer -> use B64
		return "i", Coder.toB64(n).."|"
	else
		return "n", JSON.encode(n).."|"
	end
end
encTbl["number"] = toNumber

local function fromNumber(type, str)
	local num, left = search(str, "|")

	if not num then
		error("[EncTbl] Number indicated, but not ended: \""..str:sub(1,20).."\"")
	end

	if type == "i" then
		num = Coder.fromB64(num)
	else --type == "n"
		num = JSON.decode(num)
	end
	return num, left
end
decTbl["i"] = fromNumber
decTbl["n"] = fromNumber

--encoding strings is done by JSON. note, that these encodings start and end with "
local function toString(s)
	--this puts a properly escaped string into quotes.
	local str = JSON.encode(s)

	--remove the first quote-character (the " )
	str = str:sub(2,-1)

	return '"', str
end
encTbl["string"] = toString

local function fromString(_, str)
	--we put a " in front of the string, before searching, because then
	--we guarantee at least one character in front of the ending "
	local s, tail, left = search("\""..str, "([^\\])\"")

	if not s then
		error("[EncTbl] String indicated, but not ended: \""..str:sub(1,20).."\"")
	end

	s = JSON.decode(s..tail.."\"") --the ending " was removed on search()

	return s, left
end
decTbl["\""] = fromString

local function toBool(b)
	return true, b and "+" or "-"
end
encTbl["boolean"] = toBool

local function fromBool(type, str)
	if type == "+" then
		return true, str
	else
		return false, str
	end
end
decTbl["+"] = fromBool
decTbl["-"] = fromBool

local function isIdx(i)
	return type(i)=="number" and i>0 and i==floor(i) or false
end
local function append(a,b,...) --just appends all strings together, ignoring all other values.
	local s = (type(a) == "string" and a or "")..(type(b) == "string" and b  or "")
	return select("#",...)>0 and append(s, ...) or s
end

--[[
	There are three different types of tables:
		'{'		- a default table, with no assumptions.
			example: "{_key_value_key_value}"
		'['		- a array, with no assumptions towards value types.
			example: "[_value_value_value]"
		'(_'	- a array with a specified type _ for all values.
			example: "(_valuevaluevalue)"
		'<'		- a reference to a already encoded table.
			example: "<ABC>" (num. ref. encoded with B64)
]]
local refTbl = nil -- the table to find references in
local function toTable(tbl)
	if refTbl[tbl] then
		return "<", Coder.toB64(refTbl[tbl])..">"
	else
		refTbl.num = refTbl.num + 1
		refTbl[tbl] = refTbl.num
	end

	local isArray = #tbl --holds either a number, or false. false, if its no array
	local cache = { --in this we'll gather all data nessecarry to build each of the table-types.
		key = {},
		t_val = {},
		enc_val = {},
	}
	local typed; --this one is a little confusing: nil: empty, "???" for a typed array, true for untyped

	--gather data, encode everything.
	for idx, val in pairs(tbl) do
		local t_idx, t_val = type(idx), type(val)
		if encTbl[t_idx] and encTbl[t_val] then
			local id_idx, enc_idx = encTbl[t_idx](idx)
			local id_val, enc_val = encTbl[t_val](val)

			table.insert(cache.key, append(id_idx, enc_idx))
			table.insert(cache.t_val, id_val)
			table.insert(cache.enc_val, enc_val)

			if isArray then
				if (not isIdx(idx) or idx>isArray) then
					isArray = false
				end
				typed = typed and typed~=id_val or id_val
			end
		else
			isArray = false
		end
	end

	--put stuff together to the correct table-type:
	if isArray then
		if typed and typed~=true then
			--(_???)

			local enc = table.concat(cache.enc_val)

			return "(", typed..enc..")"
		else
			--[_?_?_?]

			for i, val in ipairs(cache.enc_val) do
				cache.enc_val[i] = append(cache.t_val[i], val)
			end

			local enc = table.concat(cache.enc_val)

			return "[", enc.."]"
		end
	else
		--{_!_?_!_?_!_?}

		for i, val in ipairs(cache.enc_val) do
			cache.enc_val[i] = append(cache.key[i], cache.t_val[i], val)
		end

		local enc = table.concat(cache.enc_val)

		return "{", enc.."}"
	end
end
encTbl["table"] = toTable

refTbl = nil; --we use it here aswell - assumption: never encoding and decoding at the same time..
local function checkType(src, ignore, typ, ...)
	if typ ~= ignore and not (typ and decTbl[typ]) then
		error("[EncTbl] Unknown type in \""..src:sub(1,20).." ...\"")
	end
	return typ, ...
end
local function fromNormlTbl(_, str)

	local tbl = {}
	local t_key, v_key, t_val, v_val;

	table.insert(refTbl, tbl) --dont move this past decTbl[?](?,!) !!!

	while true do
		t_key, str = checkType(str, "}", str:match("^(.)(.*)$"))

		if t_key == "}" then
			break;
		end

		v_key, str = decTbl[t_key](t_key, str)

		t_val, str = checkType(str, false, str:match("^(.)(.*)$"))

		v_val, str = decTbl[t_val](t_val, str)

		tbl[v_key] = v_val
	end
	return tbl, str
end
decTbl["{"] = fromNormlTbl

local function fromArrayTbl(_, str)
	local i = 1;
	local tbl = {}
	local typ, val;

	table.insert(refTbl, tbl) --dont move this past decTbl[?](?,!) !!!

	while true do
		typ, str = checkType(str, "]", str:match("^(.)(.*)$"))
		if typ == "]" then
			break;
		end

		val, str = decTbl[typ](typ, str)

		tbl[i] = val
		i = i+1
	end
	return tbl, str
end
decTbl["["] = fromArrayTbl
local function fromTypedTbl(_, str)
	local i = 1;
	local tbl = {}
	local typ, val;

	table.insert(refTbl, tbl) --dont move this past decTbl[?](?,!) !!!

	typ, str = checkType(str, false, str:match("^(.)(.*)$"))

	while true do
		if str:sub(1,1) == ")" then
			str = str:sub(2,-1)
			break;
		end

		val, str = decTbl[typ](typ,str)

		tbl[i] = val
		i = i+1
	end
	return tbl, str
end
decTbl["("] = fromTypedTbl
local function fromReferTbl(_, str)
	local enc;
	enc, str = str:match("^(.-)>(.*)$")

	local idx = Coder.fromB64(enc)

	if not idx or not refTbl[idx] then
		error("[EncTbl] Referencing to non-existant table: \""..tostring(enc).."\" ("..tostring(idx)..")")
	end

	return refTbl[idx], str
end
decTbl["<"] = fromReferTbl

Coder.toEncTbl = function(tbl)
	if tbl == nil then return end
	refTbl = {num = 0}

	local t = type(tbl)
	if not encTbl[t] then
		error("[EncTbl] Type not supported for encoding: "..tostring(t))
	end

	local prefix, str = encTbl[t](tbl)

	refTbl = nil
	return append(prefix, str)
end

Coder.fromEncTbl = function(str)
	if not str or str=="" then return nil, "" end

	refTbl = {}

	local typ, val;

	typ, str = checkType(str, false, str:match("^(.)(.*)$"))
	val, str = decTbl[typ](typ, str)

	refTbl = nil
	return val, str --return the left string aswell.
end
