--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ApolloTimer = require "ApolloTimer"
local ChatSystemLib = require "ChatSystemLib"
local GameLib = require "GameLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- AutoRoll
-- Is the Collection of all AutoRolling Items; Here they get rolled for aswell.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local AutoRoll = {}
G.AutoRoll = AutoRoll

local RollTypeEnum = {
	MainSpec = 1,
	OffSpec  = 2,
	Costume  = 3,
	Pass = nil,
}
function AutoRoll:New(tRaidLooter)
	local o = {
		tLinker                = G.Linker:New(),
    tRaidLooter            = tRaidLooter,
    tMasterLoot            = tRaidLooter:GetMasterLoot(),
		tChatParser            = tRaidLooter:GetChatParser(),
    tItems                 = {}, --tItems["ChatLinkString"] = eRollType
    tLootCollections       = {}, --tLootCollections[i] = tLootCollection

		strNextCommand         = nil,
		tAutoRollTimer         = nil,
  }

	setmetatable(o, self)
	self.__index = self
	return o
end

function AutoRoll:Init(tTarget)
	self.tLinker:Link(NLLC, self)
  self.tMasterLoot:RegisterCallback(self)
	self.tChatParser:RegisterCallback(self)

	NLLC:OnceRestored("OnRestore", self)
	NLLC:OnceSaving("OnSave", self)
end

function AutoRoll:Destroy()
	self.tLinker:Destroy()
  self.tMasterLoot:UnregisterCallback(self)
	self.tChatParser:UnregisterCallback(self)

  self.tLinker = nil
  self.tRaidLooter = nil
  self.tMasterLoot = nil
  self.tItems = nil
  self.tLootCollections = nil
end

function AutoRoll:SetAutoRoll(tItem, eAutoRoll)
  local strLink = tItem:GetChatLinkString()

	if strLink then
		self.tItems[strLink] = eAutoRoll

		for _, tLootCollection in ipairs(self.tLootCollections) do
	    for i, tLootItem in ipairs(tLootCollection.tItems) do
	      if tLootItem:GetItem():GetChatLinkString() == strLink then
	        tLootItem:SetAutoRoll(eAutoRoll)
	      end
	    end
	  end
	end
end

function AutoRoll:GetAutoRoll(tItem)
  local strLink = tItem:GetChatLinkString()

	if strLink then
		return self.tItems[strLink]
	end
	return nil
end

function AutoRoll:RegisterLootItemCollection(tCollection)
	for _, Cmp in ipairs(self.tLootCollections) do
		if Cmp == tCollection then return end
	end
	table.insert(self.tLootCollections, tCollection)

  for i, tLootItem in ipairs(tCollection.tItems) do
    local strLink = tLootItem:GetItem():GetChatLinkString()
    if self.tItems[strLink] then
      tLootItem:SetAutoRoll(self.tItems[strLink])
    else
      tLootItem:SetAutoRoll(nil)
    end
  end
end

function AutoRoll:UnregisterLootItemCollection(tCollection)
	for i, Cmp in ipairs(self.tLootCollections) do
		if Cmp == tCollection then
			table.remove(self.tLootCollections, i)
			return
		end
	end
end

--from self.tMasterLoot:RegisterCallback(self)
function AutoRoll:HandleOnLootUpdate(arLoot)
  --remove all the self.tItems, which are not anymore available.
  --  (keep all those, which are still there...)

  local temp = self.tItems
  self.tItems = {}

  for i, tLoot in ipairs(arLoot) do
    local strLink = tLoot.itemDrop:GetChatLinkString()
    if strLink and temp[strLink] then
      self.tItems[strLink] = temp[strLink]
    end
  end

  if not self.tRefreshLootTimer then
    self.tRefreshLootTimer = ApolloTimer.Create(0, false, "RefreshLootItems", self)
  end
end

function AutoRoll:HandleRefreshLootItems()
  self.tRefreshLootTimer = nil

  --go through all self.tLootCollections and apply the AutoRoll

  for _, tLootCollection in ipairs(self.tLootCollections) do
    for i, tLootItem in ipairs(tLootCollection.tItems) do
      local strLink = tLootItem:GetItem():GetChatLinkString()
      if self.tItems[strLink] then
        tLootItem:SetAutoRoll(self.tItems[strLink])
      else
        tLootItem:SetAutoRoll(nil)
      end
    end
  end
end

function AutoRoll:HandleOnBeginRoll_ChatParser(strSender, tItem)
	local eAutoRoll = self:GetAutoRoll(tItem)
	if eAutoRoll then
		local strRollMax = (eAutoRoll == RollTypeEnum.OffSpec and " 50"
			or eAutoRoll == RollTypeEnum.Costume and " 33" or "")

		self.strCommand = G.Locale:TranslateCommand("Roll")..strRollMax
		if self.tAutoRollTimer then
			self.tAutoRollTimer:Stop() --ensure the time waited
			self.tAutoRollTimer:Start()
		else
			self.tAutoRollTimer = ApolloTimer.Create(1, false, "OnAutoRollTimer", self)
		end
	end
end

function AutoRoll:OnAutoRollTimer()
	if self.tAutoRollTimer then
		self.tAutoRollTimer:Stop()
		self.tAutoRollTimer = nil
	end

	if self.strCommand then
		ChatSystemLib.Command(self.strCommand)
		self.strCommand = nil
	end
end

function AutoRoll:OnSave(eLevel, tSavedData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then return end

	tSavedData.AutoRoll = self.tItems
end

function AutoRoll:OnRestore(eLevel, tSavedData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then return end

	if not tSavedData then return end

	if not tSavedData.AutoRoll then return end

	self.tItems = tSavedData.AutoRoll

	self:RefreshLootItems()
end

function AutoRoll:OnLootUpdate(...)
	G.Utils:InvokeEventHandler(self, "AutoRoll", "OnLootUpdate", ...) end
function AutoRoll:RefreshLootItems(...)
  G.Utils:InvokeEventHandler(self, "AutoRoll", "RefreshLootItems", ...) end
function AutoRoll:OnBeginRoll_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "AutoRoll", "OnBeginRoll_ChatParser", ...) end
