--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Autoloot Dropdown
-- Represents the dropdown view of the autoloot page.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local AutolootDropdown = {}
G.AutolootDropdown = AutolootDropdown
function AutolootDropdown:New(tRaidLooter)
	local o = {
		wndMain             = nil,
		wndLoadProfilesArea = nil,
		wndEditBox          = nil,
		tLinker             = G.Linker:New(),
		tRaidLooter         = tRaidLooter,
		tSettings           = tRaidLooter:GetSettings(),
		tProfiles           = {},
		btnCreateProfile    = nil,
		btnDeleteProfile    = nil,
		btnCopyProfile      = nil,
		btnLoadProfile      = nil,
		tAutolootProfiles   = nil,
		tSelectedItem       = nil,
		tCallback           = nil,
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function AutolootDropdown:Init(tTarget)
	self.tLinker:Link(NLLC, self)
	self.wndMain = self.tRaidLooter:CreateTopLevelWindow("Autoloot_ProfileDropdown", self)
	self.wndEditBox = self.wndMain:FindChild("EditBox")
	self.wndLoadProfilesArea = self.wndMain:FindChild("LoadProfilesArea")
	self.btnCreateProfile = self.wndMain:FindChild("BtnAddProfile")
	self.btnDeleteProfile = self.wndMain:FindChild("BtnDeleteProfile")
	self.btnCopyProfile = self.wndMain:FindChild("BtnCopyProfile")
	self.btnLoadProfile = self.wndMain:FindChild("BtnLoadProfile")
	self.tAutolootProfiles = self.tSettings:ReadBlock("tAutolootProfiles") -- Ref to storage

	self.btnCreateProfile:Enable(false)
	self.btnDeleteProfile:Enable(false)
	self.btnCopyProfile:Enable(false)
	self.btnLoadProfile:Enable(false)

	self:BuildProfileList()
	self.wndLoadProfilesArea:ArrangeChildrenVert(0)
	self.wndMain:Show(false, true)
end

function AutolootDropdown:Destroy()
	self.tLinker:Destroy()
	self.wndMain:Destroy()

	self.tLinker = nil
	self.wndMain = nil
	self.tCallback = nil
	self.tRaidLooter = nil
	self.tSettings = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function AutolootDropdown:GetWindow()
	return self.wndMain
end

function AutolootDropdown:Open(tPosition)
	self.wndMain:Invoke()
	self.wndMain:Move(tPosition.x, tPosition.y, self.wndMain:GetWidth(), self.wndMain:GetHeight())
end

function AutolootDropdown:SetCallback(tCallback)
	self.tCallback = tCallback
end

function AutolootDropdown:Close()
	self.wndMain:Close()
end

function AutolootDropdown:CreateNewProfileWithName(strName)
	local tProfile = G.AutolootDropdownItem:New(self.tRaidLooter, self.wndLoadProfilesArea, strName)
	tProfile:Init(self)
	tProfile:SetCallback(self)
	self.tProfiles[strName] = tProfile
	self.wndLoadProfilesArea:ArrangeChildrenVert(0)
	self.tAutolootProfiles[strName] = {
		bIsDefault = false,
		bIsLoadable = true,
		nCreatedAt = os.time(),
		strDisplayName = strName,
		tLootList = {}
	}
	self:RevalidateCanCreate()
	self.wndEditBox:SetText("")
end

function AutolootDropdown:BuildProfileList()
	for strName, tData in pairs(self.tAutolootProfiles) do
		local tProfile = G.AutolootDropdownItem:New(self.tRaidLooter, self.wndLoadProfilesArea, strName, tData)
		tProfile:Init(self)
		tProfile:SetCallback(self)
		self.tProfiles[strName] = tProfile
	end
end

function AutolootDropdown:RevalidateCanCreate()
	local strNewProfileName = self.wndEditBox:GetText()
	local bIsValid = (strNewProfileName ~= nil and strNewProfileName:len() >= 3
		and self.tAutolootProfiles[strNewProfileName] == nil)
	self.btnCreateProfile:Enable(bIsValid)
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function AutolootDropdown:HandleOnCreateNewProfile(wndHandler, wndControl, eMouseButton)
	self:CreateNewProfileWithName(self.wndEditBox:GetText())
end

function AutolootDropdown:HandleOnDeleteExistingProfile(wndHandler, wndControl, eMouseButton)
	if self.tSelectedItem ~= nil then
		local strName = self.tSelectedItem:GetName()
		self.tAutolootProfiles[strName] = nil
		self.tProfiles[strName]:Destroy()
		self.tProfiles[strName] = nil
		self.wndLoadProfilesArea:ArrangeChildrenVert(0)
	end
end

function AutolootDropdown:HandleOnCopyFromExistingProfile(wndHandler, wndControl, eMouseButton)
	if self.tSelectedItem ~= nil then
		local strName = self.tSelectedItem:GetName()
		G.Utils:FireEvent(self.tCallback, "OnCopyFromProfile", self.tAutolootProfiles[strName])
	end
end

function AutolootDropdown:HandleOnLoadProfile(wndHandler, wndControl, eMouseButton)
	if self.tSelectedItem ~= nil then
		local strName = self.tSelectedItem:GetName()
		G.Utils:FireEvent(self.tCallback, "OnLoadProfile", self.tAutolootProfiles[strName])
		self:Close()
	end
end

-- Avoid to many events to the chat here.
function AutolootDropdown:OnNewProfileNameChanged(wndHandler, wndControl, eMouseButton)
	self:RevalidateCanCreate()
end

function AutolootDropdown:HandleOnItemSelectionChanged(tAutolootDropdownItem)
	if tAutolootDropdownItem:IsSelected() then
		local tProfile = self.tAutolootProfiles[tAutolootDropdownItem:GetName()]

		self.btnDeleteProfile:Enable(not tProfile.bIsDefault)
		self.btnCopyProfile:Enable(true)
		self.btnLoadProfile:Enable(tProfile.bIsLoadable)
		self.tSelectedItem = tAutolootDropdownItem
	else
		self.btnDeleteProfile:Enable(false)
		self.btnCopyProfile:Enable(false)
		self.btnLoadProfile:Enable(false)
		self.tSelectedItem = nil
	end
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function AutolootDropdown:OnCreateNewProfile(...)
	G.Utils:InvokeEventHandler(self, "AutolootDropdown", "OnCreateNewProfile", ...) end
function AutolootDropdown:OnDeleteExistingProfile(...)
	G.Utils:InvokeEventHandler(self, "AutolootDropdown", "OnDeleteExistingProfile", ...) end
function AutolootDropdown:OnCopyFromExistingProfile(...)
	G.Utils:InvokeEventHandler(self, "AutolootDropdown", "OnCopyFromExistingProfile", ...) end
function AutolootDropdown:OnLoadProfile(...)
	G.Utils:InvokeEventHandler(self, "AutolootDropdown", "OnLoadProfile", ...) end
function AutolootDropdown:OnItemSelectionChanged(...)
	G.Utils:InvokeEventHandler(self, "AutolootDropdown", "OnItemSelectionChanged", ...) end
