--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Autoloot View
-- Represents the autoloot view of the addon.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local AutolootView = {}
G.AutolootView = AutolootView
function AutolootView:New(tRaidLooter, wndParent)
	local o = {
		wndMain            = wndParent,
		wndGrid            = wndParent:FindChild("AutolootGrid"),
		wndProfileTitle    = wndParent:FindChild("ProfileTitle"),
		btnAutolootingBtn  = wndParent:FindChild("BtnEnableAutolooting"),
		btnProfileDropdown = wndParent:FindChild("BtnProfileDropdown"),
		btnRuleDropdown    = wndParent:FindChild("BtnRuleDropdown"),
		tRules             = {},
		tRaidLooter        = tRaidLooter,
		tSettings          = tRaidLooter:GetSettings(),
		tMasterLoot        = tRaidLooter:GetMasterLoot(),
		tAutoloot          = nil,
		tAutolootProfiles  = nil,
		tLinker            = G.Linker:New(),
		tProfileDropdown   = nil,
		tRuleDropdown      = nil,
		tActiveProfile     = nil,
		nSelectedIndex     = 0,
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function AutolootView:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	-- local tAutoloot = self.tSettings:ReadSetting("tWindowLocations.tAutoloot")
	local bIsEnabled = self.tSettings:ReadSetting("tLeader.bAutolootEnabled")

	self.btnAutolootingBtn:SetCheck(bIsEnabled)
	self.tProfileDropdown = G.AutolootDropdown:New(self.tRaidLooter)
	self.tProfileDropdown:Init(self)
	self.tProfileDropdown:SetCallback(self)
	self.btnProfileDropdown:AttachWindow(self.tProfileDropdown:GetWindow())
	self.btnProfileDropdown:SetCheck(false)

	self.tRuleDropdown = G.AutolootPopup:New(self.tRaidLooter)
	self.tRuleDropdown:Init(self)
	self.tRuleDropdown:SetCallback(self)
	self.btnRuleDropdown:AttachWindow(self.tRuleDropdown:GetWindow())
	self.btnRuleDropdown:SetCheck(false)

	self.tAutolootProfiles = self.tSettings:ReadBlock("tAutolootProfiles") -- Ref to storage
	self.tAutoloot = self.tSettings:ReadBlock("tAutoloot") -- Ref to storage

	if self.tAutoloot.strProfileName == nil then
		self.tAutoloot.strProfileName = "Default"
	end

	self.tMasterLoot:RegisterCallback(self)
	self:LoadExistingProfile(self.tAutolootProfiles[self.tAutoloot.strProfileName])
end

function AutolootView:Destroy()
	self.tMasterLoot:UnregisterCallback(self)
	self.tLinker:Destroy()
	self.tRuleDropdown:Destroy()
	self.tProfileDropdown:Destroy()

	self.wndMain          = nil
	self.tLinker          = nil
	self.tProfileDropdown = nil
	self.tRuleDropdown    = nil
	self.tSettings        = nil
	self.tRaidLooter      = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function AutolootView:GetWindow()
	return self.wndMain
end

function AutolootView:AddNewRow(tRule, bPreventEvent)
	self:DrawRow(tRule)
	self.tActiveProfile.tLootList[tRule.strItemName] = {
		strRecipient = tRule.strUnitName,
		strComment = tRule.strComment
	}

	if not bPreventEvent then
		G.Utils:FireEvent(self.tRaidLooter:GetMasterLoot(), "OnAutolootChanged", self.tAutoloot.tProfileRef)
	end
end

function AutolootView:DrawRow(tRule)
	local nRow = self.wndGrid:AddRow("")
	self.wndGrid:SetCellDoc(nRow, 1,
		"<T Font=\"CRB_InterfaceSmall\" TextColor=\"UI_TextHoloBodyHighlight\">"..tRule.strUnitName.."</T>")
	self.wndGrid:SetCellDoc(nRow, 2,
		"<T Font=\"CRB_InterfaceSmall\" TextColor=\"UI_TextHoloBodyHighlight\">"..tRule.strItemName.."</T>")
	self.wndGrid:SetCellDoc(nRow, 3,
		"<T Font=\"CRB_InterfaceSmall\" TextColor=\"UI_BtnTextGrayNormal\">"..tRule.strComment.."</T>")

	self.wndGrid:SetCellLuaData(nRow, 2, tRule.strItemName)
	self.tRuleDropdown:Close()
end

function AutolootView:RepositionDropdown(btnDropdown, tDropdown, bShouldOpen)
	if bShouldOpen then
		local x, y = G.Utils:GetScreenPosition(btnDropdown)
		local nHeight = btnDropdown:GetHeight()

		tDropdown:Open({
			x = x - 5,
			y = y + 2 + nHeight
		})
	end
end

function AutolootView:LoadExistingProfile(tProfile)
	if tProfile == nil then
		tProfile = self.tAutolootProfiles["Default"]
		G.Utils:LogError("AutolootView:LoadExistingProfile", "Unable to load the requsted profile because it doesn't exist.")
	end

	self.wndProfileTitle:SetText(tProfile.strDisplayName)
	self.tAutoloot.tProfileRef = tProfile
	self.tAutoloot.strProfileName = tProfile.strDisplayName
	self.tActiveProfile = tProfile

	for strItemName, tItemInfo in pairs(tProfile.tLootList) do
		self:AddNewRow({
			strUnitName = tItemInfo.strRecipient,
			strItemName = strItemName,
			strComment = tItemInfo.strComment
		},
		true)
	end

	G.Utils:FireEvent(self.tRaidLooter:GetMasterLoot(), "OnAutolootChanged", self.tAutoloot.tProfileRef)
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function AutolootView:HandleOnToggleRuleDropdown(wndHandler, wndControl, eMouseButton)
	self:RepositionDropdown(self.btnRuleDropdown, self.tRuleDropdown, wndControl:IsChecked())
end

function AutolootView:HandleOnNewAutolootRuleAdded(tRule)
	self:AddNewRow(tRule)
end

function AutolootView:HandleOnAutolootRowDeleted()
	if self.nSelectedIndex > 0 then
		local strItemName = self.wndGrid:GetCellLuaData(self.nSelectedIndex, 2)
		self.tActiveProfile.tLootList[strItemName] = nil
		self.wndGrid:DeleteRow(self.nSelectedIndex)
	end
end

function AutolootView:HandleOnChangeAutolooting(bIsEnabled)
	local tLeaderSettings = self.tSettings:Read("tLeader", bIsEnabled)
	tLeaderSettings.bAutolootEnabled = bIsEnabled
end

function AutolootView:HandleOnDeleteAllAutolootRule()
	self.wndGrid:DeleteAll()
	self.tActiveProfile.tLootList = {}
end

function AutolootView:HandleOnToggleProfileDropdown(wndHandler, wndControl, eMouseButton)
	self:RepositionDropdown(self.btnProfileDropdown, self.tProfileDropdown, wndControl:IsChecked())
end

function AutolootView:HandleOnCopyFromProfile(tProfile)
	for strItemName, tItemInfo in pairs(tProfile.tLootList) do
		if self.tActiveProfile.tLootList[strItemName] == nil then
			local tRef = G.Utils:Deepcopy(tItemInfo)
			self.tActiveProfile.tLootList[strItemName] = tRef

			self:AddNewRow({
				strUnitName = tRef.strRecipient,
				strItemName = strItemName,
				strComment = tRef.strComment
			})
		end
	end
end

function AutolootView:HandleOnLoadProfile(tProfile)
	self.wndGrid:DeleteAll()
	self:LoadExistingProfile(tProfile)
end

function AutolootView:HandleOnGridSelectionChanged(wndControl, wndHandler, iRow, iCol, eMouseButton)
	local bDisselect = self.nSelectedIndex == iRow

	if bDisselect then
		self.nSelectedIndex = 0
	else
		self.nSelectedIndex = iRow
	end

	self.wndGrid:SetCurrentRow(self.nSelectedIndex)
	G.Utils:Log(self.wndGrid)
end

function AutolootView:HandleOnPrepareAddRuleForItem(tLootItem, strRecipientSuggestion)
	self.btnRuleDropdown:SetCheck(true)
	self:RepositionDropdown(self.btnRuleDropdown, self.tRuleDropdown, true)

	G.Utils:FireEvent(self.tRuleDropdown, "OnAutoComplete", {
		strItemName = tLootItem:GetName(),
		strRecipient = strRecipientSuggestion,
		strComment = ""
	})
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function AutolootView:OnToggleRuleDropdown(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnToggleRuleDropdown", ...) end
function AutolootView:OnNewAutolootRuleAdded(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnNewAutolootRuleAdded", ...) end
function AutolootView:OnAutolootRowDeleted(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnAutolootRowDeleted", ...) end
function AutolootView:OnDisableAutolooting(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnChangeAutolooting", false) end
function AutolootView:OnEnableAutolooting(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnChangeAutolooting", true) end
function AutolootView:OnLootUpdate(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnLootUpdate", ...) end
function AutolootView:OnDeleteAllAutolootRule(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnDeleteAllAutolootRule", ...) end
function AutolootView:OnToggleProfileDropdown( ... )
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnToggleProfileDropdown", ...) end
function AutolootView:OnCopyFromProfile( ... )
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnCopyFromProfile", ...) end
function AutolootView:OnLoadProfile( ... )
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnLoadProfile", ...) end
function AutolootView:OnGridSelectionChanged(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnGridSelectionChanged", ...) end
function AutolootView:OnPrepareAddRuleForItem(...)
	G.Utils:InvokeEventHandler(self, "AutolootView", "OnPrepareAddRuleForItem", ...) end
