--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ChatSystemLib = require "ChatSystemLib"
local ApolloTimer = require "ApolloTimer"
local GroupLib = require "GroupLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- ChatParser
-- Continuously minitors the party chat and generate events from text written.
-- This class is the core of this addon and is what drives the distribution of
-- loot.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local ChatParser = {}
G.ChatParser = ChatParser
function ChatParser:New()
	local o = {
		bIsParsing    = false,
		bIsForced     = true,
		bIsRolling    = false,
		tProcessTimer = nil,
		tAuthSenders  = {},
		tCombatants   = {},
		tCallbacks    = {},
		tOffspecWords = {"os","OS","Os","oS","off","Off","OFF","of","Offspec","offspec","ofspec","off-spec","Off-Spec",
			"Off-spec","Off spec","off spec","Off Spec","offspek"},
		tCostumeWords = {"Costume","costume","costum","Costum","cost","cos"},
		tMessageQueue = Queue:new(),
		nCombatants   = 0,
		eParseChannel = ChatSystemLib.ChatChannel_Party,
		tLinker       = G.Linker:New()
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function ChatParser:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	self.eParseChannel = G.Debug.bIsTest and ChatSystemLib.ChatChannel_Say or ChatSystemLib.ChatChannel_Party
	self.tProcessTimer = ApolloTimer.Create(5, true, "OnSlowProcessTimer", self)
	self.tProcessTimer:Stop()
end

function ChatParser:Destroy()
	self.tLinker:Destroy()
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function ChatParser:IsSenderAuthorized(strSender)
	return self.tAuthSenders[strSender] ~= nil or G.Utils:TableLength(self.tAuthSenders) == 0 or G.Debug.bIsTest
end

function ChatParser:Start()
	if not self.bIsParsing then
		G.Utils:Log("ChatParser Starting")
		Apollo.RegisterEventHandler("ChatMessage", "OnChatMessage", self)
		Apollo.RegisterEventHandler("FrameCount", "OnChatParseTimer", self)

		self.tProcessTimer:Start()
		self:OnSlowProcessTimer()
		self.bIsParsing = true
		self:Notify("OnParsingStarted_ChatParser")
	end
end

function ChatParser:ForceStop()
	G.Utils:Log("ChatParser ForceStoping")
	Apollo.RemoveEventHandler("ChatMessage", self)
	Apollo.RemoveEventHandler("FrameCount", self)

	self.bIsForced = true
	self.bIsParsing = false
	self.tProcessTimer:Stop()
	self:Notify("OnParsingStoped_ChatParser")
end

-- Should Respond to:
--     OnRoll(strUnitName, nScore)
--     OnBeginRoll(strUnitName, tItem)
--     OnEndRoll(strUnitName)
--     OnRollHint(strUnitName, tHint)
--     OnParsingStoped_ChatParser()
--     OnParsingStarted_ChatParser()
function ChatParser:RegisterCallback(tCallback)
	table.insert(self.tCallbacks, tCallback)
end

function ChatParser:UnregisterCallback(tCallback)
	for nIndex, tStoredCallback in ipairs(self.tCallbacks) do
		if tStoredCallback == tCallback then
			table.remove(self.tCallbacks, nIndex)
		end
	end
end

function ChatParser:IsParsing()
	return self.bIsParsing
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function ChatParser:OnChatParseTimer()
	local tMessagePost;

	while not self.tMessageQueue:Empty() do
		tMessagePost = self.tMessageQueue:Pop()

		if (tMessagePost.eChannelType == ChatSystemLib.ChatChannel_System and self.bIsRolling) then
			self:OnChatParseTimer_SystemParse(tMessagePost)
		elseif tMessagePost.eChannelType == self.eParseChannel then
			self:OnChatParseTimer_PartyMessage(tMessagePost)
		end
	end
end

function ChatParser:OnChatParseTimer_SystemParse(tMessagePost)
	local tMessages = tMessagePost.tMessage.arMessageSegments

	if tMessages[1] ~= nil and tMessages[1].strText ~= nil then
		self:OnSystemMessage(tMessages[1].strText, tMessagePost.tMessage.strSender)
	end
end

function ChatParser:OnChatParseTimer_PartyMessage(tMessagePost)
	local tMessages  = tMessagePost.tMessage.arMessageSegments
	local tItem      = nil
	local tExtraInfo = nil

	if tMessages[2] ~= nil then
		tItem = tMessages[2].uItem
	end

	if tMessages[3] ~= nil then
		tExtraInfo = tMessages[3].strText
	end

	if tMessages[1] ~= nil and tMessages[1].strText ~= nil then
		self:OnPartyMessage(tMessages[1].strText, tItem, tExtraInfo, tMessagePost.tMessage.strSender)
	end
end

-- Beware --
-- This is a workaround for being able to authorize the party members
-- as they post in the party chat. As we will only have the unit-name
-- of out-of-range units, another client can't verify if the poster have
-- permissions.
--
-- As a side-effect the player must have been in-range of the leader and/or
-- the assistant posting the roll message for anything to happen.
function ChatParser:OnSlowProcessTimer()
	-- WONT FIX: This can be improved by stopping when all members have
	--           been in range at least once and start listening for
	--           member flag changed events.
	--
	-- This will be called each 5 seconds.
	local nMembers = GroupLib.GetMemberCount()
	for nIndex = 1, nMembers do
		local tUnitMember = GroupLib.GetUnitForGroupMember(nIndex)
		local tMember     = GroupLib.GetGroupMember(nIndex)

		if tUnitMember ~= nil and tUnitMember:IsValid() and tMember ~= nil then
			if tMember.bRaidAssistant or tMember.bIsLeader then
				self.tAuthSenders[tUnitMember:GetName()] = true
			else
				self.tAuthSenders[tUnitMember:GetName()] = nil
			end
		end
	end
end

function ChatParser:OnChatMessage(tChannel, tMessage)
	-- Warning --
	-- Trying to store tMessage or any part of it will not work. I am guessing
	-- the real object somehow is destoryed or it might be an async issue ...
	-- However, pushing it to a Queue like in the ChatLog Addon seam to circumvent the
	-- problem.
	--
	-- So you can't put it in self here or send it to Print etc (because Wildstar Client
	-- will crash), at least that was what happend pre F2P.
	local eChannelType  = tChannel:GetType()
	local bAllowThrough;

	bAllowThrough = (eChannelType == ChatSystemLib.ChatChannel_System) and self.bIsRolling
	bAllowThrough = bAllowThrough or (eChannelType == self.eParseChannel)

	if bAllowThrough then
		self.tMessageQueue:Push({
			tMessage          = tMessage,
			eChannelType      = tChannel:GetType(),
			strChannelName    = tChannel:GetName(),
			strChannelCommand = tChannel:GetCommand(),
			idChannel         = tChannel:GetUniqueId()
		})
	end
end

----------------------------------------------------------------------------------
-- Private functions
----------------------------------------------------------------------------------
function ChatParser:IsRaidMember(tUnit)
	if G.Debug.bIsTest then
		return tUnit:IsThePlayer()
	else
		return tUnit:IsACharacter() and tUnit:IsInYourGroup()
	end
end

function ChatParser:StretchRoll(roll, key)
	if key == "(1-50)" then
		return (roll*2)-0.5
	elseif key == "(1-33)" then
		return roll*3
	end
	return roll
end

function ChatParser:OnSystemMessage(strText)
	local strForename, strLastname, strC1, strValue, strC2 = G.Utils:Split(strText, " ")

	if self:IsControlValuesMet(strC1, strC2) then
		local nScore = tonumber(strValue)

		nScore = self:StretchRoll(nScore, strC2)

		if self:IsRollInRange(nScore) then
			self:Notify("OnRoll_ChatParser", strForename.." "..strLastname, nScore)
		end
	end
	if strC2 == "(1-50)" then
    self:SendRollHintSignal(true, strForename.." "..strLastname)
  elseif strC2 == "(1-33)" then
		self:SendRollHintSignal(false, strForename.." "..strLastname)
	end
end

function ChatParser:OnPartyMessage(strText, tItem, strExtraInfo, strSender)
	if self.bIsRolling then
		if self:IsAnOffspecHint(strText) then
			self:SendRollHintSignal(true, strSender)
		elseif self:IsAnCostumeHint(strText) then
			self:SendRollHintSignal(false, strSender)
		elseif self:IsSenderAuthorized(strSender) then
			if self:IsNoRolls(strText) then
				self:Notify("OnEndRoll_ChatParser", strSender)
				self.bIsRolling = false
			else
				local bRollsEnding = self:ParsePotentailRolls(strText, strSender)
				self.bIsRolling = not bRollsEnding
			end
		end
	elseif self:IsARollStart(strText, tItem, strSender) then
		local bAscending = strExtraInfo ~= nil and strExtraInfo == "(lowest wins)"
		self:Notify("OnBeginRoll_ChatParser", strSender, tItem, bAscending)
		self.bIsRolling = true
	end
end

function ChatParser:ParsePotentailRolls(strText, strSender)
	local strTwoFirstWords = self:GetFirstTwoWords(strText)

	if self:IsOneWinner(strTwoFirstWords) or self:IsATie(strTwoFirstWords) then
		self:Notify("OnEndRoll_ChatParser", strSender)
		return true --found roll-ending phrase
	end
	return false
end

function ChatParser:GetFirstTwoWords(strText)
	local strWord1, strWord2 = G.Utils:Split(strText, " ")

	if strWord1 ~= nil and strWord2 ~= nil then
		return strWord1.." "..strWord2
	end

	return nil
end

function ChatParser:SendRollHintSignal(bIsOffspec, strSender)
	local tHint = {
		bIsOffspec = bIsOffspec,
		bIsCostume = not bIsOffspec
	}

	self:Notify("OnRollHint_ChatParser", strSender, tHint)
end

function ChatParser:MatchesAnyOf(tAvailable, strWord)
	for nIndex, strAvailableWord in pairs(tAvailable) do
		if strWord == strAvailableWord then
			return true
		end
	end

	return false
end

function ChatParser:IsControlValuesMet(strC1, strC2)
	return strC2 == "(1-100)" or strC2 == "(1-50)" or strC2 == "(1-33)" or false
end

function ChatParser:IsRollInRange(nScore)
	return nScore ~= nil and 1 <= nScore and nScore <= 100
end

function ChatParser:IsARollStart(strText, tItem, strSender)
	return self:IsSenderAuthorized(strSender) and (strText == "Roll for: " or strText == "Reroll for: ") and tItem ~= nil
end

function ChatParser:IsNoRolls(strText)
	return self:IsTextMatching(strText, "No rolls.")
end

function ChatParser:IsOneWinner(strText)
	return self:IsTextMatching(strText, "Best roll:")
end

function ChatParser:IsATie(strText)
	return self:IsTextMatching(strText, "Tie between:")
end

function ChatParser:IsTextMatching(strText, strActualText)
	return strText ~= nil and strText == strActualText
end

function ChatParser:IsAnOffspecHint(strText)
	return self:MatchesAnyOf(self.tOffspecWords, strText)
end

function ChatParser:IsAnCostumeHint(strText)
	return self:MatchesAnyOf(self.tCostumeWords, strText)
end

function ChatParser:Notify(strEventName, ...)
	G.Utils:Log("Notify => " .. strEventName)
	for _, tCallback in ipairs(self.tCallbacks) do
		G.Utils:FireEvent(tCallback, strEventName, ...)
	end
end
