--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- ContextMenu
-- Represents the context menu in RaidLooter
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local ContextMenu = {}
G.ContextMenu = ContextMenu

local RollTypeEnum = {
	MainSpec = 1,
	OffSpec  = 2,
	Costume  = 3,
	Pass = nil,
}

function ContextMenu:New(tRaidLooter, tLootItem)
	local o = {
		tRaidLooter = tRaidLooter,
		tAutoRoll = tRaidLooter:GetAutoRoll(),
		tLootItem = tLootItem,
		tLinker = G.Linker:New(),
		wndMain = nil,
		btnAddRule = nil,
		btnUnqueue = nil,
		btnViewDetails = nil,
		btnLinkToChat = nil,
		btnPreview = nil,
		btnCrbContextMenu = nil,
		tLootQueue = tRaidLooter:GetLootQueue(),
		tCursor = nil
	}
	setmetatable(o, self)
	self.__index = self
	return o
end

function ContextMenu:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	self.wndMain = self.tRaidLooter:CreateTooltipWindow("LootItem_ContextMenu", self)
	self.btnAddRule = self.wndMain:FindChild("btnAddRule")
	self.btnUnqueue = self.wndMain:FindChild("btnUnqueue")
  self.btnViewDetails = self.wndMain:FindChild("btnViewDetails")
	self.btnLinkToChat = self.wndMain:FindChild("btnLinkToChat")
	self.btnPreview = self.wndMain:FindChild("btnPreview")
	self.btnAutoRoll = self.wndMain:FindChild("btnAutoRoll")
	self.btnMore = self.wndMain:FindChild("btnMore")

	local tContextMenuAddon = Apollo.GetAddon("ContextMenuItem")
	local bCanPreview = true
	if tContextMenuAddon ~= nil and tContextMenuAddon["HelperValidatePreview"] ~= nil then
		bCanPreview = tContextMenuAddon:HelperValidatePreview(self.tLootItem:GetItem())
	end

	self.btnPreview:Enable(bCanPreview)
	self.btnUnqueue:Enable(G.Utils:IsLeader())
	self.btnAddRule:Enable(G.Utils:IsLeader())
end

function ContextMenu:Destroy()
	self.tLinker:Destroy()
	self.wndMain:Destroy()
	self.wndMain = nil
	self.tLootItem = nil
	self.tRaidLooter = nil
	self.tLootQueue = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function ContextMenu:Open()
	local knXCursorOffset = 10
	local knYCursorOffset = 25

	self.tCursor = Apollo.GetMouse()
	self.wndMain:Invoke()
	self.wndMain:Move(self.tCursor.x - knXCursorOffset, self.tCursor.y - knYCursorOffset,
		self.wndMain:GetWidth(), self.wndMain:GetHeight())
end

function ContextMenu:Close()
	if self.wndMain ~= nil then
		self.wndMain:Close()
	end
end

----------------------------------------------------------------------------------
-- Getters & Setters
----------------------------------------------------------------------------------
function ContextMenu:GetPosition() return self.tCursor end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function ContextMenu:HandleOnActionSelected(wndHandler, wndControl, eMouseButton)
	local strButtonName = wndControl:GetName()

	if strButtonName == "btnAddRule" then
		self:FireAddRuleEvent()
	elseif strButtonName == "btnUnqueue" then
		self.tRaidLooter:GetLootQueue():Remove(self.tLootItem)
	elseif strButtonName == "btnViewDetails" then
		self.tRaidLooter:ShowItemDetails(self.tLootItem)
	elseif strButtonName == "btnLinkToChat" then
		Event_FireGenericEvent("GenericEvent_LinkItemToChat", self.tLootItem:GetItem())
	elseif strButtonName == "btnPreview" then
		self:FirePreviewEvent()
	elseif strButtonName == "btnAutoRoll" then
		local tItem = self.tLootItem:GetItem()
		local eAutoRoll = self.tAutoRoll:GetAutoRoll(tItem)
		if not eAutoRoll then
			eAutoRoll = RollTypeEnum.MainSpec
		elseif eAutoRoll >= RollTypeEnum.Costume then
			eAutoRoll = nil
		else
			eAutoRoll = eAutoRoll+1
		end
		self.tAutoRoll:SetAutoRoll(tItem, eAutoRoll)
		return --do not close the context menu
	elseif strButtonName == "btnMore" then
		Event_FireGenericEvent("GenericEvent_ContextMenuItem", self.tLootItem:GetItem())
	end

	self:Close()
end

----------------------------------------------------------------------------------
-- Private functions
----------------------------------------------------------------------------------
function ContextMenu:FirePreviewEvent()
	local tItem = self.tLootItem:GetItem()
	local nDecorId = tItem:GetHousingDecorInfoId()
	if nDecorId and nDecorId ~= 0 then
		Event_FireGenericEvent("GenericEvent_LoadDecorPreview", nDecorId)
	else
		Event_FireGenericEvent("GenericEvent_LoadItemPreview", tItem)
	end
end

function ContextMenu:FireAddRuleEvent()
	local tManagementWindow = self.tRaidLooter:GetActiveWindow()
	G.Utils:FireEvent(tManagementWindow, "OnPrepareAddRuleForItem", self.tLootItem)
end
----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function ContextMenu:OnActionSelected(...) G.Utils:InvokeEventHandler(self, "ContextMenu", "OnActionSelected", ...) end
