--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ChatSystemLib = require "ChatSystemLib"
local GameLib = require "GameLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Tests
-- Defines varaibles and functions used while developing.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local Debug = {}
G.Debug = Debug
function Debug:Init(tRaidLooter)
	self.bIsTest            = false
	self.bIsTracingEvents    = false
	self.bIsDebug             = false
	self.bShouldLink           = false
	self.bIsLootSourceInventory = false
	self.bIsLeader             = false
	self.bIsAssistant         = false
	self.bIsInRaid           = false
	self.tMasterLoot        = {}
	self.tRecipients       = {}
	self.tRaidLooter      = tRaidLooter
	self.tLinker          = G.Linker:New()
	self.tLinker:Link(NLLC, self)
end

function Debug:Destroy() self.tLinker:Destroy() end

----------------------------------------------------------------------------------
-- Save & Restore
----------------------------------------------------------------------------------
function Debug:OnSave(eLevel)
	if self.tSettings == nil then
		return
	end

	local tDebug = self.tSettings:Read("tDebug")
	if tDebug ~= nil then
		tDebug.bIsTest                = self.bIsTest
		tDebug.bIsTracingEvents       = self.bIsTracingEvents
		tDebug.bIsDebug               = self.bIsDebug
		tDebug.bShouldLink            = self.bShouldLink
		tDebug.bIsLootSourceInventory = self.bIsLootSourceInventory
		tDebug.bIsLeader              = self.bIsLeader
		tDebug.bIsAssistant           = self.bIsAssistant
		tDebug.bIsInRaid              = self.bIsInRaid
	end
end

function Debug:OnRestore()
	if self.tSettings == nil then
		return
	end

	self.bIsTest                = self.tSettings:Read("tDebug.bIsTest")
	self.bIsTracingEvents       = self.tSettings:Read("tDebug.bIsTracingEvents")
	self.bIsDebug               = self.tSettings:Read("tDebug.bIsDebug")
	self.bShouldLink            = self.tSettings:Read("tDebug.bShouldLink")
	self.bIsLootSourceInventory = self.tSettings:Read("tDebug.bIsLootSourceInventory")
	self.bIsLeader              = self.tSettings:Read("tDebug.bIsLeader")
	self.bIsAssistant           = self.tSettings:Read("tDebug.bIsAssistant")
	self.bIsInRaid              = self.tSettings:Read("tDebug.bIsInRaid")
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function Debug:SetSettings(tSettings)
	self.tSettings = tSettings
end

function Debug:SetDevMode(bEnable)
	self.bIsTest  = bEnable
	self.bIsDebug = bEnable
	self.bIsTracingEvents = bEnable
	self.bShouldLink = bEnable
	self:SetLootSource(bEnable and "inventory" or "masterloot")
	self:SetRole(bEnable and "leader" or "normal")

	ChatSystemLib.Command("/reloadui")
end

function Debug:SetTestMode(bEnable)
	self.bIsTest = false
	self.bIsDebug = bEnable
	self.bIsTracingEvents = bEnable
	self.bShouldLink = bEnable
	self:SetLootSource("masterloot")
	self:SetRole("normal")

	ChatSystemLib.Command("/reloadui")
end

function Debug:SetTraceEvents(bEnable)
	self.bIsTracingEvents = bEnable
end

function Debug:SetBetterErrors(bEnable)
	self.bShouldLink = bEnable
	ChatSystemLib.Command("/reloadui")
end

function Debug:SetLootSource(strSource)
	self.bIsLootSourceInventory = strSource == "inventory"
end

function Debug:SetRole(strRole)
	self.bIsLeader    = strRole == "leader"
	self.bIsAssistant = strRole == "assist"
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function Debug:OnLoad()
	if self.bIsTest then
		Apollo.RegisterSlashCommand("rl-test", "OnRaidLooterTestCmd", self)
	end
end

function Debug:OnRaidLooterTestCmd(_, params)
	if not self.bIsTest then
		return
	end
	local cmd, value, value2 = G.Utils:Split(params, " ")

	if cmd == "gen-loot" then
		local tLoot = self:GenerateTestLoot(tonumber(value))
		local tMasterLoot = self.tRaidLooter:GetMasterLoot()
		tMasterLoot:AddLoot(tLoot)
	elseif cmd == "rand" then
		G.Utils:PrintRandomNumbers(tonumber(value))
	elseif cmd == "settings" then
		self:PrintSettings(value)
	elseif cmd == "read-setting" then
		self:PrintSetting(value)
	elseif cmd == "update-loot" then
		local tMasterLoot = self.tRaidLooter:GetMasterLoot()
		tMasterLoot:Update()
	elseif cmd == "mass-roll" then
		self:MakeRolles(value)
	elseif cmd == "set" then
		if value == "loot-source" then
			self:SetLootSource(value2)
		elseif value == "role" then
			self:SetRole(value2)
		elseif value == "trace-events" then
			self:SetTraceEvents(value2 == "on")
		elseif value == "better-errors" then
			self:SetBetterErrors(value2 == "on")
		end
	elseif cmd == "trigger-toggle-leader" then
		self.tRaidLooter:HandleOnMemberGroupFlagsChanged(0, true, {})
	else
		ChatSystemLib.Command("/rl")
		ChatSystemLib.Command("/rover")

		local tLoot = self:GenerateTestLoot(15)
		local tMasterLoot = self.tRaidLooter:GetMasterLoot()
		tMasterLoot:AddLoot(tLoot)
	end
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function Debug:GenerateTestLoot(nX)
	local tPlayerUnit    = GameLib.GetPlayerUnit()
	local tInventoryLoot = tPlayerUnit:GetInventoryItems()
	local nItems         = G.Utils:TableLength(tInventoryLoot)
	local tMasterLoot    = {}

	for i=1,nX do
		local nIndex = i==1 and 1 or G.Utils:Random(nItems)
		local tItem  = tInventoryLoot[nIndex]

		if i==1 then
			G.Utils:Log(tItem.itemInBag)
		end

		table.insert(tMasterLoot, {
			bIsMaster = true,
			itemDrop  = tItem.itemInBag,
			nLootId   = i==1 and 42 or G.Utils:Random(1000000),
			tLooters  = {
				[1] = tPlayerUnit
			},
			tLootersOutOfRange = {
				[1] = "Aaven Dragonborn",
				[2] = "Chromid Chan",
				[3] = "Tobbu Meatsheild",
				[4] = "Ethyria Torviir",
				[5] = "Pol Gara"
			},
		})
	end
	return tMasterLoot
end

function Debug:PrintSetting(strKeySequnce)
	G.Utils:Log(self.tRaidLooter:GetSettings():ReadSetting(strKeySequnce))
end

function Debug:PrintSettings(value)
	if value == "default" then
		G.Utils:Log("Default Settings")
		self:PrintSettingsHelper(self.tRaidLooter:GetSettings():GetDefaultTable(), "")
	elseif value == "new" then
		G.Utils:Log("Saved Settings")
		self:PrintSettingsHelper(self.tRaidLooter:GetSettings():GetNewTable(), "")
	else
		G.Utils:Log("Current Settings")
		self:PrintSettingsHelper(self.tRaidLooter:GetSettings():GetTable(), "")
	end
end

function Debug:PrintSettingsHelper(tSettings, indent)
	for strKey, value in pairs(tSettings) do
		if type(value) == 'table' then
			G.Utils:Log(indent..strKey .. " => {")
			self:PrintSettingsHelper(value, indent.."  ")
			G.Utils:Log(indent.."}")
		else
			G.Utils:Log(indent..strKey .. " => " .. tostring(value))
		end
	end
end

function Debug:MakeRolles(nRolls)
	for nIndex=1,nRolls do
		ChatSystemLib.Command(G.Locale:TranslateCommand("Roll"))
	end
end
