--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ApolloTimer = require "ApolloTimer"
local ChatSystemLib = require "ChatSystemLib"
local Tooltip = require "Tooltip"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Distribution View
-- Represents that distribution view/tab of the management window. This is the
-- tab the raid leader will use to distribute loot to other raiders.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local DistributionView = {}
G.DistributionView = DistributionView
function DistributionView:New(tRaidLooter, wndParent)
	local o = {
		wndParent       = wndParent,
		wndLoot         = wndParent:FindChild("AvailableItemContainer"),
		wndRecp         = wndParent:FindChild("RecipientsList"),
		wndRollerList   = wndParent:FindChild("RollTopList-Dist"),
		wndRollerSort   = wndParent:FindChild("RollSortWindow"),
		wndItemStop     = wndParent:FindChild("ItemStop"),
		wndElapsedTime  = wndParent:FindChild("Elapsed"),
		wndSpec         = wndParent:FindChild("RecipientSpec"),
		btnAssign       = wndParent:FindChild("AssignButton"),
		btnAssingMinor  = wndParent:FindChild("AssignMinorButton"),
		btnAssingMajor  = wndParent:FindChild("AssignMajorButton"),
		btnSpec         = wndParent:FindChild("RecipientSpecButton"),
		btnBeginRoll    = wndParent:FindChild("BeginRollButton"),
		btnEndRoll      = wndParent:FindChild("EndRollButton"),
		btnRandomize    = wndParent:FindChild("RandomizeButton"),
		btnReroll       = wndParent:FindChild("RerollButton"),
		btnSelectAll    = wndParent:FindChild("SelectAllItemsButton"),
		btnUnselectAll  = wndParent:FindChild("UnselectAllItemsButton"),
		btnFilterNone   = wndParent:FindChild("Filter_AllButton"),
		btnFilterLight  = wndParent:FindChild("Filter_LButton"),
		btnFilterMedium = wndParent:FindChild("Filter_MButton"),
		btnFilterHeavy  = wndParent:FindChild("Filter_HButton"),
		btnFilterOptionRecipients = wndParent:FindChild("Filter_Option_Recipients"),
		btnFilterOptionEquipable = wndParent:FindChild("Filter_Option_Equipable"),
		btnFilterOptionCommodity = wndParent:FindChild("Filter_Option_Commodity"),
		nElapsedTime    = 0,
		tElapsedTimer   = nil,
		tLootSet        = nil,
		tRecipientSet   = nil,
		tRollSet        = nil,
		tSelectedRaider = {
			tRecipient = nil,
			bIsLocked  = false,
			bOverwrittenSpec = nil,
		},
		tSelectedLootItem = {
			wndIcon     = wndParent:FindChild("SelectedItemIcon"),
			wndName     = wndParent:FindChild("SelectedItemName"),
			tItem       = nil,
			bIsLocked   = false,
			bIsMultiple = false
		},
		tRaidLooter = tRaidLooter,
		tSettings   = tRaidLooter:GetSettings(),
		tMasterLoot = tRaidLooter:GetMasterLoot(),
		tChatParser = tRaidLooter:GetChatParser(),
		bIsRolling  = false,
		tLinker     = G.Linker:New()
	}
	setmetatable(o, self)
	self.__index = self
	return o
end

function DistributionView:Init(tTarget)
	self.tLinker:Link(NLLC, self)
	self.tElapsedTimer = ApolloTimer.Create(1, true, "OnOneSecondElapsed", self)
	self.tElapsedTimer:Stop()

	-- Set the initial state for some controls.
	self.btnReroll:Show(false, true)
	self.wndParent:FindChild("Filter_AllButton"):SetCheck(true)
	self.wndParent:FindChild("Sort_ByLevel"):SetCheck(true)
	self.wndItemStop:Show(false, true)
	self:DisplayNoItemSelected()

	-- Collection of raiders, loot and rollers
	self.tLootSet      = G.LootItemCollection:New(self.tRaidLooter, true, true, self.wndLoot)
	self.tRecipientSet = G.RecipientCollection:New(self.tRaidLooter, self.wndRecp)
	self.tRollSet      = G.RollerCollection:New(self.tRaidLooter, true, self.wndRollerList, self.wndRollerSort)

	self.tLootSet:Init(self)
	self.tRecipientSet:Init(self)
	self.tRollSet:Init(self)

	self.tLootSet:SetCompareFunction(function(tItem)
		return tItem:GetEquippedItemForItemType()
	end)

	self.tLootSet:SetCallback(self)
	self.tRecipientSet:SetCallback(self)
	self.tRollSet:SetCallback(self, true)

	self.tMasterLoot:RegisterCallback(self)
	self.tChatParser:RegisterCallback(self)
end

function DistributionView:Destroy()
	self.tMasterLoot:UnregisterCallback(self)
	self.tChatParser:UnregisterCallback(self)
	self.tLootSet:SetCallback(nil)
	self.tRecipientSet:SetCallback(nil)
	self.tRollSet:SetCallback(nil)

	self.tLinker:Destroy()
	self.tLootSet:Destroy()
	self.tRecipientSet:Destroy()
	self.wndRollerList:Destroy()

	self.tLootSet      = nil
	self.tRecipientSet = nil
	self.wndRollerList = nil
	self.tRaidLooter   = nil
	self.tSettings     = nil
	self.tMasterLoot   = nil
end

function DistributionView:SetFiltersEnable(bEnable)
	self.btnFilterNone:Enable(bEnable)
	self.btnFilterLight:Enable(bEnable)
	self.btnFilterMedium:Enable(bEnable)
	self.btnFilterHeavy:Enable(bEnable)
	self.btnFilterOptionRecipients:Enable(bEnable)
	self.btnFilterOptionEquipable:Enable(bEnable)
	self.btnFilterOptionCommodity:Enable(bEnable)
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function DistributionView:HandleOnLootUpdate(tLoot)
	self:UpdateRecipients()
end

function DistributionView:HandleOnRollerClicked(tRoller)
	if tRoller then
		local strName = tRoller:GetName()
		local tRaider = self.tRecipientSet:FindRecipientByName(strName)
		if tRaider then
			tRaider:Select()
		end
	end
end

function DistributionView:HandleOnLootItemSelectionChanged(tSelectedItems)
	local nNumSelected        = #tSelectedItems
	local bIsNoneSelected     = nNumSelected == 0
	local bIsMultipleSelected = nNumSelected > 1

	self.tSelectedRaider.bOverwrittenSpec = nil

	if bIsNoneSelected then
		self:Reset()
	elseif bIsMultipleSelected then
		self:DisplayMultipleItemsSelected()
	else
		local tSelectedItem = tSelectedItems[1]
		self.tSelectedLootItem.tItem = tSelectedItem
		self:DisplaySingleItemSelect(tSelectedItem)
	end

	self:UpdateRecipients()
end

function DistributionView:HandleOnRecipientSelectionChanged(tOldRecipient, tNewRecipient)
	self.btnAssign:Enable(tNewRecipient ~= nil)
	self.btnAssingMinor:Enable(tNewRecipient ~= nil)
	self.btnAssingMajor:Enable(tNewRecipient ~= nil)
	self.btnSpec:Show(tNewRecipient ~= nil)
	self.wndSpec:Show(tNewRecipient ~= nil)
	self.tSelectedRaider.tRecipient = tNewRecipient
	self.tSelectedRaider.bOverwrittenSpec = nil

	local _, strSpec = self:GetCurrentSpec()
	self.wndSpec:SetText(strSpec)
end

function DistributionView:HandleOnAssignLoot()
	self:EndRoll(true)
	G.Utils:PlaySound(236)
	local tSelectedItems = self.tLootSet:GetSelectedItems()
	local tRaider = self.tRecipientSet:GetSelectedRecipient()

	if tRaider == nil or tSelectedItems == nil then
		return -- No recipient selected.
	end

	local tUnitInfo = {
		tUnit = tRaider:GetUnit(),
		strUnitName = tRaider:GetFullname()
	}

	self.tMasterLoot:Assign(tSelectedItems, tUnitInfo)
	self:Reset()
	self.tLootSet:SelectFirst()
end

function DistributionView:GetCurrentSpec()
	if self.tSelectedRaider.bOverwrittenSpec == true then
		return true, "Mainspec"
	elseif self.tSelectedRaider.bOverwrittenSpec == false then
		return false, "Offspec"
	end

	if not self:RolledForSelectedItem() then return true, "Mainspec" end

	local tRaider = self.tRecipientSet:GetSelectedRecipient()
	if not tRaider then return true, "Mainspec" end
	local strUnitName = tRaider:GetFullname()
	local tRoller = self.tRollSet:GetRoller(strUnitName)
	if not tRoller then return true, "Mainspec" end
	if tRoller:IsRollingMainspec() then
		return true, "Mainspec"
	else
		return false, "Offspec"
	end
end

function DistributionView:RolledForSelectedItem()
	local arSelectedItems = self.tLootSet:GetSelectedItems()
	if #arSelectedItems ~= 1 then return false end

	local tLootItem = self.tRollSet:GetItem()
	if not tLootItem then return false end

	local tItem = tLootItem:GetItem()
	if not tItem then return false end

	local strLink1 = tItem:GetChatLinkString()
	local strLink2 = arSelectedItems[1]:GetItem():GetChatLinkString()

	return strLink1 == strLink2
end

function DistributionView:HandleOnAssignMinorLoot()
	local arSelectedItems = self.tLootSet:GetSelectedItems()
	if #arSelectedItems > 0 then
		local tRaider = self.tRecipientSet:GetSelectedRecipient()
		local strUnitName = tRaider:GetFullname()
		local bMain = self:GetCurrentSpec()
		NLLC.Ranker:ApplyLootOffset(strUnitName, bMain, -1)
		self:ShareAssignedLoot(arSelectedItems, strUnitName, bMain, "Minor")
	end
	self:HandleOnAssignLoot()
end

function DistributionView:HandleOnAssignMajorLoot()
	local arSelectedItems = self.tLootSet:GetSelectedItems()
	if #arSelectedItems > 0 then
		local tRaider = self.tRecipientSet:GetSelectedRecipient()
		local strUnitName = tRaider:GetFullname()
		local bMain = self:GetCurrentSpec()
		NLLC.Ranker:ApplyLootOffset(strUnitName, bMain, -2)
		self:ShareAssignedLoot(arSelectedItems, strUnitName, bMain, "Major")
	end
	self:HandleOnAssignLoot()
end

function DistributionView:ShareAssignedLoot(tItems, strUnitName, bMain, strSuffix)
	local nChNum   = G.Debug.bIsTest and ChatSystemLib.ChatChannel_Say or ChatSystemLib.ChatChannel_Party
	local tChannel = ChatSystemLib.GetChannels()[nChNum]

	local strMain = bMain and "Main" or "Off"

	for i, tItem in ipairs(tItems) do
		local strMsg = tItem:GetItem():GetChatLinkString().. " - " .. strUnitName .. " - " .. strMain .. " - " .. strSuffix

		tChannel:Send(strMsg)
	end
end

function DistributionView:HandleOnSwitchSpec()
	local bSpec = self:GetCurrentSpec()
	self.tSelectedRaider.bOverwrittenSpec = not bSpec
	local _, strSpec = self:GetCurrentSpec()
	self.wndSpec:SetText(strSpec)
end

function DistributionView:HandleOnBeginRoll(bIsReroll)
	local tLootItem = self.tSelectedLootItem.tItem
	local bIsAscending = self.tSettings:ReadSetting("tLeader.bAscending")

	G.Utils:PlaySound(173)
	self.bIsRolling = true
	self.nElapsedTime = 0
	self.tElapsedTimer:Start()
	self.wndElapsedTime:SetText("")
	self.tRollSet:Clear()
	self.tRollSet:BeginRollingForItem(tLootItem, bIsAscending)
	self.tLootSet:Lock()
	self.btnReroll:Enable(false)
	self.wndItemStop:Show(true)
	self.btnBeginRoll:Enable(false)
	self.btnEndRoll:Enable(true)
	self.btnUnselectAll:Enable(false)
	self.btnSelectAll:Enable(false)
	self:SetFiltersEnable(false)

	if self.tSettings:ReadSetting("tGeneral.bUseRollPopups") then
		-- self.tRaidLooter:ShowRollDialogForItem(tLootItem:GetItemId())
		self.tRaidLooter:ShowRollDialogForLootItem(tLootItem)
	end

	local strTieString = self.tRollSet:FormatTopTie()
	self:SendBeginRollToChat(bIsReroll, strTieString, bIsAscending)
end

function DistributionView:HandleOnEndRoll()
	G.Utils:PlaySound(173)
	self:EndRoll()
end

function DistributionView:HandleOnRandomizeLoot()
	self:EndRoll(true)
	G.Utils:PlaySound(173)
	local tSelectedItems = self.tLootSet:GetSelectedItems()
	local tLootItems = {}
	for _, tSelectedItem in ipairs(tSelectedItems) do
		table.insert(tLootItems, tSelectedItem)
	end

	self.tMasterLoot:Randomize(tLootItems)
	self:Reset()
	self.tLootSet:SelectFirst()
end

function DistributionView:HandleOnReroll()
	self:SendEndRollToChat()
	self:HandleOnBeginRoll(true)
	self.btnReroll:Enable(false)
	self.btnReroll:Show(false)
	self.btnBeginRoll:Show(true)
end

function DistributionView:HandleOnRollerEvent_Tied(bTied)
	self.btnReroll:Enable(bTied)
	self.btnReroll:Show(bTied)
	-- self.btnBeginRoll:Enable(false)
	self.btnBeginRoll:Show(not bTied)
end

function DistributionView:HandleOnRollerEvent_NewLeader(tRaider, nScore)
	self.btnReroll:Enable(false)
	self.btnReroll:Show(false)
	self.btnBeginRoll:Show(true)
end

function DistributionView:HandleOnRoll_ChatParser(strUnitName, nScore)
	if self.bIsRolling then
		self.tRollSet:AddRoller(strUnitName, nScore)
	end
end

function DistributionView:HandleOnRollHint_ChatParser(strUnitName, tHint)
	self.tRollSet:UpdateRollHint(strUnitName, tHint)
	if (self.tSelectedRaider.tRecipient and self.tSelectedRaider.bOverwrittenSpec == nil
			and self.tSelectedRaider.tRecipient:GetFullname() == strUnitName) then
		local _, strSpec = self:GetCurrentSpec()
		self.wndSpec:SetText(strSpec)
	end
end

function DistributionView:HandleOnBeginRoll_ChatParser(strUnitName, tItem) end
function DistributionView:HandleOnEndRoll_ChatParser(strUnitName) end
function DistributionView:HandleOnParsingStarted_ChatParser() end
function DistributionView:HandleOnParsingStoped_ChatParser() end

function DistributionView:HandleOnOneSecondElapsed( ... )
	self.nElapsedTime = self.nElapsedTime + 1
	self.wndElapsedTime:SetText(G.Utils:FormatSecondsToReadable(self.nElapsedTime))
end

function DistributionView:HandleOnSelectAllItems()
	self.tLootSet:SelectAll()
end

function DistributionView:HandleOnUnselectAllItems()
	self.tLootSet:UnselectAll()
end

----------------------------------------------------------------------------------
-- Private functions
----------------------------------------------------------------------------------
function DistributionView:UpdateRecipients()
	local tSelectedItems = self.tLootSet:GetSelectedItems()
	local tEligibleRecipients = self:GetIntersectionForEligibleRecipients(tSelectedItems)
	local bShouldKeepSelection = true

	self.tRecipientSet:Update(tEligibleRecipients, bShouldKeepSelection)
end

function DistributionView:SendBeginRollToChat(bIsReroll, strTieString, bIsAscending)
	local tLootItem    = self.tSelectedLootItem.tItem
	local nChNum       = G.Debug.bIsTest and ChatSystemLib.ChatChannel_Say or ChatSystemLib.ChatChannel_Party
	local tChannel     = ChatSystemLib.GetChannels()[nChNum]
	local strPrefix    = "Roll for: "
	local strAscending = ""
	local strItem      = tLootItem:GetItem():GetChatLinkString()

	if bIsReroll then
		tChannel:Send(strTieString)
		strPrefix = "Reroll for: "
	end

	if bIsAscending then
		strAscending = " (lowest wins)"
	end

	-- Push the accepted format for a begin roll to the party channel.
	tChannel:Send(strPrefix .. strItem .. strAscending)
end

function DistributionView:EndRoll(bSkipAutoSelection)
	if self.bIsRolling then
		self.bIsRolling = false
		self.tElapsedTimer:Stop()
		self.tRollSet:EndRolling()
		self.tLootSet:Unlock()
		self.wndItemStop:Show(false)
		self.btnUnselectAll:Enable(true)
		self.btnSelectAll:Enable(true)
		self.btnReroll:Enable(false)
		self.btnReroll:Show(false)
		self.btnBeginRoll:Enable(true)
		self.btnBeginRoll:Show(true)
		self.btnEndRoll:Enable(false)
		self:SetFiltersEnable(true)

		if not bSkipAutoSelection then
			self:AutoSelectTopRoller()
		end

		self:SendEndRollToChat()
	end
end

function DistributionView:AutoSelectTopRoller()
	local tRoller, bIsTied = self.tRollSet:GetDistictTop()
	if tRoller ~= nil and bIsTied ~= nil and not bIsTied then
		local strUnitName = tRoller:GetName()
		local tRaider     = self.tRecipientSet:GetRaider(strUnitName)
		if tRaider then
			tRaider:Select()
		end
	end
end

function DistributionView:SendEndRollToChat()
	local nChNum     = G.Debug.bIsTest and ChatSystemLib.ChatChannel_Say or ChatSystemLib.ChatChannel_Party
	local tChannel   = ChatSystemLib.GetChannels()[nChNum]
	local strSummary = self.tRollSet:GetSummaryString()

	-- Push the accepted format for a end roll to the party channel.
	tChannel:Send(strSummary)
end

function DistributionView:Reset()
	self.tLootSet:UnselectAll()
	self.tRecipientSet:Clear()
	self.tSelectedRaider.tRecipient = nil
	self.tSelectedRaider.bOverwrittenSpec = nil
	self.tSelectedLootItem.tItem = nil
	self:DisplayNoItemSelected()
end

function DistributionView:DisplayNoItemSelected()
	if self.tSelectedLootItem ~= nil then
		self.tSelectedLootItem.wndName:SetText("No item selected")
		self.tSelectedLootItem.wndName:SetTextColor("White")
		self.tSelectedLootItem.wndIcon:SetSprite("achievements:sprAchievements_Icon_Group")
	end

	self.btnBeginRoll:Enable(false)
	self.btnRandomize:Enable(false)
	self.btnAssign:Enable(false)
	self.btnAssingMinor:Enable(false)
	self.btnAssingMajor:Enable(false)
	self.btnSpec:Show(false)
	self.wndSpec:Show(false)
	self.btnEndRoll:Enable(false)
	self.wndElapsedTime:SetText("-")
end

function DistributionView:DisplayMultipleItemsSelected()
	self.tSelectedLootItem.bIsMultiple = true
	self.tSelectedLootItem.tItem       = nil
	self.tSelectedLootItem.wndName:SetText("Multiple items selected.")
	self.tSelectedLootItem.wndName:SetTextColor("White")
	self.tSelectedLootItem.wndIcon:SetSprite("achievements:sprAchievements_Icon_Group")

	self.btnBeginRoll:Enable(false)
	self.btnEndRoll:Enable(false)
	self.btnRandomize:Enable(true)
end

function DistributionView:GetIntersectionForEligibleRecipients(tSelectedItems)
	if #tSelectedItems == 0 then
		return {}
	end

	local tEligibleRecipients = tSelectedItems[1]:GetEligibleRecipients()
	for _, tSelectedItem in ipairs(tSelectedItems) do
		tEligibleRecipients = G.Utils:GetIntersection(
			tEligibleRecipients,
			tSelectedItem:GetEligibleRecipients(),
			function(lhs, rhs)
				return lhs.strUnitName == rhs.strUnitName
			end
		)
	end
	return tEligibleRecipients
end

function DistributionView:DisplaySingleItemSelect(tSelectedItem)
	local strTextColor = G.Utils:GetItemQualityString(tSelectedItem)
	local wndItemName  = self.tSelectedLootItem.wndName
	local wndItemIcon  = self.tSelectedLootItem.wndIcon

	wndItemName:SetText(tSelectedItem:GetName())
	wndItemName:SetTextColor(strTextColor)
	wndItemIcon:SetSprite(tSelectedItem:GetIcon())

	self.btnAssign:Enable(self.tSelectedRaider.tRecipient ~= nil)
	self.btnAssingMinor:Enable(self.tSelectedRaider.tRecipient ~= nil)
	self.btnAssingMajor:Enable(self.tSelectedRaider.tRecipient ~= nil)
	self.btnSpec:Show(self.tSelectedRaider.tRecipient ~= nil)
	self.wndSpec:Show(self.tSelectedRaider.tRecipient ~= nil)
	self.btnRandomize:Enable(true)
	self.btnBeginRoll:Enable(true)

	local _, strSpec = self:GetCurrentSpec()
	self.wndSpec:SetText(strSpec)

	local tItem = tSelectedItem:GetItem()
	if tItem ~= nil then
		Tooltip.GetItemTooltipForm(self, wndItemIcon, tItem, {
			bPrimary    = true,
			bSelling    = false,
			itemCompare = tItem
		})
	end
end

function DistributionView:GetSelectedRecipient() return self.tRecipientSet:GetSelectedRecipient() end
function DistributionView:GetLootItem(nLootId) return self.tLootSet:GetItemWithId(nLootId) end
function DistributionView:GetLootItems() return self.tLootSet end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function DistributionView:OnRollerClicked(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnRollerClicked", ...) end
function DistributionView:OnLootUpdate(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnLootUpdate", ...) end
function DistributionView:OnLootItemSelectionChanged(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnLootItemSelectionChanged", ...) end
function DistributionView:OnRecipientSelectionChanged(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnRecipientSelectionChanged", ...) end
function DistributionView:OnAssignLoot(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnAssignLoot", ...) end
function DistributionView:OnAssignMinorLoot(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnAssignMinorLoot", ...) end
function DistributionView:OnAssignMajorLoot(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnAssignMajorLoot", ...) end
function DistributionView:OnSwitchSpec(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnSwitchSpec", ...) end
function DistributionView:OnBeginRoll()
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnBeginRoll") end
function DistributionView:OnEndRoll()
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnEndRoll") end
function DistributionView:OnRandomizeLoot(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnRandomizeLoot", ...) end
function DistributionView:OnReroll(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnReroll", ...) end
function DistributionView:OnRollerEvent_Tied(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnRollerEvent_Tied", ...) end
function DistributionView:OnRollerEvent_NewLeader(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnRollerEvent_NewLeader", ...) end
function DistributionView:OnClearLatestRolles(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnClearLatestRolles", ...) end
function DistributionView:OnRoll_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnRoll_ChatParser", ...) end
function DistributionView:OnBeginRoll_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnBeginRoll_ChatParser", ...) end
function DistributionView:OnEndRoll_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnEndRoll_ChatParser", ...) end
function DistributionView:OnRollHint_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnRollHint_ChatParser", ...) end
function DistributionView:OnParsingStarted_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnParsingStarted_ChatParser", ...) end
function DistributionView:OnParsingStoped_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnParsingStoped_ChatParser", ...) end
function DistributionView:OnOneSecondElapsed(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnOneSecondElapsed", ...) end
function DistributionView:OnSelectAllItems(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnSelectAllItems", ...) end
function DistributionView:OnUnselectAllItems(...)
	G.Utils:InvokeEventHandler(self, "DistributionView", "OnUnselectAllItems", ...) end
function DistributionView:OnSortChange(...)
	G.Utils:FireEvent(self.tLootSet, "OnSortChange", ...) end
function DistributionView:OnFilterEquippable(...)
	G.Utils:FireEvent(self.tLootSet, "OnFilterEquippable", ...) end
function DistributionView:OnFilterCommodity(...)
	G.Utils:FireEvent(self.tLootSet, "OnFilterCommodity", ...) end
function DistributionView:OnFilterRecipients(...)
	G.Utils:FireEvent(self.tRecipientSet, "OnFilterRecipients", ...) end
function DistributionView:OnFilterChange(...)
	G.Utils:FireEvent(self.tLootSet, "OnFilterChange", ...)
	G.Utils:FireEvent(self.tRecipientSet, "OnFilterChange", ...)
end
