--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ChatSystemLib = require "ChatSystemLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- ItemRollDialog
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local ItemRollDialog = {}
G.ItemRollDialog = ItemRollDialog
local RollTypeEnum = {
	MainSpec = 1,
	OffSpec  = 2,
	Costume  = 3
}

function ItemRollDialog:New(tRaidLooter, tItem)
	local o = {
		wndHandle       = nil,
		wndItemInset    = nil,
		wndExItemBorder = nil,
		wndItemSparks   = nil,
		wndItemName     = nil,
		wndItemType     = nil,
		bIsValid        = true,
		tLootItem       = nil,
		tItem           = tItem,
		eRollType       = RollTypeEnum.MainSpec,
		btnMainSpec     = nil,
		btnOffSpec      = nil,
		btnCostume      = nil,
		tRaidLooter     = tRaidLooter,
		tSettings       = tRaidLooter:GetSettings(),
		tLinker         = G.Linker:New()
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function ItemRollDialog:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	self.wndHandle    = self.tRaidLooter:CreateTopLevelWindow("LootItem_RollDialog", self)
	self.wndHandle:FindChild("Controls:PassButton"):SetText(G.Locale:Translate("Pass"))
	self.wndItemInset = self.wndHandle:FindChild("ItemInset")
	self.tLootItem    = G.LootItem:New(self.tRaidLooter, self.wndItemInset, self.tItem, false)
	self.tLootItem:Init(self)
	self.bIsValid     = self.tLootItem:IsEligible(G.Utils:GetPlayerName())

	if self.bIsValid then
		self.wndExItemBorder = self.wndHandle:FindChild("ItemToRollFor")
		self.wndItemSparks   = self.wndHandle:FindChild("Sparks")
		self.wndItemName     = self.wndHandle:FindChild("ItemName")
		self.wndItemType     = self.wndHandle:FindChild("ItemType")
		self.btnMainSpec     = self.wndHandle:FindChild("MainSpecButton")
		self.btnOffSpec      = self.wndHandle:FindChild("OffspecButton")
		self.btnCostume      = self.wndHandle:FindChild("CostumeButton")

		local strTextColor = G.Utils:GetItemQualityString(self.tLootItem)
		local strExSprite  = G.Utils:GetItemExtendedRarity(self.tLootItem)

		self.wndItemType:SetText(self.tLootItem:GetItem():GetItemTypeName())
		self.wndItemName:SetText(self.tLootItem:GetName())
		self.wndItemName:SetTextColor(strTextColor)
		self.wndExItemBorder:SetSprite(strExSprite)
		self.btnMainSpec:SetCheck(true)
		self.btnMainSpec:SetBGColor(strTextColor)
		self.btnOffSpec:SetBGColor(strTextColor)
		self.btnCostume:SetBGColor(strTextColor)

		self.tLootItem:SetCompareFunction(function(tItem)
			return tItem:GetEquippedItemForItemType()
		end)
		self.tLootItem:GenerateTooltip()
		self.wndHandle:Show(true)
	else
		self:Destroy()
	end

	G.Utils:PlayNewRollSound()
end

function ItemRollDialog:Destroy()
	if self.tLinker ~= nil then
		self.tLinker:Destroy()
		self.wndHandle:Destroy()
		self.tLootItem:Destroy()

		self.tLinker     = nil
		self.wndHandle   = nil
		self.tLootItem   = nil
		self.tRaidLooter = nil
		self.tSettings   = nil
	end
end

----------------------------------------------------------------------------------
-- Getters & Setters
----------------------------------------------------------------------------------
function ItemRollDialog:IsValid() return self.bIsValid end

function ItemRollDialog:SetAnchor(tAnchor)
	local tP = tAnchor.tPoints
	local tO = tAnchor.tOffsets

	self.wndHandle:SetAnchorPoints(tP[1], tP[2], tP[3], tP[4])
	self.wndHandle:SetAnchorOffsets(tO[1], tO[2], tO[3], tO[4])
end

function ItemRollDialog:GetSize()
	return {
		nWidth = self.wndHandle:GetWidth(),
		nHeight = self.wndHandle:GetHeight()
	}
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function ItemRollDialog:HandleOnRollTypeChanged(wndHandler, wndControl, eMouseButton)
	if wndControl:IsChecked() then
		local strButtonName = wndControl:GetName()

		if strButtonName == "MainSpecButton" then
			self.eRollType = RollTypeEnum.MainSpec
		elseif strButtonName == "OffspecButton" then
			self.eRollType = RollTypeEnum.OffSpec
		elseif strButtonName == "CostumeButton" then
			self.eRollType = RollTypeEnum.Costume
		else
			self.eRollType = RollTypeEnum.MainSpec
		end
	end
end

function ItemRollDialog:HandleOnItemDialogClosed()
	self:Destroy()
end

function ItemRollDialog:HandleOnRollForItem()
	local strRollMax = (self.eRollType == RollTypeEnum.OffSpec and " 50"
		or self.eRollType == RollTypeEnum.Costume and " 33" or "")
	ChatSystemLib.Command(G.Locale:TranslateCommand("Roll")..strRollMax)
	-- local nChNum     = G.Debug.bIsTest and ChatSystemLib.ChatChannel_Say or ChatSystemLib.ChatChannel_Party
	-- local tChannel   = ChatSystemLib.GetChannels()[nChNum]
	--
	-- if self.eRollType == RollTypeEnum.OffSpec then
	-- 	tChannel:Send("os")
	-- elseif self.eRollType == RollTypeEnum.Costume then
	-- 	tChannel:Send("costume")
	-- end

	self:Destroy()
	G.Utils:PlaySound(174)
end

function ItemRollDialog:HandleOnPassForItem()
	self:Destroy()
	G.Utils:PlaySound(200)
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function ItemRollDialog:OnRollTypeChanged(...)
	G.Utils:InvokeEventHandler(self, "ItemRollDialog", "OnRollTypeChanged", ...) end
function ItemRollDialog:OnItemDialogClosed(...)
	G.Utils:InvokeEventHandler(self, "ItemRollDialog", "OnItemDialogClosed", ...) end
function ItemRollDialog:OnRollForItem(...)
	G.Utils:InvokeEventHandler(self, "ItemRollDialog", "OnRollForItem", ...) end
function ItemRollDialog:OnPassForItem(...)
	G.Utils:InvokeEventHandler(self, "ItemRollDialog", "OnPassForItem", ...) end
