--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ApolloTimer = require "ApolloTimer"
local ChatSystemLib = require "ChatSystemLib"
local Tooltip = require "Tooltip"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Loot Item
-- Represents an undistributed item.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local LootItem = {}
G.LootItem = LootItem

local RollTypeEnum = {
	MainSpec = 1,
	OffSpec  = 2,
	Costume  = 3,
	Pass = nil,
}

function LootItem:New(tRaidLooter, wndParent, tLootItem, bIsSelectable)
	local o = {
		tGenTooltipTimer = nil,
		tCallback        = nil,
		tDrop            = tLootItem,
		tItem            = tLootItem.itemDrop,
		tLooters         = tLootItem.tLooters,
		tLootersOoR      = tLootItem.tLootersOutOfRange,
		tDetails         = {
			nDropTime = os.time(),
			bIsQueued = false,
			strRecipient = nil,
			eAutoRoll = RollTypeEnum.Pass,
		},
		nUUID            = tLootItem.nLootId,
		wndParent        = wndParent,
		wndHandle        = nil,
		wndHighlight     = nil,
		wndBorder        = nil,
		wndIcon          = nil,
		wndInlineInfo    = nil,
		wndTooltip       = nil,
		wndTooltipComp   = nil,
		wndLabel         = nil,
		bIsGenTooltip    = false,
		bIsSelectable    = bIsSelectable,
		bIsSelected      = false,
		bIsRolledFor     = false,
		bIsLocked        = false,
		btn              = nil,
		fCompare         = function() return nil end,
		tRaidLooter      = tRaidLooter,
		tSettings        = tRaidLooter:GetSettings(),
		tLinker          = G.Linker:New(),
		strSortFuncName  = "ByLevel"
	}

	setmetatable(o, self)
	self.__index = self
	self.__lt = function(a, b) return o:LessThen(a, b) end
	return o
end

function LootItem:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	self.wndHandle         = self.tRaidLooter:CreateWindow("LootItem_Square", self.wndParent, self)
	self.wndHighlight      = self.wndHandle:FindChild("Highlight")
	self.wndBorder         = self.wndHandle:FindChild("Border")
	self.wndIcon           = self.wndHandle:FindChild("Icon")
	self.wndQueueHighlight = self.wndHandle:FindChild("QueueHighlight")
	self.wndLabel          = self.wndHandle:FindChild("LabelAutoloot")
	self.btn               = self.wndHandle:FindChild("Button")

	self.wndQueueHighlight:Show(false, true)
	self.wndHighlight:Show(false, true)
	self.wndIcon:SetSprite(self:GetIcon())
	self.wndBorder:SetSprite(G.Utils:GetItemRarity(self))

	self.btn:AddEventHandler("MouseButtonUp", "OnMouseUp", self)

	-- Events
  if self.bIsSelectable then
		self.btn:AddEventHandler("ButtonSignal", "OnButtonPressed", self)
	else
		self.wndBorder:SetBGColor("ffffffff")
	end

	self.tEligibleRecipients = G.Utils:MergeLootersForDrop(self.tDrop)
	self:RestoreDetails()
end

function LootItem:Destroy()
	self.tLinker:Destroy()
	self.wndHandle:Destroy()

	if self.wndTooltip ~= nil then
		self.wndTooltip:Destroy()
	end

	if self.wndTooltipCompand ~= nil then
		self.wndTooltipCompand:Destroy()
	end

	if self.bIsGenTooltip then
		self.tGenTooltipTimer:Stop()
	end

	if self.btn ~= nil then
		self.btn:RemoveEventHandler("ButtonSignal", self)

		if self.bIsSelectable then
			self.btn:RemoveEventHandler("ButtonSignal", self)
		end
	end

	self.tGenTooltipTimer    = nil
	self.tLinker             = nil
	self.wndHandle           = nil
	self.wndParent           = nil
	self.wndTooltip          = nil
	self.wndTooltipCompand   = nil
	self.wndLabel            = nil
	self.btn                 = nil
	self.tRaidLooter         = nil
	self.tSettings           = nil
	self.tEligibleRecipients = nil
	self.tDrop               = nil
	self.tItem               = nil
	self.tLooters            = nil
	self.tLootersOoR         = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function LootItem:RestoreDetails()
	-- Look through the saved settings and restore the details if possible.
	local tLootDetails = self.tSettings:ReadBlock("tLootDetails") -- Get ref
	if tLootDetails ~= nil then
		local tDetails = tLootDetails[self.nUUID]
		if tDetails ~= nil then
			self.tDetails = tDetails
		else
			tLootDetails[self.nUUID] = self.tDetails
		end
	end

	if self.tDetails.bIsQueued and self.tDetails.strRecipient ~= nil then
		self.tRaidLooter:GetLootQueue():Push(self, self.tDetails.strRecipient)
	end
end

function LootItem:ClearDetails()
	-- Clears the details from the settings.
	local tLootDetails = self.tSettings:ReadBlock("tLootDetails") -- Get ref
	if tLootDetails ~= nil then
		tLootDetails[self.nUUID] = nil
		self.tDetails = nil
	end
end

function LootItem:Update(tLootItem)
	self.tLooters    = tLootItem.tLooters
	self.tLootersOoR = tLootItem.tLootersOutOfRange

	self.tEligibleRecipients = G.Utils:MergeLootersForDrop(tLootItem)
end

function LootItem:GenerateTooltip()
	-- No need to do anything if generating tooltip is ongoing.
	if self.bIsGenTooltip then
		return
	end

	-- In case we have a valid tooltip nothing more can be done.
	if self.wndTooltip ~= nil and self.wndTooltip:IsValid() then
		return
	end

	-- Beware --
	-- Calling 'OnGenerateTooltip' from here will not work "to earily"!
	-- Postponing the generation to be post 'OnLoad' works. The timer is
	-- unlikely to run more then once for each item.
	--
	-- Also note this workaround only affect tooltips for loot that exists
	-- when the addon is loaded (for instance if you do /reloadui while having
	-- undistributed loot).
	self.bIsGenTooltip = true
	self.tGenTooltipTimer = ApolloTimer.Create(1, true, "OnGenerateTooltip", self)
end

function LootItem:LinkToChat()
	local nChNum   = G.Debug.bIsTest and ChatSystemLib.ChatChannel_Say or ChatSystemLib.ChatChannel_Party
	local tChannel = ChatSystemLib.GetChannels()[nChNum]
	local strItem  = self.tItem:GetChatLinkString()

	tChannel:Send(strItem) --strTieString
end

function LootItem:Unselect(bShouldPreventEvent)
	if self.bIsSelectable and not self.bIsLocked then
		self.wndHighlight:Show(false, false)
		self.bIsSelected = false

		if not bShouldPreventEvent then
			G.Utils:FireEvent(self.tCallback, "OnItemUnselected", self)
		end
	end
end

function LootItem:Select(bShouldPreventEvent)
	if self.bIsSelectable and not self.bIsLocked then
		self.wndHighlight:Show(true, false)
		self.bIsSelected = true

		if not bShouldPreventEvent then
			G.Utils:PlaySound(86)
			G.Utils:FireEvent(self.tCallback, "OnItemSelected", self)
		end
	elseif self.bIsLocked then
		G.Utils:PlaySound(220)
	end
end

function LootItem:IsEligible(strUnitName)
	for nIndex, tLooterUnit in ipairs(self.tLooters) do
		if tLooterUnit:GetName() == strUnitName then
			return true
		end
	end

	return false
end

function LootItem:Show()
	self.wndHandle:Enable(true)
	self.wndHandle:Show(true)
end

function LootItem:Hide()
	self.wndHandle:Enable(false)
	self.wndHandle:Show(false)
end

function LootItem:IsOfType(strType, bEquippable, bCommodity)
	-- This is a very specific function used for filtering.
	if bCommodity and self.tItem:IsCommodity() then
		return true
	end

	local strLight = G.Locale:Translate("Light")
	local strMedium = G.Locale:Translate("Medium")
	local strHeavy = G.Locale:Translate("Heavy")

	local tTypes = { L = strLight, M = strMedium, H = strHeavy }
	local strShouldInclude = tTypes[strType]
	local tShouldExclude = {}
	if strType == "L" then
		table.insert(tShouldExclude, strMedium)
		table.insert(tShouldExclude, strHeavy)
	elseif strType == "M" then
		table.insert(tShouldExclude, strLight)
		table.insert(tShouldExclude, strHeavy)
	elseif strType == "H" then
		table.insert(tShouldExclude, strMedium)
		table.insert(tShouldExclude, strLight)
	end

	if strShouldInclude == nil then
		return true
	end

	local strTypeName = self.tItem:GetItemTypeName()
	for _, strShouldExclude in ipairs(tShouldExclude) do
		if string.find(strTypeName, strShouldExclude) then
			return false
		end
	end

	return bEquippable and self.tItem:IsEquippable() or string.find(strTypeName, strShouldInclude)
end

function LootItem:ByLevelSortFunction(a, b)
	local bIsEqual = a:GetQuality() == b:GetQuality()

	if bIsEqual then
		return a:GetItemLevel() > b:GetItemLevel()
	end

	return a:GetQuality() > b:GetQuality()
end

function LootItem:ByNameSortFunction(a, b)
	return a:GetName():lower() <  b:GetName():lower()
end

function LootItem:LessThen(a, b)
	if self.strSortFuncName == "ByLevel" then
		return self:ByLevelSortFunction(a, b)
	else
		return self:ByNameSortFunction(a, b)
	end
end

function LootItem:SetSortRule(strSortFuncName)
	self.strSortFuncName = strSortFuncName
end

function LootItem:SetAutoRoll(eAutoRoll)
	self.tDetails.eAutoRoll = eAutoRoll
	if eAutoRoll == RollTypeEnum.MainSpec then
		self.wndLabel:SetText("M")
	elseif eAutoRoll == RollTypeEnum.OffSpec then
		self.wndLabel:SetText("O")
	elseif eAutoRoll == RollTypeEnum.Costume then
		self.wndLabel:SetText("C")
	else
		self.wndLabel:SetText("")
	end
end

function LootItem:GetAutoRoll()
	return self.tDetails.eAutoRoll
end

----------------------------------------------------------------------------------
-- Getters & Setters
----------------------------------------------------------------------------------
function LootItem:GetTypeName() return self.tItem:GetItemTypeName() end
function LootItem:GetDropTime() return self.tDetails.nDropTime end
function LootItem:GetDrop() return self.tDrop end
function LootItem:GetId() return self.nUUID end
function LootItem:HaveId(nId) return self.nUUID == nId end
function LootItem:GetName() return self.tItem:GetName() end
function LootItem:GetItemTypeName() return self:GetItemTypeName() end
function LootItem:GetIcon() return self.tItem:GetIcon() end
function LootItem:GetQuality() return self.tItem:GetItemQuality() end
function LootItem:GetPower() return self.tItem:GetItemPower() end
function LootItem:GetItemLevel() return self.tItem:GetEffectiveLevel() end
function LootItem:GetItem() return self.tItem end
function LootItem:GetQueueInfo() return self.tDetails.strRecipient end
function LootItem:GetItemId() return self.tItem:GetItemId() end
function LootItem:IsRolledFor() return self.bIsRolledFor end
function LootItem:IsSelected() return self.bIsSelected end
function LootItem:GetEligibleRecipients() return self.tEligibleRecipients end
function LootItem:SetLock(bShouldLock) self.bIsLocked = bShouldLock or self.tDetails.bIsQueued end
function LootItem:IsQueued() return self.tDetails.bIsQueued end
function LootItem:SetCompareFunction(fCompare) self.fCompare = fCompare end
function LootItem:SetIsRolledFor(bVal) self.bIsRolledFor = bVal end
function LootItem:SetCallback(tCallback) self.tCallback = tCallback end

function LootItem:SetAnchor(tAnchor)
	self.wndHandle:SetAnchorPoints(unpack(tAnchor.tPoints))
	self.wndHandle:SetAnchorOffsets(unpack(tAnchor.tOffsets))
end

function LootItem:GetStackSize()
	local tItemDetails = self.tItem:GetDetailedInfo()
	if tItemDetails ~= nil and tItemDetails.tPrimary ~= nil and tItemDetails.tPrimary.tStack then
		return tItemDetails.tPrimary.tStack.nCount
	end

	return 1
end

function LootItem:GetSize()
	return {
		nWidth  = self.wndHandle:GetWidth(),
		nHeight = self.wndHandle:GetHeight()
	}
end

function LootItem:SetIsQueued(bIsQueued, strRecipient)
	self.tDetails.bIsQueued = bIsQueued
	self.tDetails.strRecipient = strRecipient
	self:Unselect()
	self:SetLock(bIsQueued)
	self.wndQueueHighlight:Show(bIsQueued)
	self.wndQueueHighlight:Enable(bIsQueued)
	if bIsQueued and strRecipient ~= nil then
		local strTip = G.Utils:IsLeader() and "Right click and select \"Unqueue\" to unqueue the item." or ""
		self.wndQueueHighlight:SetTooltip("This item is currently queued for "..strRecipient..". "..strTip)
	else
		self.wndQueueHighlight:SetTooltip("")
	end

	if bIsQueued then
		if self.wndTooltip ~= nil then
			self.wndTooltip:Destroy()
		end
		self.wndTooltip = nil

		if self.wndTooltipComp ~= nil then
			self.wndTooltipComp:Destroy()
		end
		self.wndTooltipComp = nil
	else
		self:GenerateTooltip()
	end
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function LootItem:HandleOnGenerateTooltip()
	self.wndTooltip, self.wndTooltipComp = Tooltip.GetItemTooltipForm(self, self.btn, self.tItem, {
		bPrimary    = true,
		bSelling    = false,
		itemCompare = self.fCompare(self.tItem)
	})

	if self.wndTooltip ~= nil and self.wndTooltip:IsValid() then
		self.tGenTooltipTimer:Stop()
		self.tGenTooltipTimer = nil
		self.bIsGenTooltip = false
	end
end

function LootItem:HandleOnButtonPressed(wndHandler, wndControl, eMouseButton)
	if self.bIsSelected then
		self:Unselect()
	else
		self:Select()
	end
end

function LootItem:HandleOnMouseUp(wndHandler, wndControl, eMouseButton)
	self.tRaidLooter:ShowContextMenu(self)
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function LootItem:OnGenerateTooltip(...) G.Utils:InvokeEventHandler(self, "LootItem", "OnGenerateTooltip", ...) end
function LootItem:OnButtonPressed(...) G.Utils:InvokeEventHandler(self, "LootItem", "OnButtonPressed", ...) end
function LootItem:OnMouseUp(...) G.Utils:InvokeEventHandler(self, "LootItem", "OnMouseUp", ...) end
