--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Loot Item Collection
-- Represents a collection of undistributed items.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local LootItemCollection = {}
G.LootItemCollection = LootItemCollection

function LootItemCollection:New(tRaidLooter, shouldCompareToSelf, bIsSelectable, wndTarget)
	local o = {
		bShouldCompareToSelf = shouldCompareToSelf,
		bIsSelectable        = bIsSelectable,
		bShouldPreventEvents = false,
		tSelectedItems       = {},
		tItems               = {},
		tCallback            = nil,
		wndHandle            = wndTarget,
		fCompare             = function() return nil end,
		tRaidLooter          = tRaidLooter,
		tSettings            = tRaidLooter:GetSettings(),
		tMasterLoot          = tRaidLooter:GetMasterLoot(),
		tAutoRoll            = tRaidLooter:GetAutoRoll(),
		tLoot                 = {},
		tLinker                = G.Linker:New(),
		bIsMultiSelectionForced = false,
		strCurrentFilter       = "A",
		strCurrentSort        = "ByLevel",
		bFilterEquipable     = true,
		bFilterCommodity    = true
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function LootItemCollection:Init(tTarget)
	self.tLinker:Link(NLLC, self)
	self.tMasterLoot:RegisterCallback(self)
	self.tAutoRoll:RegisterLootItemCollection(self)
end

function LootItemCollection:Destroy()
	self.tMasterLoot:UnregisterCallback(self)
	self.tAutoRoll:UnregisterLootItemCollection(self)
	self.tLinker:Destroy()
	for _, tItem in pairs(self.tItems) do
		tItem:Destroy()
	end

	self.tSelectedItems = nil
	self.tItems         = nil
	self.tCallback      = nil
	self.wndHandle      = nil
	self.tMasterLoot    = nil
	self.tRaidLooter    = nil
	self.tSettings      = nil
	self.tCallback      = nil
	self.tLoot          = nil
	self.tAutoRoll      = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function LootItemCollection:AddItem(tItem)
	tItem:SetSortRule(self.strCurrentSort)

	local nNewLastIndex = #self.tItems + 1 -- New last index
    local nInsertIndex = self:GetNextIndex(tItem, nNewLastIndex)

    table.insert(self.tItems, nInsertIndex, tItem)
end

function LootItemCollection:RemoveItem(tItem, bPreventArrange)
	-- This item must lose its selection.
	tItem:Unselect()

	for nIndex, tExistingItem in ipairs(self.tItems) do
		if tItem:GetId() == tExistingItem:GetId() then
			table.remove(self.tItems, nIndex)
			tItem:ClearDetails()
			tItem:Destroy()

			if not bPreventArrange then
				self:ArrangeAndFilter()
			end

			break
		end
	end
end

function LootItemCollection:RemoveAll()
	self:UnselectAll()

	for nIndex, tItem in ipairs(self.tItems) do
		tItem:Destroy()
	end

	self.tItems = {}
end

function LootItemCollection:ApplyFilter()
	self:ArrangeAndFilter()
	self:UnselectAll()
end

function LootItemCollection:ArrangeAndFilter()
	if G.Utils:IsTableEmpty(self.tItems) then
		return
	else
		-- (Pre)calculate some variables
		local nSW     = 6                                          -- Estiamted width of the scroll bar.
		local nWidth  = self.wndHandle:GetWidth()
		-- local nHeight = self.wndHandle:GetHeight()
		local nSize   = self.tItems[1]:GetSize()                   -- Assumes equal sized items.
		local nIpr    = math.floor(nWidth / nSize.nWidth)          -- Items Per Row
		-- local nIpc    = math.floor(nHeight / nSize.nHeight)        -- (Visible) Items Per Column
		local nPadH   = 0.5 * (nWidth - nSize.nWidth * nIpr - nSW) -- Horizontal padding
		-- local nPadV   = 0.5 * (nHeight - nSize.nHeight * nIpc)     -- Vertical padding (Not used now)
		local nSkipCount = 0

		for nIndex, tItem in pairs(self.tItems) do
			nIndex = nIndex - 1 - nSkipCount

			if tItem:IsOfType(self.strCurrentFilter, not self.bFilterEquipable, not self.bFilterCommodity) then
				local nRow = math.floor(nIndex / nIpr)
				local nCol = nIndex % nIpr

				-- Offsets for readability
				local nLeft   = nCol * (nSize.nWidth + nPadH)
				local nTop    = nRow * (nSize.nHeight - 4)
				local nRight  = nLeft + nSize.nWidth
				local nBottom = nTop + nSize.nHeight

				tItem:SetAnchor({
					tPoints  = {0, 0, 0, 0},
				    tOffsets = {nLeft, nTop, nRight, nBottom}
				})
				tItem:Show()
			else
				tItem:Hide()
				nSkipCount = nSkipCount + 1
			end
		end
	end
end

function LootItemCollection:ContainsItemWithUUID(nUUID)
	for nIndex, tExistingItem in pairs(self.tItems) do
		if tExistingItem.nUUID == nUUID then
			return true
		end
	end
	return false
end

function LootItemCollection:GetItemWithUUID(nUUID)
	for nIndex, tExistingItem in pairs(self.tItems) do
		if tExistingItem.nUUID == nUUID then
			return tExistingItem
		end
	end
	return nil
end

-- function LootItemCollection:FindItem

function LootItemCollection:Lock()
	for nIndex, tItem in ipairs(self.tItems) do
		tItem:SetLock(true)
	end
end

function LootItemCollection:Unlock()
	for nIndex, tItem in ipairs(self.tItems) do
		tItem:SetLock(false)
	end
end

function LootItemCollection:SetCallback(tCallback)
	self.tCallback = tCallback
end

function LootItemCollection:GetSelectedItems()
	local tSelectedItems = {}
	for _, tItem in ipairs(self.tSelectedItems) do
		table.insert(tSelectedItems, tItem)
	end
	return tSelectedItems
end

function LootItemCollection:SelectAll()
	self.bIsMultiSelectionForced = true
	self:RunInEventProtectedScope(function()
		for nIndex, tItem in ipairs(self.tItems) do
			if not tItem:IsSelected() then
				tItem:Select()
			end
		end
	end)
	self.bIsMultiSelectionForced = false
end

function LootItemCollection:UnselectAll()
	self:RunInEventProtectedScope(function()
		for nIndex, tSelectedItem in ipairs(self.tSelectedItems) do
			tSelectedItem:Unselect(true) -- Prevent events
		end

		self.tSelectedItems = {}
	end)
end

function LootItemCollection:SelectFirst()
	if #self.tItems > 0 then
		self:RunInEventProtectedScope(function()
			self:UnselectAll()
			for _, tItemToSelect in ipairs(self.tItems) do
				tItemToSelect:Select(true) -- Prevent event

				if tItemToSelect:IsSelected() then
					table.insert(self.tSelectedItems, tItemToSelect)
					return
				end
			end
		end)
	end
end

function LootItemCollection:IsMultiSelectionAllowed()
	return self.bIsMultiSelectionForced or Apollo.IsControlKeyDown()
end

function LootItemCollection:SetCompareFunction(fCompare)
	self.fCompare = fCompare
end

-- Protects against multiple event invocations during fRun. If an event is emitted it
-- happens on the end. Note RunInEventProtectedScope may be called nested
function LootItemCollection:RunInEventProtectedScope(fRun, ...)
	local bShouldPreventEvent = self.bShouldPreventEvents
	self.bShouldPreventEvents = true

	-- Block all events generated by fRun.
	fRun(...)

	self.bShouldPreventEvents = bShouldPreventEvent

	-- Attempt fire an event.
	self:Notify("OnLootItemSelectionChanged", self.tSelectedItems)
end

function LootItemCollection:Notify(eventName, ...)
	if (not self.bShouldPreventEvents) and self.tCallback ~= nil then
		local func = self.tCallback[eventName]

		if func ~= nil then
			-- IMPORTANT --
			-- Make sure to block all events that may be triggered by running 'func'
			-- This will block potentail infinite recursion.
			self.bShouldPreventEvents = true
			func(self.tCallback, ...)
			self.bShouldPreventEvents = false
		end
	end
end

function LootItemCollection:GetItemWithId(nLootId)
	for _, tItem in pairs(self.tItems) do
		if tItem:HaveId(nLootId) then
			return tItem
		end
	end

	return nil
end

function LootItemCollection:GetItem(tItem)
	for _, tExistingItem in pairs(self.tItems) do
		if tExistingItem:GetItem():GetChatLinkString() == tItem:GetChatLinkString() then
			return tExistingItem
		end
	end

	return nil
end

function LootItemCollection:GetNextIndex(tItem, nLast)
	for nIndex, tExistingItem in ipairs(self.tItems) do
		if tItem < tExistingItem then
			return nIndex
		end
  end
  return nLast
end

function LootItemCollection:ApplySort()
	table.sort(self.tItems)
	self:ArrangeAndFilter()
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function LootItemCollection:HandleOnLootUpdate(tLoot)
	-- Reference all current items.
	self.tLoot = tLoot
	local tKillItems = {}
	for _, tItem in pairs(self.tItems) do
		tKillItems[tItem:GetId()] = tItem
	end

	-- Attempt to insert new items.
	for _, tLootItem in pairs(tLoot) do
		local tExistingItem = self:GetItemWithUUID(tLootItem.nLootId)

		if tExistingItem ~= nil then
			tExistingItem:Update(tLootItem)
			tKillItems[tLootItem.nLootId] = nil
		else
			local tNewLootItem = G.LootItem:New(self.tRaidLooter, self.wndHandle, tLootItem, self.bIsSelectable)
			tNewLootItem:Init(self)
			tNewLootItem:SetCompareFunction(self.fCompare)
			tNewLootItem:SetCallback(self)
			tNewLootItem:GenerateTooltip()
			tNewLootItem:SetSortRule(self.strCurrentSort)
			self:AddItem(tNewLootItem)
		end
	end

	-- Remove remaining items.
	for nKey, tItem in pairs(tKillItems) do
		self:RemoveItem(tItem, true)
	end

	-- Arrange items according to there position in self.tItems.
	self:ArrangeAndFilter()
	self.wndHandle:RecalculateContentExtents()
end

function LootItemCollection:HandleOnItemSelected(tItem)
	if self.bIsSelectable then
		self:RunInEventProtectedScope(function()
			if not self:IsMultiSelectionAllowed() then
				self:UnselectAll()
			end

			table.insert(self.tSelectedItems, tItem)
		end)
	end
end

function LootItemCollection:HandleOnItemUnselected(tItem)
	if self.bIsSelectable then
		self:RunInEventProtectedScope(function()
			if self:IsMultiSelectionAllowed() then
				for nIndex, tSelectedItem in ipairs(self.tSelectedItems) do
					if tSelectedItem == tItem then
						table.remove(self.tSelectedItems, nIndex)
						break
					end
				end
			else
				local wasMultiSelect = #self.tSelectedItems > 1

				self:UnselectAll()

				if wasMultiSelect then
					tItem:Select(false)
				end
			end
		end)
	end
end

function LootItemCollection:HandleOnSortChange(wndHandler, wndControl, eMouseButton)
	if wndControl:IsChecked() then
		local strButtonName = wndControl:GetName()
		local _, strButtonId = G.Utils:Split(strButtonName, "_")

		if strButtonId ~= nil and self.strCurrentSort ~= strButtonId then
			G.Utils:Log("Sort settings changed to: " .. strButtonId)

			self.strCurrentSort = strButtonId
			self:SetSortRule()
			self:ApplySort()
		end
	end
end

function LootItemCollection:SetSortRule()
	for nIndex, tItem in ipairs(self.tItems) do
		tItem:SetSortRule(self.strCurrentSort)
	end
end

function LootItemCollection:HandleOnFilterChange(wndHandler, wndControl, eMouseButton)
	if wndControl:IsChecked() then
		local strButtonName = wndControl:GetName()
		local _, strButtonId = G.Utils:Split(strButtonName, "_")
		local tMap = { AllButton = "A", HButton = "H", MButton = "M", LButton = "L"}
		local strFilter  = tMap[strButtonId]

		if strFilter == nil then
			strFilter = "A"
		end

		if strFilter ~= self.strCurrentFilter then
			G.Utils:Log("Filter settings changed to: " .. strFilter)

			self.strCurrentFilter = strFilter
			self:ApplyFilter()
		end
	end
end

function LootItemCollection:HandleOnFilterEquippable(wndHandler, wndControl, eMouseButton)
	self.bFilterEquipable = not wndControl:IsChecked()
	self:ApplyFilter()
end

function LootItemCollection:HandleOnFilterCommodity(wndHandler, wndControl, eMouseButton)
	self.bFilterCommodity = not wndControl:IsChecked()
	self:ApplyFilter()
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------

function LootItemCollection:OnLootUpdate(...)
	G.Utils:InvokeEventHandler(self, "LootItemCollection", "OnLootUpdate", ...) end
function LootItemCollection:OnItemSelected(...)
	G.Utils:InvokeEventHandler(self, "LootItemCollection", "OnItemSelected", ...) end
function LootItemCollection:OnItemUnselected(...)
	G.Utils:InvokeEventHandler(self, "LootItemCollection", "OnItemUnselected", ...) end
function LootItemCollection:OnSortChange(...)
	G.Utils:InvokeEventHandler(self, "LootItemCollection", "OnSortChange", ...) end
function LootItemCollection:OnFilterChange(...)
	G.Utils:InvokeEventHandler(self, "LootItemCollection", "OnFilterChange", ...) end
function LootItemCollection:OnFilterEquippable(...)
	G.Utils:InvokeEventHandler(self, "LootItemCollection", "OnFilterEquippable", ...) end
function LootItemCollection:OnFilterCommodity(...)
	G.Utils:InvokeEventHandler(self, "LootItemCollection", "OnFilterCommodity", ...) end
