--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ChatSystemLib = require "ChatSystemLib"
local Tooltip = require "Tooltip"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- LootWindow
-- Represents the loot window i.e. the window everyone can see.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local LootWindow = {}
G.LootWindow = LootWindow
local RollTypeEnum = {
	MainSpec = 1,
	OffSpec  = 2,
	Costume  = 3
}

function LootWindow:New(tRaidLooter)
	local o = {
		wndMain       = nil,
		wndRollerList = nil,
		wndRollerSort = nil,
		wndLoot       = nil,
		wndAmount     = nil,
		wndLootItemsRegion = nil,
		wndLootRollRegion = nil,
		wndRollOutset = nil,
		tItemDialog   = nil,
		tInlineItem   = {
			wnd        = nil,
			wndIcon    = nil,
	        wndName    = nil,
			btnMain    = nil,
			btnOff     = nil,
			btnCostume = nil,
			btnRoll    = nil,
			btnPass    = nil,
			eRollType  = RollTypeEnum.MainSpec,
		},
		wndOptions      = nil,
		btnOptions      = nil,
		btnExpand       = nil,
		btnRoll         = nil,
		btnTab          = nil,
		bIsRolling      = false,
		tRollSet        = nil,
		tLootCollection = nil,
		tRaidLooter     = tRaidLooter,
		tSettings       = tRaidLooter:GetSettings(),
		tMasterLoot     = tRaidLooter:GetMasterLoot(),
		tChatParser     = tRaidLooter:GetChatParser(),
		tAutoRoll       = tRaidLooter:GetAutoRoll(),
		tLinker         = G.Linker:New(),
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function LootWindow:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	self.wndMain                = self.tRaidLooter:CreateTopLevelWindow("LootWindow", self)
	self.wndMain:FindChild("InlineItemBase:Footer:ButtonControls:InlinePassButton"):SetText(G.Locale:Translate("Pass"))
	self.wndLootItemsRegion     = self.wndMain:FindChild("LootItemsRegion")
	self.wndLootRollRegion      = self.wndMain:FindChild("LootRollRegion")
	self.wndRollOutset          = self.wndMain:FindChild("RollOutset")
	self.wndLoot                = self.wndMain:FindChild("UndistributedLootContainer")
	self.wndRollerList          = self.wndMain:FindChild("Loot_RollTopList")
	self.wndRollerSort          = self.wndMain:FindChild("RollSortWindow")
	self.btnExpand              = self.wndMain:FindChild("ExpandButton")
	self.tInlineItem.wnd        = self.wndMain:FindChild("InlineItemBase")
	self.tInlineItem.wndItem    = self.wndMain:FindChild("LootInlineItemRoller")
	self.tInlineItem.wndName    = self.wndMain:FindChild("ItemRolledForName")
	self.tInlineItem.wndIcon    = self.wndMain:FindChild("ItemRolledForIcon")
	self.tInlineItem.btnMain    = self.wndMain:FindChild("Inline_MainSpecButton")
	self.tInlineItem.btnOff     = self.wndMain:FindChild("Inline_OffSpecButton")
	self.tInlineItem.btnCostume = self.wndMain:FindChild("Inline_CostumeButton")
	self.tInlineItem.btnRoll    = self.wndMain:FindChild("InlineRollButton")
	self.tInlineItem.btnPass    = self.wndMain:FindChild("InlinePassButton")
	self.wndOptions             = self.wndMain:FindChild("OptionsDialog")
	self.btnOptions             = self.wndMain:FindChild("ConfigureButton")
	self.tRollSet               = G.RollerCollection:New(self.tRaidLooter, true, self.wndRollerList, self.wndRollerSort)
	self.tLootCollection        = G.LootItemCollection:New(self.tRaidLooter, true, false, self.wndLoot)

	self.tRollSet:Init(self)
	self.tLootCollection:Init(self)
	self.tChatParser:RegisterCallback(self)

	self.btnExpand:AttachWindow(self.wndRollOutset)
	self.btnExpand:SetCheck(false)
	self.btnOptions:AttachWindow(self.wndOptions)
	self.btnOptions:SetCheck(false)
	self.tInlineItem.wnd:Show(false)
	self.tInlineItem.btnMain:SetCheck(true)

	self.tLootCollection:SetCompareFunction(function(tItem)
		return tItem:GetEquippedItemForItemType()
	end)

	self.tRaidLooter:RestoreOptions()
	G.Utils:MoveWindow(self.wndMain, self.tSettings:ReadBlock("tWindowLocations.wndLootWindowLoc"))
end

function LootWindow:Destroy()
	self.tChatParser:UnregisterCallback(self)

	self.tLinker:Destroy()
	self.wndMain:Destroy()
	self.tRollSet:Destroy()
	self.tLootCollection:Destroy()

	self.wndMain = nil
	self.tLinker = nil
	self.tRollSet = nil
	self.tLootCollection = nil
	self.tChatParser = nil
	self.tAutoRoll = nil
	self.tRaidLooter = nil
	self.tMasterLoot = nil
	self.tSettings = nil
end

----------------------------------------------------------------------------------
-- Save & Restore
----------------------------------------------------------------------------------
function LootWindow:OnSave(eLevel)
	if self.wndMain ~= nil then
		local tLocations = self.tSettings:ReadBlock("tWindowLocations")
		tLocations.wndLootWindowLoc = self.wndMain:GetLocation():ToTable()
	end
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function LootWindow:FindLootItem(tItem)
	if self.tLootCollection ~= nil then
		return self.tLootCollection:GetItem(tItem)
	end
	return nil
end

function LootWindow:Hide(bImmediate)
	if bImmediate == nil then
		bImmediate = true
	end
	self.wndMain:Show(false, bImmediate)
end

function LootWindow:Show(bImmediate)
	if bImmediate == nil then
		bImmediate = true
	end
	self.wndMain:Show(true, bImmediate)
end

function LootWindow:Close()
	G.Utils:PlaySound(191)
	self.btnExpand:SetCheck(false)
	self.wndMain:Close()
end

function LootWindow:Open()
	G.Utils:PlaySound(190)
	self.wndMain:Invoke()
end

----------------------------------------------------------------------------------
-- Getters & Setters
----------------------------------------------------------------------------------
function LootWindow:IsManagementWindow() return false end
function LootWindow:GetWindow() return self.wndMain end
function LootWindow:IsShown() return self.wndMain:IsShown() end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function LootWindow:HandleOnOpenManagementWindow(wndHandler, wndControl, eMouseButton)
	local tManagementWindow = self.tRaidLooter:GetManagementWindow()
	if tManagementWindow ~= nil then
		if tManagementWindow:IsShown() then
			tManagementWindow:Close()
		else
			tManagementWindow:Open()
		end
	end
end

function LootWindow:HandleOnCloseOptions()
	G.Utils:PlaySound(191)
end

function LootWindow:HandleOnOpenOptions()
	self.btnExpand:SetCheck(false)
	G.Utils:PlaySound(190)
end

function LootWindow:HandleOnClearLatestRolles()
	if self.tRollSet ~= nil then
		self.tRollSet:Clear()
	end
end

function LootWindow:HandleOnRoll_ChatParser(strUnitName, nScore)
	if self.bIsRolling then
		if self.tRollSet:IsEmpty() then
			self.btnExpand:SetCheck(true)
		end

		self.tRollSet:AddRoller(strUnitName, nScore)
	end

	if strUnitName == G.Utils:GetPlayerName() then
		self:DisableInlineItem()

		if self.tItemDialog ~= nil then
			self.tItemDialog:Destroy()
			self.tItemDialog = nil
		end
	end
end

function LootWindow:HandleOnBeginRoll_ChatParser(strUnitName, tItem, bAscending)
	if self.bIsRolling then
		self:HandleOnEndRoll_ChatParser(strUnitName)
	end

	local tLootItem = self:FindLootItem(tItem)
	if tLootItem ~= nil then
		self.tRollSet:Clear()
		self.tRollSet:BeginRollingForItem(tLootItem, bAscending)
		self.bIsRolling = true

		if self.tSettings:ReadSetting("tGeneral.bUseRollPopups") then
			--self.tRaidLooter:ShowRollDialogForItem(tItem:GetItemId())
			self.tRaidLooter:ShowRollDialogForLootItem(tLootItem)
		else
			self:ShowInlineItem(tLootItem)
		end
	else
		G.Utils:Log("Unable to find loot item for id {" .. tItem:GetItemId() .. "}")
	end
end

function LootWindow:HandleOnEndRoll_ChatParser(strUnitName)
	self.tRollSet:EndRolling()
	self.bIsRolling = false

	self:ResetInlineItem()

	if self.tItemDialog ~= nil then
		self.tItemDialog:Destroy()
		self.tItemDialog = nil
	end
end

function LootWindow:HandleOnRollHint_ChatParser(strUnitName, tHint)
	self.tRollSet:UpdateRollHint(strUnitName, tHint)

	if strUnitName== G.Utils:GetPlayerName() then
		if tHint.bIsOffspec then
			self.eRollType = RollTypeEnum.OffSpec
			self.tInlineItem.btnOff:SetCheck(true)
		elseif tHint.bIsCostume then
			self.eRollType = RollTypeEnum.Costume
			self.tInlineItem.btnCostume:SetCheck(true)
		end
	end
end

function LootWindow:HandleOnMakeRoll(wndHandler, wndControl, eMouseButton)
	local strRollMax = (self.eRollType == RollTypeEnum.OffSpec and " 50"
		or self.eRollType == RollTypeEnum.Costume and " 33" or "")
	ChatSystemLib.Command(G.Locale:TranslateCommand("Roll")..strRollMax)

	self:DisableInlineItem()
	G.Utils:PlaySound(174)
end

function LootWindow:HandleOnMakePass(wndHandler, wndControl, eMouseButton)
	self:DisableInlineItem()
	G.Utils:PlaySound(200)
end

function LootWindow:HandleOnToggleLatestRolls(wndHandler, wndControl, eMouseButton)
	self.btnOptions:SetCheck(false)
end

function LootWindow:HandleOnLootWindowOpen(wndHandler, wndControl, eMouseButton)
	self:Open()
end

function LootWindow:HandleOnLootWindowClose(wndHandler, wndControl, eMouseButton)
	self:Close()
end

function LootWindow:HandleOnOptionUnselected(wndHandler, wndControl, eMouseButton)
	self.tRaidLooter:OptionSelectionHelper(wndHandler:GetName(), false)
end

function LootWindow:HandleOnOptionSelected(wndHandler, wndControl, eMouseButton)
	self.tRaidLooter:OptionSelectionHelper(wndHandler:GetName(), true)
end

function LootWindow:HandleOnInlinePopupClose(wndHandler, wndControl, eMouseButton)
	self:DisableInlineItem()
end

function LootWindow:HandleOnInlineRollTypeChanged(wndHandler, wndControl, eMouseButton)
	if wndControl:IsChecked() then
		local strButtonName = wndControl:GetName()

		if strButtonName == "Inline_MainSpecButton" then
			self.eRollType = RollTypeEnum.MainSpec
		elseif strButtonName == "Inline_OffSpecButton" then
			self.eRollType = RollTypeEnum.OffSpec
		elseif strButtonName == "Inline_CostumeButton" then
			self.eRollType = RollTypeEnum.Costume
		else
			self.eRollType = RollTypeEnum.MainSpec
		end
	else
		self.eRollType = RollTypeEnum.MainSpec
	end
end

----------------------------------------------------------------------------------
-- Private functions
----------------------------------------------------------------------------------
function LootWindow:ShowInlineItem(tLootItem)
	local strTextColor = G.Utils:GetItemQualityString(tLootItem)

	self.tInlineItem.wnd:Show(true)
	self.tInlineItem.wndName:SetText(tLootItem:GetName())
	self.tInlineItem.wndName:SetTextColor(strTextColor)
	self.tInlineItem.wndIcon:SetSprite(tLootItem:GetIcon())
	self.tInlineItem.btnMain:SetCheck(true)
	self.tInlineItem.btnOff:SetCheck(false)
	self.tInlineItem.btnCostume:SetCheck(false)

	local tItem         = tLootItem:GetItem()
	local tItemEquipped = tItem:GetEquippedItemForItemType()

	Tooltip.GetItemTooltipForm(self, self.tInlineItem.wndItem, tItem, {
		bPrimary = true,
		bSelling = false,
		itemCompare = tItemEquipped
	})

	local bIsNotInline = self.tSettings:ReadSetting("tGeneral.bUseRollPopups")
	bIsNotInline = bIsNotInline or self.tAutoRoll:GetAutoRoll(tLootItem:GetIItem()) ~= nil

	if bIsNotInline then
		self:DisableInlineItem()
	else
		G.Utils:PlayNewRollSound()
	end
end

function LootWindow:ResetInlineItem()
	self.tInlineItem.btnRoll:Enable(true)
	self.tInlineItem.btnPass:Enable(true)
	self.tInlineItem.btnMain:Enable(true)
	self.tInlineItem.btnOff:Enable(true)
	self.tInlineItem.btnCostume:Enable(true)
	self.tInlineItem.wnd:Show(false, false)
	self.eRollType = RollTypeEnum.MainSpec
end

function LootWindow:DisableInlineItem()
	self.tInlineItem.btnRoll:Enable(false)
	self.tInlineItem.btnPass:Enable(false)
	self.tInlineItem.btnMain:Enable(false)
	self.tInlineItem.btnOff:Enable(false)
	self.tInlineItem.btnCostume:Enable(false)
	self.tInlineItem.wnd:Show(false, false)
	self.eRollType = RollTypeEnum.MainSpec
end

function LootWindow:GetLootItem(nLootId) return self.tLootCollection:GetItemWithId(nLootId) end
function LootWindow:GetLootItems() return self.tLootCollection end
function LootWindow:HandleOnCallRanksWindow() ChatSystemLib.Command("/ranks") end

-----------------------------------------------------------------------------------------------
-- Events
-----------------------------------------------------------------------------------------------
function LootWindow:OnOpenManagementWindow(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnOpenManagementWindow", ...) end
function LootWindow:OnCallRanksWindow(...)
	G.Utils:InvokeEventHandler(self, "ManagementWindow", "OnCallRanksWindow", ...) end
function LootWindow:OnCloseOptions(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnCloseOptions", ...) end
function LootWindow:OnOpenOptions(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnOpenOptions", ...) end
function LootWindow:OnClearLatestRolles(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnClearLatestRolles", ...) end
function LootWindow:OnRoll_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnRoll_ChatParser", ...) end
function LootWindow:OnBeginRoll_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnBeginRoll_ChatParser", ...) end
function LootWindow:OnEndRoll_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnEndRoll_ChatParser", ...) end
function LootWindow:OnRollHint_ChatParser(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnRollHint_ChatParser", ...) end
function LootWindow:OnMakeRoll(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnMakeRoll", ...) end
function LootWindow:OnMakePass(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnMakePass", ...) end
function LootWindow:OnClearRolles_Loot(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnClearLatestRolles", ...) end
function LootWindow:OnToggleLatestRolls(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnToggleLatestRolls", ...) end
function LootWindow:OnLootWindowOpen(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnLootWindowOpen", ...) end
function LootWindow:OnLootWindowClose(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnLootWindowClose", ...) end
function LootWindow:OnOptionUnselected(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnOptionUnselected", ...) end
function LootWindow:OnOptionSelected(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnOptionSelected", ...) end
function LootWindow:OnInlinePopupClose(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnInlinePopupClose", ...) end
function LootWindow:OnInlineRollTypeChanged(...)
	G.Utils:InvokeEventHandler(self, "LootWindow", "OnInlineRollTypeChanged", ...) end
