--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local ChatSystemLib = require "ChatSystemLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- ManagementWindow
-- Represents the management window. This is the window that the leader/
-- assistents see.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local ManagementWindow = {}
G.ManagementWindow = ManagementWindow
function ManagementWindow:New(tRaidLooter)
	local o = {
		strClassName = "ManagementWindow",
		wndMain = nil,
		wndOptions = nil,
		btnOptions = nil,
		btnAutoLootTab = nil,
		btnDistributionTab = nil,
		tDistributionView = nil,
		tAutolootView = nil,
		tRaidLooter = tRaidLooter,
		tSettings = tRaidLooter:GetSettings(),
		tMasterLoot = tRaidLooter:GetMasterLoot(),
		tChatParser = tRaidLooter:GetChatParser(),
		tLinker = G.Linker:New()
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function ManagementWindow:Init(tTarget)
	self.tLinker:Link(NLLC, self)
	self.wndMain = self.tRaidLooter:CreateTopLevelWindow("ManagementWindow", self)

	-- Setup each tab
	local wndDistributionPage = self.wndMain:FindChild("DistributionPage")
	local btnDistributionTab = self.wndMain:FindChild("DistributionPageButton")
	self.tDistributionView = G.DistributionView:New(self.tRaidLooter, wndDistributionPage)
	self.tDistributionView:Init(self)
	btnDistributionTab:AttachWindow(wndDistributionPage)
	btnDistributionTab:SetCheck(true)
	btnDistributionTab:Enable(true)
	self.btnDistributionTab = btnDistributionTab

	local wndAutolootPage = self.wndMain:FindChild("AutolootPage")
	local btnAutoLootTab = self.wndMain:FindChild("AutolootPageButton")
	self.tAutolootView = G.AutolootView:New(self.tRaidLooter, wndAutolootPage)
	self.tAutolootView:Init(self)
	btnAutoLootTab:AttachWindow(wndAutolootPage)
	btnAutoLootTab:SetCheck(false)
	btnAutoLootTab:Enable(true)
	self.btnAutoLootTab = btnAutoLootTab

	self.wndOptions = self.wndMain:FindChild("OptionsDialog")
	self.btnOptions = self.wndMain:FindChild("ConfigureButton")
	self.btnOptions:AttachWindow(self.wndOptions)
	self.btnOptions:SetCheck(false)

	self.tRaidLooter:RestoreOptions()
	G.Utils:MoveWindow(self.wndMain, self.tSettings:ReadBlock("tWindowLocations.wndManageWindowLoc"))
end

function ManagementWindow:Destroy()
	self.tLinker:Destroy()
	self.wndMain:Destroy()
	self.tDistributionView:Destroy()
	self.tAutolootView:Destroy()

	self.wndMain           = nil
	self.tRaidLooter       = nil
	self.tMasterLoot       = nil
	self.tChatParser       = nil
	self.tDistributionView = nil
	self.tAutolootView     = nil
	self.tSettings         = nil
end

----------------------------------------------------------------------------------
-- Save & Restore
----------------------------------------------------------------------------------
function ManagementWindow:OnSave(eLevel)
	if self.wndMain ~= nil then
		local tLocations = self.tSettings:ReadBlock("tWindowLocations")
		tLocations.wndManageWindowLoc = self.wndMain:GetLocation():ToTable()
	end
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function ManagementWindow:UnloadTabs()
	if self.tDistributionView ~= nil then
		self.tDistributionView:Destroy()
	end

	self.tDistributionView = nil
end

function ManagementWindow:Hide(bImmediate)
	if bImmediate == nil then
		bImmediate = true
	end
	self.wndMain:Show(false, bImmediate)
end

function ManagementWindow:Show(bImmediate)
	if bImmediate == nil then
		bImmediate = true
	end
	self.wndMain:Show(true, bImmediate)
end

function ManagementWindow:Close()
	G.Utils:PlaySound(191)
    self.wndMain:Close()
end

function ManagementWindow:CloseOptions() G.Utils:PlaySound(191) end
function ManagementWindow:OpenOptions() G.Utils:PlaySound(190) end
function ManagementWindow:Open() self.wndMain:Invoke() end

----------------------------------------------------------------------------------
-- Getters & Setters
----------------------------------------------------------------------------------
function ManagementWindow:GetWindow() return self.wndMain end
function ManagementWindow:IsShown() return self.wndMain:IsShown() end
function ManagementWindow:GetLootItems() return self.tDistributionView:GetLootItems() end
function ManagementWindow:GetLootItem(nLootId) return self.tDistributionView:GetLootItem(nLootId) end
function ManagementWindow:IsManagementWindow() return true end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function ManagementWindow:HandleOnOptionUnselected(wndHandler, wndControl, eMouseButton)
	self.tRaidLooter:OptionSelectionHelper(wndHandler:GetName(), false)
end

function ManagementWindow:HandleOnOptionSelected(wndHandler, wndControl, eMouseButton)
	self.tRaidLooter:OptionSelectionHelper(wndHandler:GetName(), true)
end

function ManagementWindow:HandleOnPrepareAddRuleForItem(tLootItem)
	self.btnAutoLootTab:SetCheck(true)
	self.btnDistributionTab:SetCheck(false)
	local tRaider = self.tDistributionView:GetSelectedRecipient()
	local strNameSuggestion

	if tRaider ~= nil then
		strNameSuggestion = tRaider:GetFullname()
	end

	G.Utils:FireEvent(self.tAutolootView, "OnPrepareAddRuleForItem", tLootItem, strNameSuggestion)
end

function ManagementWindow:HandleOnTabChecked() G.Utils:PlaySound(184) end
function ManagementWindow:HandleOnManagementWindowClose() self:Close() end
function ManagementWindow:HandleOnCallRanksWindow() ChatSystemLib.Command("/ranks") end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function ManagementWindow:OnTabChecked(...)
	G.Utils:InvokeEventHandler(self, "ManagementWindow", "OnTabChecked", ...) end
function ManagementWindow:OnManagementWindowClose(...)
	G.Utils:InvokeEventHandler(self, "ManagementWindow", "OnManagementWindowClose", ...) end
function ManagementWindow:OnCallRanksWindow(...)
	G.Utils:InvokeEventHandler(self, "ManagementWindow", "OnCallRanksWindow", ...) end
function ManagementWindow:OnOptionUnselected(...)
	G.Utils:InvokeEventHandler(self, "ManagementWindow", "OnOptionUnselected", ...) end
function ManagementWindow:OnOptionSelected(...)
	G.Utils:InvokeEventHandler(self, "ManagementWindow", "OnOptionSelected", ...) end
function ManagementWindow:OnPrepareAddRuleForItem(...)
	G.Utils:InvokeEventHandler(self, "ManagementWindow", "OnPrepareAddRuleForItem", ...) end
function ManagementWindow:OnAssignLoot(...)
	G.Utils:FireEvent(self.tDistributionView, "OnAssignLoot", ...) end
function ManagementWindow:OnAssignMinorLoot(...)
	G.Utils:FireEvent(self.tDistributionView, "OnAssignMinorLoot", ...) end
function ManagementWindow:OnAssignMajorLoot(...)
	G.Utils:FireEvent(self.tDistributionView, "OnAssignMajorLoot", ...) end
function ManagementWindow:OnSwitchSpec(...)
	G.Utils:FireEvent(self.tDistributionView, "OnSwitchSpec", ...) end
function ManagementWindow:OnBeginRoll(...)
	G.Utils:FireEvent(self.tDistributionView, "OnBeginRoll", ...) end
function ManagementWindow:OnEndRoll(...)
	G.Utils:FireEvent(self.tDistributionView, "OnEndRoll", ...) end
function ManagementWindow:OnRandomizeLoot(...)
	G.Utils:FireEvent(self.tDistributionView, "OnRandomizeLoot", ...) end
function ManagementWindow:OnReroll(...)
	G.Utils:FireEvent(self.tDistributionView, "OnReroll", ...) end
function ManagementWindow:OnCurrentRecipientDeselect(...)
	G.Utils:FireEvent(self.tDistributionView, "OnCurrentRecipientDeselect", ...) end
function ManagementWindow:OnCurrentItemDeselect(...)
	G.Utils:FireEvent(self.tDistributionView, "OnCurrentItemDeselect", ...)  end
function ManagementWindow:OnClearLatestRolles(...)
	G.Utils:FireEvent(self.tDistributionView, "OnClearLatestRolles", ...) end
function ManagementWindow:OnRoll_ChatParser(...)
	G.Utils:FireEvent(self.tDistributionView, "OnRoll_ChatParser", ...) end
function ManagementWindow:OnBeginRoll_ChatParser(...)
	G.Utils:FireEvent(self.tDistributionView, "OnBeginRoll_ChatParser", ...) end
function ManagementWindow:OnEndRoll_ChatParser(...)
	G.Utils:FireEvent(self.tDistributionView, "OnEndRoll_ChatParser", ...) end
function ManagementWindow:OnRollHint_ChatParser(...)
	G.Utils:FireEvent(self.tDistributionView, "OnRollHint_ChatParser", ...) end
function ManagementWindow:OnParsingStarted_ChatParser(...)
	G.Utils:FireEvent(self.tDistributionView, "OnParsingStarted_ChatParser", ...) end
function ManagementWindow:OnParsingStoped_ChatParser(...)
	G.Utils:FireEvent(self.tDistributionView, "OnParsingStoped_ChatParser", ...) end
function ManagementWindow:OnSelectAllItems(...)
	G.Utils:FireEvent(self.tDistributionView, "OnSelectAllItems", ...) end
function ManagementWindow:OnUnselectAllItems(...)
	G.Utils:FireEvent(self.tDistributionView, "OnUnselectAllItems", ...) end
function ManagementWindow:OnSortChange(...)
	G.Utils:FireEvent(self.tDistributionView, "OnSortChange", ...) end
function ManagementWindow:OnFilterChange(...)
	G.Utils:FireEvent(self.tDistributionView, "OnFilterChange", ...) end
function ManagementWindow:OnFilterEquippable(...)
	G.Utils:FireEvent(self.tDistributionView, "OnFilterEquippable", ...) end
function ManagementWindow:OnFilterCommodity(...)
	G.Utils:FireEvent(self.tDistributionView, "OnFilterCommodity", ...) end
function ManagementWindow:OnFilterRecipients(...)
	G.Utils:FireEvent(self.tDistributionView, "OnFilterRecipients", ...) end
function ManagementWindow:OnAddNewAutolootRule(...)
	G.Utils:FireEvent(self.tAutolootView, "OnAddNewAutolootRule", ...) end
function ManagementWindow:OnDisableAutolooting(...)
	G.Utils:FireEvent(self.tAutolootView, "OnDisableAutolooting", ...) end
function ManagementWindow:OnEnableAutolooting(...)
	G.Utils:FireEvent(self.tAutolootView, "OnEnableAutolooting", ...) end
function ManagementWindow:OnDeleteAllAutolootRule(...)
	G.Utils:FireEvent(self.tAutolootView, "OnDeleteAllAutolootRule", ...) end
function ManagementWindow:OnToggleProfileDropdown(...)
	G.Utils:FireEvent(self.tAutolootView, "OnToggleProfileDropdown", ...) end
function ManagementWindow:OnToggleRuleDropdown(...)
	G.Utils:FireEvent(self.tAutolootView, "OnToggleRuleDropdown", ...) end
function ManagementWindow:OnGridSelectionChanged(...)
	G.Utils:FireEvent(self.tAutolootView, "OnGridSelectionChanged", ...) end
function ManagementWindow:OnAutolootRowDeleted(...)
	G.Utils:FireEvent(self.tAutolootView, "OnAutolootRowDeleted", ...) end
