--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local GameLib = require "GameLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- MasterLoot
-- Represents the master loot
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local MasterLoot = {}
G.MasterLoot = MasterLoot
function MasterLoot:New(tRaidLooter, tLootQueue)
	local o = {
		tRaidLooter = tRaidLooter,
		tSettings  = tRaidLooter:GetSettings(),
		tLootQueue = tLootQueue,
		tLoot      = {},
		tCallbacks = {},
		tProfile   = nil,
		nLootCount = 0,
		nPreviousLootCount = -1,
		tLinker = G.Linker:New()
	}

	setmetatable(o, self)
	self.__index = self
	o.tLootQueue:SetMasterLoot(o)
	return o
end

function MasterLoot:Init(tTarget)
	self.tLinker:Link(NLLC, self)
	Apollo.RegisterEventHandler("MasterLootUpdate", "OnMasterLootUpdate", self)
	Apollo.RegisterEventHandler("LootAssigned", "OnLootAssigned", self)
end

function MasterLoot:Destroy()
	self.tLinker:Destroy()
	Apollo.RemoveEventHandler("MasterLootUpdate", self)
	Apollo.RemoveEventHandler("LootAssigned", self)

	self.tCallbacks  = nil
	self.tRaidLooter = nil
	self.tSettings   = nil
	self.tLinker     = nil
	self.tLootQueue  = nil
	self.tLoot       = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function MasterLoot:Update(bPreventEvent)
	if not G.Debug.bIsLootSourceInventory then
		self.tLoot = GameLib.GetMasterLoot()
	end

	local nLootCount = #self.tLoot
	self.nLootCount = nLootCount
	if not bPreventEvent then
		self:Notify("OnLootUpdate", self.tLoot)
	end
	self.nPreviousLootCount = nLootCount

	if self.tSettings:ReadSetting("tLeader.bAutolootEnabled") then
		self:UpdateAutoloot(self.tLoot)
	end
end

function MasterLoot:IsEmpty() return self.nLootCount == 0 end
function MasterLoot:RecentlyHadLoot() return self.nPreviousLootCount > 0 end
function MasterLoot:RecentlyHadLootDropped() return self.nLootCount > self.nPreviousLootCount end
function MasterLoot:GetNumberOfItems()	return self.nLootCount end

function MasterLoot:Assign(tLootItems, tUnitInfo)
	local bShouldEnqueue = tUnitInfo.tUnit == nil or not tUnitInfo.tUnit:IsValid()

	for _, tLootItem in ipairs(tLootItems) do
		local nLootId = tLootItem:GetId()
		if bShouldEnqueue then
			self.tLootQueue:Push(tLootItem, tUnitInfo.strUnitName)
		else
			if G.Debug.bIsTest then
				self:HandOutDebugLoot(nLootId, tUnitInfo.tUnit)
			else
				GameLib.AssignMasterLoot(nLootId, tUnitInfo.tUnit)
			end
		end
	end
end

function MasterLoot:Randomize(tLootItems)
	for _, tLootItem in ipairs(tLootItems) do
		local tEligibleRecipients = tLootItem:GetEligibleRecipients()
		local nIndex = G.Utils:Random(#tEligibleRecipients)
		local tUnitInfo = tEligibleRecipients[nIndex]
		local tTmpList = {}
		table.insert(tTmpList, tLootItem)
		self:Assign(tTmpList, tUnitInfo)
	end
end

function MasterLoot:AddLoot(tLoot)
	for _, tLootDrop in ipairs(tLoot) do
		table.insert(self.tLoot, tLootDrop)
	end

	self:Update()
end

-- Should Respond to:
--     OnLootUpdate
function MasterLoot:RegisterCallback(tCallback)
	table.insert(self.tCallbacks, tCallback)
end

function MasterLoot:UnregisterCallback(tCallback)
	for nIndex, tStoredCallback in ipairs(self.tCallbacks) do
		if tStoredCallback == tCallback then
			table.remove(self.tCallbacks, nIndex)
		end
	end
end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function MasterLoot:HandleOnLootAssigned(tLootInfo)
	local strItem = tLootInfo.itemLoot:GetChatLinkString()
	local nCount = tLootInfo.itemLoot:GetStackCount()
	if nCount > 1 then
		strItem = String_GetWeaselString(Apollo.GetString("CombatLog_MultiItem"), nCount, strItem)
	end

	Event_FireGenericEvent("GenericEvent_LootChannelMessage",
		String_GetWeaselString(Apollo.GetString("CRB_MasterLoot_AssignMsg"), strItem, tLootInfo.strPlayer))
end

function MasterLoot:HandleOnMasterLootUpdate()
	self:Update()
end

function MasterLoot:HandleOnAutolootChanged()
	self:Update()
end

----------------------------------------------------------------------------------
-- Private functions
----------------------------------------------------------------------------------
function MasterLoot:HandOutDebugLoot(nLootId, tUnit)
	for nIndex, tDrop in ipairs(self.tLoot) do
		if tDrop.nLootId == nLootId then
			table.remove(self.tLoot, nIndex)
			self.nLootCount = self.nLootCount - 1

			self:OnLootAssigned({
				itemLoot = tDrop.itemDrop,
				strPlayer = tUnit:GetName()
			})
			self:Notify("OnLootUpdate", self.tLoot)
			return
		end
	end

	G.Utils:Log("Unable to find item to assign")
end

function MasterLoot:FindDropWithId(nLootId)
	for _, tDrop in ipairs(self.tLoot) do
		if tDrop.nLootId == nLootId then
			return tDrop
		end
	end
end

function MasterLoot:FindFirstDropWithItemId(nItemId)
	for _, tDrop in ipairs(self.tLoot) do
		if tDrop.itemDrop:GetItemId() == nItemId then
			return tDrop
		end
	end
end

function MasterLoot:Notify(strEventName, ...)
	G.Utils:Log("Notify => " .. strEventName)
	for _, tCallback in ipairs(self.tCallbacks) do
		G.Utils:FireEvent(tCallback, strEventName, ...)
	end
end

function MasterLoot:UpdateAutoloot(tLoot)
	self:UpdateProfileBinding()
	if self.tProfile ~= nil and self.tProfile.tLootList ~= nil then
		local tRules = self.tProfile.tLootList

		for nIndex, tDropInfo in ipairs(tLoot) do
			local strExistingItemName = tDropInfo.itemDrop:GetName()
			for strItemName, tRule in pairs(tRules) do
				if strExistingItemName == strItemName then
					if self:TryAutoloot(tDropInfo, tRule.strRecipient) then
						break
					end
				end
			end
		end
	end
end

function MasterLoot:UpdateProfileBinding()
	self.tProfile = self.tSettings:ReadBlock("tAutoloot.tProfileRef") -- Ref from storage
end

function MasterLoot:TryAutoloot(tDrop, strRecipient)
	local tLootItem = self.tRaidLooter:GetLootItem(tDrop.nLootId)

	if not tLootItem or tLootItem:IsQueued() then
		return false
	end

	local tLootItems = {}
	table.insert(tLootItems, tLootItem)

	if strRecipient == "Leader" then
		local tPlayerUnit = GameLib.GetPlayerUnit()
		self:Assign(tLootItems, {
			tUnit = tPlayerUnit,
			strUnitName = tPlayerUnit:GetName()
		})
		return true
	else
		if strRecipient == "Randomize" then
			self:Randomize(tLootItems)
			return true
		else
			local bIsEligible, tUnit = self:IsEligible(strRecipient, tDrop)
			if bIsEligible then
				self:Assign(tLootItems, {
					tUnit = tUnit,
					strUnitName = strRecipient
				})
				return true
			end
		end
	end

	return false
end

function MasterLoot:IsEligible(strRecipient, tDrop)
	local tEligibleRecipients = G.Utils:MergeLootersForDrop(tDrop)
	for _, tEligibleRecipient in ipairs(tEligibleRecipients) do
		if tEligibleRecipient.strUnitName == strRecipient then
			return true, tEligibleRecipient.tUnit
		end
	end

	return false
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function MasterLoot:OnLootAssigned(...) G.Utils:InvokeEventHandler(self, "MasterLoot", "OnLootAssigned", ...) end
function MasterLoot:OnMasterLootUpdate(...)
	G.Utils:InvokeEventHandler(self, "MasterLoot", "OnMasterLootUpdate", ...) end
function MasterLoot:OnAutolootChanged(...) G.Utils:InvokeEventHandler(self, "MasterLoot", "OnAutolootChanged", ...) end
