--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
require "Window"
local GameLib = require "GameLib"
local Apollo = require "Apollo"
local Item = require "Item"
local GroupLib = require "GroupLib"
local ChatSystemLib = require "ChatSystemLib"
local ApolloTimer = require "ApolloTimer"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

-- Version information
local MAJOR = 2
local MINOR = 0
local REV   = 0
G.CURRENT_VERSION = MAJOR.."."..MINOR.."."..REV

-- Constatns
G.ItemInfo = {
	[Item.CodeEnumItemQuality.Inferior]  = {
		strName = "ItemQuality_Inferior",
		strRarityBorder = "BK3:UI_RarityBorder_Grey",
		strExtendedRarityBorder = "CRB_Tooltips:sprTooltip_Header_Silver"},
	[Item.CodeEnumItemQuality.Average]   = {
		strName = "ItemQuality_Average",
		strRarityBorder = "BK3:UI_RarityBorder_White",
		strExtendedRarityBorder = "CRB_Tooltips:sprTooltip_Header_White"},
	[Item.CodeEnumItemQuality.Good]      = {
		strName = "ItemQuality_Good",
		strRarityBorder = "BK3:UI_RarityBorder_Green",
		strExtendedRarityBorder = "CRB_Tooltips:sprTooltip_Header_Green"},
	[Item.CodeEnumItemQuality.Excellent] = {
		strName = "ItemQuality_Excellent",
		strRarityBorder = "BK3:UI_RarityBorder_Blue",
		strExtendedRarityBorder = "CRB_Tooltips:sprTooltip_Header_Blue"},
	[Item.CodeEnumItemQuality.Superb]    = {
		strName = "ItemQuality_Superb",
		strRarityBorder = "BK3:UI_RarityBorder_Purple",
		strExtendedRarityBorder = "CRB_Tooltips:sprTooltip_Header_Purple"},
	[Item.CodeEnumItemQuality.Legendary] = {
		strName = "ItemQuality_Legendary",
		strRarityBorder = "BK3:UI_RarityBorder_Orange",
		strExtendedRarityBorder = "CRB_Tooltips:sprTooltip_Header_Orange"},
	[Item.CodeEnumItemQuality.Artifact]  = {
		strName = "ItemQuality_Artifact",
		strRarityBorder = "BK3:UI_RarityBorder_Magenta",
		strExtendedRarityBorder = "CRB_Tooltips:sprTooltip_Header_Pink"}
}
G.ClassInfo = {
	[GameLib.CodeEnumClass.Esper] 		 = { strIcon = "Icon_Windows_UI_CRB_Esper",        strArmorType = "L" },
	[GameLib.CodeEnumClass.Medic] 		 = { strIcon = "Icon_Windows_UI_CRB_Medic",        strArmorType = "M" },
	[GameLib.CodeEnumClass.Stalker] 	 = { strIcon = "Icon_Windows_UI_CRB_Stalker",      strArmorType = "M" },
	[GameLib.CodeEnumClass.Warrior]      = { strIcon = "Icon_Windows_UI_CRB_Warrior",      strArmorType = "H" },
	[GameLib.CodeEnumClass.Engineer] 	 = { strIcon = "Icon_Windows_UI_CRB_Engineer",     strArmorType = "H" },
	[GameLib.CodeEnumClass.Spellslinger] = { strIcon = "Icon_Windows_UI_CRB_Spellslinger", strArmorType = "L" }
}



--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- RaidLooter
-- The main addon class.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local RaidLooter = {}
G.RaidLooter = RaidLooter
function RaidLooter:New()
	local o = {
		strClassName      = "RaidLooter",
		tLinker           = G.Linker:New(),
		wndOverlay        = nil,
		wndAmount         = nil,
		tActiveWindow     = nil,
		tMasterLoot       = nil,
		tLootQueue         = nil,
		tXMLs               = nil,
		tChatParser          = nil,
		tSettings             = nil,
		tXMLFactory            = nil,
		tLootWindowFactory      = nil,
		tManagementWindowFactory = nil,
		tContextMenuFactory      = nil,
		tLootDetailsWindowFactory = nil,
		tRollDialogWindowFactory = nil,
		tItemDetailsView        = nil,
		tContextMenu           = nil,
		tItemDialog           = nil,
		bIsLoaded            = false
	}

  setmetatable(o, self)
  self.__index = self
	NLLC:OnceSaving("OnSave", o)
	NLLC:OnceLoaded("OnLoad", o)
	NLLC:OnceRestored("OnRestore", o)
	return o
end

function RaidLooter:Startup(tDependencies)
	-- Create objects and wire them.
	self.tLootQueue  = tDependencies.LootQueue:New()
	self.tSettings   = tDependencies.Settings:New()
	self.tMasterLoot = tDependencies.MasterLoot:New(self, self.tLootQueue)
	self.tChatParser = tDependencies.ChatParser:New()
	self.tXMLFactory = tDependencies.XMLFactory:New()
	self.tAutoRoll   = tDependencies.AutoRoll:New(self)

	-- Reference to deps to break cycles.
	self.tManagementWindowFactory = tDependencies.ManagementWindow
	self.tLootWindowFactory = tDependencies.LootWindow
	self.tContextMenuFactory = tDependencies.ContextMenu
	self.tLootDetailsWindowFactory = tDependencies.LootDetailsWindow
	self.tRollDialogWindowFactory = tDependencies.ItemRollDialog

	self.tLinker:Link(NLLC, self)
	-- Apollo.RegisterAddon(self) --this is no more a standalone addon!
end

function RaidLooter:OnLoad()
	-- LoadAddon should be called from here or the OnRestore not
	-- both. So we use a delayed load check using a timer as we prefer restoring
	-- from existing settings.
	--
	-- This delayed loading will be triggered when there are no settings.
	self.tDelayedLoadTimer = ApolloTimer.Create(3, true, "OnDelayedLoad", self)
end

function RaidLooter:OnSave(eLevel, tSavedData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
      return nil
  end

	G.Utils:FireEvent(self.tActiveWindow, "OnSave", eLevel)
	G.Utils:FireEvent(G.Debug, "OnSave", eLevel)

	if self.wndOverlay ~= nil then
		local tLocations = self.tSettings:ReadBlock("tWindowLocations")
		tLocations.wndOverlayLoc = self.wndOverlay:GetLocation():ToTable()
	end

	tSavedData.RaidLooter = self.tSettings:Save(eLevel)
end

function RaidLooter:OnDelayedLoad()
	self.tDelayedLoadTimer:Stop()
	self.tDelayedLoadTimer = nil
	if not self.bIsLoaded then
		self.tSettings:Reset()
		self:LoadAddon()
	end
end

function RaidLooter:OnRestore(eLevel, tSavedData)
	local tData = tSavedData and tSavedData.RaidLooter
	if not tData then return end
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return nil
	end
	if not self.bIsLoaded then
		self.tSettings:Load(tData)
		self:LoadAddon()
	end
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function RaidLooter:ActivateAddon()
	if self:ShouldBeActive() and not self:IsActive() then
		self.tXMLFactory:LoadAll("OnActivateAddon", self)
	end
end

function RaidLooter:DeactivateAddon()
	if not self:ShouldBeActive() and self:IsActive() then
		self:OnDeactivateAddon()
	end
end

function RaidLooter:LoadAddon()
	self.bIsLoaded = true
	Apollo.RegisterSlashCommand("rl",                                  "OnRaidLooterCmd", self)
	Apollo.RegisterEventHandler("Group_Join",                          "OnJoinedGroup",  self)
	Apollo.RegisterEventHandler("Group_LootRulesChanged",              "OnGroupLootRulesChanged", self)
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded",          "OnInterfaceMenuListHasLoaded", self)
	Apollo.RegisterEventHandler("RaidLooterInterfaceList",             "OnRaidLooterCmd", self)
	Apollo.RegisterEventHandler("GenericEvent_ToggleGroupBag",         "OnToggleGroupBag", self)
	Apollo.RegisterEventHandler("GenericEvent_Raid_UncheckMasterLoot", "OnMasterLootUnchecked", self)

	-- Init singletons.
	G.Locale:Init(self)
	G.Utils:Init(self)
	G.Debug:Init(self)
	G.Debug:SetSettings(self.tSettings)
	G.Debug:OnRestore()
	G.Debug:OnLoad()

	self.tLootQueue:Init(self)
	self.tMasterLoot:Init(self)
	self.tXMLFactory:Init(self)
	self.tChatParser:Init(self)
	self.tSettings:Init(self)
	self.tAutoRoll:Init(self)

	self.tMasterLoot:RegisterCallback(self)
	self.tChatParser:RegisterCallback(self.tMasterLoot)
	self:ActivateAddon()
end

function RaidLooter:ShowRollDialogForItem(nItemId)
	if self.tItemDialog ~= nil then
		self.tItemDialog:Destroy()
	end

	local tDrop = self.tMasterLoot:FindFirstDropWithItemId(nItemId)
	if tDrop ~= nil then
		self.tItemDialog = self.tRollDialogWindowFactory:New(self, tDrop)
		self.tItemDialog:Init(self)
		local tSize = self.tItemDialog:GetSize()
		local nWidthOverTwo  = 0.5 * tSize.nWidth
		local nHeightOverTwo = 0.5 * tSize.nHeight

		self.tItemDialog:SetAnchor({
			tPoints  = {0.5, 0.30, 0.5, 0.30},
			tOffsets = {-nWidthOverTwo, -nHeightOverTwo, nWidthOverTwo, nHeightOverTwo }
		})
	end
end

function RaidLooter:ShowRollDialogForLootItem(tLootItem)
	if self.tItemDialog ~= nil then
		self.tItemDialog:Destroy()
	end

	if tLootItem:GetAutoRoll() ~= nil then return end --dont show the dialog for auto-rolled items.

	local tDrop = tLootItem.tDrop

	self.tItemDialog = self.tRollDialogWindowFactory:New(self, tDrop)
	self.tItemDialog:Init(self)
	local tSize = self.tItemDialog:GetSize()
	local nWidthOverTwo  = 0.5 * tSize.nWidth
	local nHeightOverTwo = 0.5 * tSize.nHeight

	self.tItemDialog:SetAnchor({
		tPoints  = {0.5, 0.30, 0.5, 0.30},
		tOffsets = {-nWidthOverTwo, -nHeightOverTwo, nWidthOverTwo, nHeightOverTwo }
	})
end

function RaidLooter:ShowContextMenu(tLootItem)
	if self.tContextMenu ~= nil then
		self.tContextMenu:Close()
		self.tContextMenu:Destroy()
		self.tContextMenu = nil
	end

	if self.tItemDetailsView ~= nil then
		self.tItemDetailsView:Close()
		self.tItemDetailsView:Destroy()
		self.tItemDetailsView = nil
	end

	self.tContextMenu = self.tContextMenuFactory:New(self, tLootItem)
	self.tContextMenu:Init(self)
	self.tContextMenu:Open()
end

function RaidLooter:ShowItemDetails(tLootItem)
	local tPosition
	if self.tContextMenu ~= nil then
		tPosition = self.tContextMenu:GetPosition()
		self.tContextMenu:Close()
		self.tContextMenu:Destroy()
		self.tContextMenu = nil
	end

	if self.tItemDetailsView ~= nil then
		self.tItemDetailsView:Close()
		self.tItemDetailsView:Destroy()
		self.tItemDetailsView = nil
	end

	self.tItemDetailsView = self.tLootDetailsWindowFactory:New(self, tLootItem)
	self.tItemDetailsView:Init(self)
	self.tItemDetailsView:Open(tPosition)
end

function RaidLooter:ToggleActiveWindow()
	local tActiveWindow = self:GetActiveWindow()
	if tActiveWindow ~= nil then
		if tActiveWindow:IsShown() then
			tActiveWindow:Close()
		else
			tActiveWindow:Open()
			self.tMasterLoot:Update()
		end
	end
end

function RaidLooter:RestoreOptions()
	-- Read from settings.
	local wndMain = self:GetActiveWindow():GetWindow()
	wndMain:FindChild("opt-btn_tGeneral_bUseRollPopups"):SetCheck(self.tSettings:Read("tGeneral.bUseRollPopups"))
	wndMain:FindChild("opt-btn_tGeneral_bOpenOnLoot"):SetCheck(self.tSettings:Read("tGeneral.bOpenOnLoot"))
	wndMain:FindChild("opt-btn_tGeneral_bUseOverlay"):SetCheck(self.tSettings:Read("tGeneral.bUseOverlay"))
	wndMain:FindChild("opt-btn_tGeneral_bUseSound"):SetCheck(self.tSettings:Read("tGeneral.bUseSound"))
	wndMain:FindChild("opt-btn_tGeneral_bUseRollSound"):SetCheck(self.tSettings:Read("tGeneral.bUseRollSound"))
	wndMain:FindChild("opt-btn_tGeneral_bUseLootSound"):SetCheck(self.tSettings:Read("tGeneral.bUseLootSound"))

	local btnLowstRollWins = wndMain:FindChild("opt-btn_tLeader_bAscending")
	local btnOpenOnLoot = wndMain:FindChild("opt-btn_tGeneral_bOpenOnLoot")
	local btnUseSound = wndMain:FindChild("opt-btn_tGeneral_bUseSound")
	local btnOverlay = wndMain:FindChild("opt-btn_tGeneral_bUseOverlay")
	local btnRoll = wndMain:FindChild("opt-btn_tGeneral_bUseRollSound")
	local btnLoot = wndMain:FindChild("opt-btn_tGeneral_bUseLootSound")

	if btnLowstRollWins ~= nil then
		btnLowstRollWins:SetCheck(self.tSettings:Read("tLeader.bAscending"))
	end

	if not btnOpenOnLoot:IsChecked() then
		self:OptionSelectionButtonLinkerHelper(btnOverlay, self.tSettings:Read("tGeneral.bUseOverlay"),
			"tGeneral", "bUseOverlay")
	end

	if not btnUseSound:IsChecked() then
		self:OptionSelectionButtonLinkerHelper(btnRoll, self.tSettings:Read("tGeneral.bUseRollSound"),
			"tGeneral", "bUseRollSound")
		self:OptionSelectionButtonLinkerHelper(btnLoot, self.tSettings:Read("tGeneral.bUseLootSound"),
			"tGeneral", "bUseLootSound")
	end
end

function RaidLooter:OptionSelectionHelper(strName, bIsChecked)
	local strControl, strSection, strVarName = G.Utils:Split(strName, "_")
	local wndMain = self:GetActiveWindow():GetWindow()

	if strControl == "opt-btn" then
		if strSection ~= nil and strVarName ~= nil then
			wndMain:FindChild(strName):SetCheck(bIsChecked)
			local tSetting = self.tSettings:Read(strSection)
			tSetting[strVarName] = bIsChecked
			self:OptionSelectionLinkerHelper(strName, bIsChecked)

			if (strName == "opt-btn_tGeneral_bUseOverlay") then
				if bIsChecked then
					if not self.tMasterLoot:IsEmpty() then
						self.wndAmount:SetText(self.tMasterLoot:GetNumberOfItems())
						self.wndOverlay:Show(true)
					end
				else
					self.wndOverlay:Show(false)
				end
			end
		end
	end
end

function RaidLooter:OptionSelectionLinkerHelper(strName, bIsChecked)
	local wndMain = self:GetActiveWindow():GetWindow()
	if strName == "opt-btn_tGeneral_bOpenOnLoot" then
		local btnOverlay = wndMain:FindChild("opt-btn_tGeneral_bUseOverlay")

		self:OptionSelectionButtonLinkerHelper(btnOverlay, bIsChecked, "tGeneral", "bUseOverlay")

		if not bIsChecked then
			self.wndOverlay:Show(false)
		end
	elseif strName == "opt-btn_tGeneral_bUseSound" then

		local btnRoll = wndMain:FindChild("opt-btn_tGeneral_bUseRollSound")
		local btnLoot = wndMain:FindChild("opt-btn_tGeneral_bUseLootSound")

		self:OptionSelectionButtonLinkerHelper(btnRoll, bIsChecked, "tGeneral", "bUseRollSound")
		self:OptionSelectionButtonLinkerHelper(btnLoot, bIsChecked, "tGeneral", "bUseLootSound")
	end
end

function RaidLooter:OptionSelectionButtonLinkerHelper(btn, bIsChecked, strSection, strOption)
	if bIsChecked then
		btn:Enable(true)
		btn:SetTextColor("UI_TextHoloTitle")
		btn:SetBGColor("ffffffff")
	else
		btn:SetTextColor("6f909090")
		btn:SetBGColor("4fffffff")
		btn:Enable(false)
		btn:SetCheck(false)
		local tSetting = self.tSettings:Read(strSection)
		tSetting[strOption] = false
	end
end

----------------------------------------------------------------------------------
-- Getters & Setters
----------------------------------------------------------------------------------
function RaidLooter:IsActive() return self.tActiveWindow ~= nil end
function RaidLooter:ShouldBeActive() return G.Debug.bIsTest or G.Utils:IsMasterLootingOn() end
function RaidLooter:GetSettings() return self.tSettings end
function RaidLooter:GetMasterLoot() return self.tMasterLoot end
function RaidLooter:GetLootQueue() return self.tLootQueue end
function RaidLooter:GetChatParser() return self.tChatParser end
function RaidLooter:GetAutoRoll() return self.tAutoRoll end
function RaidLooter:CreateWindow(strAlias, wndParent, tEventHandler)
	return self.tXMLFactory:CreateWindow(strAlias, wndParent, tEventHandler) end
function RaidLooter:CreateTopLevelWindow(strAlias, tEventHandler)
	return self.tXMLFactory:CreateWindow(strAlias, nil, tEventHandler) end
function RaidLooter:CreateTooltipWindow(strAlias, tEventHandler)
	return self.tXMLFactory:CreateWindow(strAlias, "TooltipStratum", tEventHandler) end
function RaidLooter:GetLootItem(nLootId) return self:GetActiveWindow():GetLootItem(nLootId) end
function RaidLooter:GetLootItems() return self:GetActiveWindow():GetLootItems() end
function RaidLooter:GetActiveWindow() return self.tActiveWindow end

----------------------------------------------------------------------------------
-- Event handlers
----------------------------------------------------------------------------------
function RaidLooter:HandleOnLootUpdate(tLoot)
	local bOpenOnLoot = self.tSettings:Read("tGeneral.bOpenOnLoot")
	local bUseOverlay = self.tSettings:Read("tGeneral.bUseOverlay")
	local wndOpen = self:GetActiveWindow()

	if bOpenOnLoot then
		if self.tMasterLoot:IsEmpty() then
			self.wndOverlay:Show(false)

			if self.tMasterLoot:RecentlyHadLoot() then
				wndOpen:Close()
			end
		else
			if bUseOverlay then
				self.wndAmount:SetText(self.tMasterLoot:GetNumberOfItems())
				self.wndOverlay:Show(true)
			end

			if self.tMasterLoot:RecentlyHadLootDropped() then
				G.Utils:PlayNewLootSound()

				if not bUseOverlay then
					wndOpen:Open()
				end
			end
		end
	else
		self.wndOverlay:Show(false)
	end
end

function RaidLooter:HandleOnRaidLooterCmd(cmd, params)
	local value
	if params then
		cmd, value = G.Utils:Split(params, " ")
	end

	G.Utils:Log(self)

	if cmd == "dev-mode" then
		G.Debug:SetDevMode(value == "on")
	elseif cmd == "test-mode" then
		G.Debug:SetTestMode(value == "on")
	elseif cmd == "reset" then
		self.tSettings:Reset()
		ChatSystemLib.Command("/reloadui")
	else
		if self:IsActive() then
			self:ToggleActiveWindow()
		else
			G.Utils:Log("Currently in the deactive state need to load XMLs ...")
			self.tXMLFactory:LoadAll("OnForceLoad", self)
		end
	end
end

function RaidLooter:HandleOnInterfaceMenuListHasLoaded()
	local strEventName = "InterfaceMenuList_NewAddOn"
	local tInfo = {"RaidLooterInterfaceList", "", "BK3:sprHolo_Alert_COMMAttachment_DruseraIcon"}
    Event_FireGenericEvent(strEventName, "RaidLooter", tInfo)
end

function RaidLooter:HandleOnJoinedGroup() self:ActivateAddon() end
function RaidLooter:HandleOnLeftGroup() self:DeactivateAddon() end
function RaidLooter:HandleOnDisbandedGroup() self:DeactivateAddon() end
function RaidLooter:HandleOnGroupLootRulesChanged()
	if self:ShouldBeActive() then
		self:ActivateAddon()
	else
		self:DeactivateAddon()
	end
end
function RaidLooter:HandleOnToggleGroupBag() self:ToggleActiveWindow() end

function RaidLooter:HandleOnGroupFlagsChanged()
	if G.Utils:IsMasterLootingOn() then
		self:ActivateAddon()
	else
		self:DeactivateAddon()
	end
end

function RaidLooter:HandleOnMemberGroupFlagsChanged(nMemberIdx, bFromPromotion, tChangedFlags)
	if self:IsActive() then
		local tMember = GroupLib.GetGroupMember(nMemberIdx)
		local tUnit   = GroupLib.GetUnitForGroupMember(nMemberIdx)

		if tMember ~= nil and tUnit ~= nil then
			local bIsConcerningYou = tUnit:IsThePlayer()

			if bIsConcerningYou and bFromPromotion then
				local bBecomeLeader = not tMember.bIsLeader

				self:UpdateActiveWindows(bBecomeLeader, self.tActiveWindow:IsShown())
				self.tMasterLoot:Update()
			end
		end
	end
end

function RaidLooter:UpdateActiveWindows(bBecomeLeader, bWasOpen)
	if (    bBecomeLeader and not self.tActiveWindow:IsManagementWindow()) or
	   (not bBecomeLeader and     self.tActiveWindow:IsManagementWindow())
	then
		self.tActiveWindow:Hide()
		self.tActiveWindow:Destroy()
		self.tActiveWindow = nil
	end

	if bBecomeLeader then
		self.tActiveWindow = self.tManagementWindowFactory:New(self)
	else
		self.tActiveWindow = self.tLootWindowFactory:New(self)
	end

	self.tActiveWindow:Init(self)

	if bWasOpen then
		self.tActiveWindow:Open()
	end
end

function RaidLooter:HandleOnMasterLootUnchecked()
	if G.Utils:IsMasterLootingOn() then
		if self:IsActive() then
			self:OnDeactivateAddon()
		end

		self:ActivateAddon()
	else
		self:DeactivateAddon()
	end
end

function RaidLooter:HandleOnActivateAddon()
	self.wndOverlay = self:CreateTopLevelWindow("Overlay_Square", self)
	self.wndAmount = self.wndOverlay:FindChild("Amount")
	G.Utils:MoveWindow(self.wndOverlay, self.tSettings:ReadBlock("tWindowLocations.wndOverlayLoc"))

	if G.Utils:IsLeader() then
		self.tActiveWindow = self.tManagementWindowFactory:New(self)
	else
		self.tActiveWindow = self.tLootWindowFactory:New(self)
	end

	self.tActiveWindow:Init(self)
	self.tActiveWindow:Hide()

	Apollo.RegisterEventHandler("Group_FlagsChanged",       "OnGroupFlagsChanged", self)
	Apollo.RegisterEventHandler("Group_Leave",              "OnLeftGroup", self)
	Apollo.RegisterEventHandler("Group_Disband",            "OnDisbandedGroup", self)
	Apollo.RegisterEventHandler("Group_MemberFlagsChanged", "OnMemberGroupFlagsChanged", self)

	self.tMasterLoot:Update()
	self.tChatParser:Start()

	if not self.tMasterLoot:IsEmpty() and not self.tSettings:Read("tGeneral.bUseOverlay") then
		self.tActiveWindow:Open()
	end
end

function RaidLooter:HandleOnDeactivateAddon()
	self.tChatParser:ForceStop()
	self.tXMLFactory:UnloadAll()

	if self.tActiveWindow ~= nil then
		self.tActiveWindow:Hide()
		self.tActiveWindow:Destroy()
		self.tActiveWindow = nil
	end

	Apollo.RemoveEventHandler("Group_FlagsChanged", self)
	Apollo.RemoveEventHandler("Group_Leave", self)
	Apollo.RemoveEventHandler("Group_Disband", self)
	Apollo.RemoveEventHandler("Group_MemberFlagsChanged", self)
end

function RaidLooter:HandleOnForceLoad()
	self:OnActivateAddon()
	self:ToggleActiveWindow()
	self.tMasterLoot:Update()
	G.Utils:PlaySound(190)
end

function RaidLooter:HandleOnToggleWindows(wndHandler, wndControl, eMouseButton)
	self:ToggleActiveWindow()
end

function RaidLooter:HandleOnOverlayLockToggled(wndHandler, wndControl, eMouseButton)
	local nLeft, nTop, nRight, nBottom = self.wndOverlay:GetAnchorOffsets()
	local nChange = 25

	if wndControl:IsChecked() then
		self.wndOverlay:SetBGColor("0x3AFFFFFF")
		nLeft   = nLeft - nChange
		nTop    = nTop - nChange
		nRight  = nRight + nChange
		nBottom = nBottom + nChange
	else
		self.wndOverlay:SetBGColor("0x00FFFFFF")
		nLeft   = nLeft + nChange
		nTop    = nTop + nChange
		nRight  = nRight - nChange
		nBottom = nBottom - nChange
	end

	self.wndOverlay:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function RaidLooter:OnRaidLooterCmd(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnRaidLooterCmd", ...) end
function RaidLooter:OnInterfaceMenuListHasLoaded(...)
	G.Utils:InvokeEventHandler(self, "RaidLooter", "OnInterfaceMenuListHasLoaded", ...) end
function RaidLooter:OnJoinedGroup(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnJoinedGroup", ...) end
function RaidLooter:OnGroupLootRulesChanged(...)
	G.Utils:InvokeEventHandler(self, "RaidLooter", "OnGroupLootRulesChanged", ...) end
function RaidLooter:OnLeftGroup(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnLeftGroup", ...) end
function RaidLooter:OnDisbandedGroup(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnDisbandedGroup", ...) end
function RaidLooter:OnGroupFlagsChanged(...)
	G.Utils:InvokeEventHandler(self, "RaidLooter", "OnGroupFlagsChanged", ...) end
function RaidLooter:OnMemberGroupFlagsChanged(...)
	G.Utils:InvokeEventHandler(self, "RaidLooter", "OnMemberGroupFlagsChanged", ...) end
function RaidLooter:OnMasterLootUnchecked(...)
	G.Utils:InvokeEventHandler(self, "RaidLooter", "OnMasterLootUnchecked", ...) end
function RaidLooter:OnToggleGroupBag(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnToggleGroupBag", ...) end
function RaidLooter:OnActivateAddon(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnActivateAddon", ...) end
function RaidLooter:OnDeactivateAddon(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnDeactivateAddon", ...) end
function RaidLooter:OnForceLoad(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnForceLoad", ...) end
function RaidLooter:OnLootUpdate(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnLootUpdate", ...) end
function RaidLooter:OnToggleWindows(...) G.Utils:InvokeEventHandler(self, "RaidLooter", "OnToggleWindows", ...) end
function RaidLooter:OnOverlayLockToggled(...)
	G.Utils:InvokeEventHandler(self, "RaidLooter", "OnOverlayLockToggled", ...) end
