--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local GameLib = require "GameLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raider
-- Represents a member of the raid party
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local Raider = {}
G.Raider = Raider
function Raider:New(tRaidLooter, wndParent, tUnit, strUnitName)
	local o = {
		wndParent      = wndParent,
		wndHandle      = nil,
		wndIcon        = nil,
		wndName        = nil,
		wndBG          = nil,
		wndAnim        = nil,
		wndInspectAnim = nil,
		wndWarn        = nil,
		wndInset       = nil,
		btn            = nil,
		tCallback      = nil,
		tUnit          = tUnit,
		strName        = "",
		strUnitName    = strUnitName,
		strClassSprite = "",
		bIsMouseInside = false,
		bIsSelected    = false,
		bIsOutOfRange  = false,
		tRaidLooter    = tRaidLooter,
		tLinker        = G.Linker:New(),
		nClassId       = -1,
	}
	setmetatable(o, self)
	self.__index = self
	return o
end

function Raider:Init(tTarget)
	self.tLinker:Link(NLLC, self)

    self.wndHandle      = self.tRaidLooter:CreateWindow("Raider_HalfRow", self.wndParent, self)
	self.btn            = self.wndHandle:FindChild("Button")
	self.wndIcon        = self.wndHandle:FindChild("ClassIcon")
	self.wndName        = self.wndHandle:FindChild("Name")
	self.wndInset       = self.wndHandle:FindChild("Inset")
	self.wndBG          = self.wndHandle:FindChild("Background")
	self.wndAnim        = self.wndHandle:FindChild("SelectedAnimation")
	self.wndWarn        = self.wndHandle:FindChild("Indicator")

	self.wndAnim:Show(true, true)
	self.wndAnim:SetBGColor("00ffffff")
	self.wndWarn:Show(false, true)
	self.wndName:SetText(self.strName)

	-- Events
	self.btn:AddEventHandler("MouseEnter",   "OnMouseEnter",    self)
	self.btn:AddEventHandler("MouseExit",    "OnMouseExit",     self)
	self.btn:AddEventHandler("ButtonSignal", "OnButtonPressed", self)
	self.btn:AddEventHandler("MouseButtonUp", "OnMouseButtonUp", self)
	self.bIsOutOfRange = self.tUnit == nil or not self.tUnit:IsValid()
	self:UpdateOutOfRange(self.tUnit, self.strUnitName)
end

function Raider:Destroy()
	self.tLinker:Destroy()
	self.wndHandle:Destroy()

	self.btn:RemoveEventHandler("MouseEnter", self)
	self.btn:RemoveEventHandler("MouseExit", self)
	self.btn:RemoveEventHandler("ButtonSignal", self)
	self.btn:RemoveEventHandler("MouseButtonUp", self)

	self.wndHandle   = nil
	self.tLinker     = nil
	self.tRaidLooter = nil
	self.tLinker     = nil
	self.btn         = nil
	self.tCallback   = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function Raider:UpdateOutOfRange(tUnit, strUnitName)
	local bIsOutOfRange = tUnit == nil or not tUnit:IsValid()
	local strText

	if bIsOutOfRange then
		self.strName        = strUnitName or self.strName
		self.strClassSprite = self.strClassSprite or ""
		self.nClassId       = self.nClassId ~= -1 and self.nClassId or -1
		strText = "?"
	else
		self.strName        = tUnit:GetName() or self.strName
		self.strClassSprite = G.Utils:GetClassIcon(tUnit)
		self.nClassId       = tUnit:GetClassId()
		strText = ""
	end

	if self.strClassSprite == nil then
		self.wndIcon:SetText(strText)
	end
	self.wndIcon:SetSprite(self.strClassSprite)

	local strFullName = self.strName
	if self.strName ~= nil then
		self.strName = self.strName:gmatch("[^ ]+")()
	end

	self:SetIsOutOfRange(bIsOutOfRange)
	self.wndName:SetText(self.strName)

	if strFullName then
		self.wndHandle:SetTooltip(strFullName..(bIsOutOfRange and " (Out of Range)" or ""))
	end
end

function Raider:Unselect()
	if self.bIsSelected then
		self.bIsSelected = false
		self.btn:SetCheck(self.bIsSelected)

		-- Workaround :Show(false) triggers MouseExit/MouseEnter
		self.wndAnim:SetBGColor("00ffffff")

		if not self.bIsMouseInside then
			self.wndBG:SetBGColor("0fffffff")
		end

		G.Utils:FireEvent(self.tCallback, "OnRecipientUnselected", self)
	end
end

function Raider:Select()
	if not self.bIsSelected then
		self.bIsSelected = true
		self.btn:SetCheck(self.bIsSelected)

		-- Workaround :Show(false) triggers MouseExit/MouseEnter
		self.wndAnim:SetBGColor("ffffffff")

		if not self.bIsMouseInside then
			self.wndBG:SetBGColor("2fffffff")
		end

		G.Utils:PlaySound(86)
		G.Utils:FireEvent(self.tCallback, "OnRecipientSelected", self)
	end
end

function Raider:Show()
	self.wndHandle:Enable(true)
	self.wndHandle:Show(true)
end

function Raider:Hide()
	self.wndHandle:Enable(false)
	self.wndHandle:Show(false)
end

function Raider:SetUnit(tUnit)
	self.tUnit = tUnit
	self:UpdateOutOfRange(tUnit)
end

function Raider:SetAnchor(tAnchor)
	self.wndHandle:SetAnchorPoints(unpack(tAnchor.tPoints))
	self.wndHandle:SetAnchorOffsets(unpack(tAnchor.tOffsets))
end

function Raider:SetIsOutOfRange(bIsOutOfRange) self.bIsOutOfRange = bIsOutOfRange end
function Raider:SetCallback(tCallback) self.tCallback = tCallback end
function Raider:GetClassSprite() return self.strClassSprite end
function Raider:GetFullname() return self.strUnitName end
function Raider:GetUnit() return self.tUnit end
function Raider:IsNamed(strUnitName) return self.strUnitName == strUnitName end
function Raider:IsSelected() return self.bIsSelected end
function Raider:GetClassId() return self.nClassId end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function Raider:OnButtonPressed(wndHandler, wndControl, eMouseButton)
	if eMouseButton == GameLib.CodeEnumInputMouse.Left then
		if self.bIsSelected then
			self:Unselect()
		else
			self:Select()
		end
	end
end

function Raider:OnMouseEnter(wndHandler, wndControl, x, y)
	self.bIsMouseInside = true
	self.wndBG:SetBGColor("5fffffff")
end

function Raider:OnMouseExit(wndHandler, wndControl, x, y)
	local a = self.bIsSelected and 2 or 0
	self.bIsMouseInside = false
	self.wndBG:SetBGColor(a .. "fffffff")
end
