--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- RecipientCollection
-- Represents a collection of raiders that are eligible recipients for a
-- specific undistributed item.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local RecipientCollection = {}
G.RecipientCollection = RecipientCollection

----------------------------------------------------------------------------------
-- Initialization
----------------------------------------------------------------------------------
function RecipientCollection:New(tRaidLooter, wndTarget)
	local o = {
		nMaxRows         = 10,
		nColumns         = 2,
		wndHandle        = wndTarget,
		tRecipients      = {},
		tSelection       = nil,
		tInspector       = nil,
		tCallback        = nil,
		strCurrentFilter = "A",
		tRaidLooter      = tRaidLooter,
		tLinker          = G.Linker:New(),
		bShouldFilter    = true
	}

	setmetatable(o, self)
	self.__index = self
	return o
end

function RecipientCollection:Init(tTarget)
	self.tLinker:Link(NLLC, self)
end

function RecipientCollection:Destroy()
	self.tLinker:Destroy()
	for _, tRecipient in pairs(self.tRecipients) do
		tRecipient:Destroy()
	end

	self.tRecipients = nil
	self.wndHandle   = nil
	self.tLinker     = nil
	self.tCallback   = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function RecipientCollection:Clear()
	self.tSelection = nil

	for _, tRecipient in pairs(self.tRecipients) do
		tRecipient:Destroy()
	end

	self.tRecipients = {}
end

function RecipientCollection:FindRecipientByName(strName)
	for _, tRecipient in pairs(self.tRecipients) do
		if tRecipient:IsNamed(strName) then
			return tRecipient
		end
	end

	return nil
end

function RecipientCollection:Update(tEligibleRecipients, bShouldKeepSelection)
	-- Copy all current recipients.
	local tKillRecipients = {}
	for _, tRecipients in ipairs(self.tRecipients) do
		tKillRecipients[tRecipients:GetFullname()] = tRecipients
	end

	-- Insert/update recipients.
	for _, tLooterInfo in pairs(tEligibleRecipients) do
		if tLooterInfo.strUnitName ~= nil and tLooterInfo.strUnitName ~= "" then
			local tExistingRecipient = self:FindRecipientByName(tLooterInfo.strUnitName)

			if tExistingRecipient ~= nil then
				local bIsOutOfRange = tLooterInfo.tUnit == nil or not tLooterInfo.tUnit:IsValid()
				tExistingRecipient:SetUnit(tLooterInfo.tUnit)
				tExistingRecipient:SetIsOutOfRange(bIsOutOfRange)
				tKillRecipients[tExistingRecipient:GetFullname()] = nil
			else
				local tNewRecipient = G.Raider:New(self.tRaidLooter, self.wndHandle, tLooterInfo.tUnit, tLooterInfo.strUnitName)
				tNewRecipient:Init(self)
				self:AddRecipient(tNewRecipient)
			end
		end
	end

	-- Remove remaining recipients.
	for nKey, tRecipients in pairs(tKillRecipients) do
		self:RemoveRecipient(tRecipients)
	end

	self:ArrangeAndFilter()

	if bShouldKeepSelection == nil or not bShouldKeepSelection then
		self:UnselectCurrent()
	end
end

function RecipientCollection:ApplyFilter()
	self:ArrangeAndFilter()
end

function RecipientCollection:AddRecipient(tRecipient)
	tRecipient:SetCallback(self)
	table.insert(self.tRecipients, tRecipient)
end

function RecipientCollection:RemoveRecipient(tRecipient)
	if tRecipient == self.tSelection then
		self.tSelection = nil
	end

	for nIndex, tExistingRecipient in ipairs(self.tRecipients) do
		if tRecipient:GetFullname() == tExistingRecipient:GetFullname() then
			tRecipient:Destroy()
			table.remove(self.tRecipients, nIndex)
			break
		end
	end
end

function RecipientCollection:ArrangeAndFilter()
	local nHF = 1.0 / self.nMaxRows -- Horizontal factor
	local nWF = 1.0 / self.nColumns -- Vertical factor
	local nSkipCount = 0

	for nIndex, tRecipient in pairs(self.tRecipients) do
		nIndex = nIndex - 1 - nSkipCount

		local bShouldNotFilterOut = true
		if self.bShouldFilter then
			local nClassId = tRecipient:GetClassId()
			if nClassId ~= -1 then
				bShouldNotFilterOut = G.Utils:HaveArmorType(nClassId, self.strCurrentFilter)
			end
		end

		if bShouldNotFilterOut then
			local nRow = math.floor(nIndex / self.nColumns)
			local nCol = nIndex % self.nColumns

			tRecipient:SetAnchor({
				tPoints  = {nWF * nCol, nHF * nRow, nWF * (nCol+1), nHF * (nRow+1)},
			    tOffsets = {0, 0, -1, -1}
			})
			tRecipient:Show()
		else
			tRecipient:Hide()
			nSkipCount = nSkipCount + 1
		end
	end
end

function RecipientCollection:SetCallback(tCallback)
	self.tCallback = tCallback
end

function RecipientCollection:GetSelectedRecipient()
	return self.tSelection
end

function RecipientCollection:GetRaider(strUnitName)
	for _, tRecipient in pairs(self.tRecipients) do
		if tRecipient:IsNamed(strUnitName) then
			return tRecipient
		end
	end

	return nil
end

function RecipientCollection:UnselectCurrent()
	if self.tSelection ~= nil then
		self.tSelection:Unselect()
	end
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function RecipientCollection:OnRecipientSelected(tRecipient)
	if self.tSelection ~= nil then
		local tToUnselect = self.tSelection -- Break the recursion
		self.tSelection = nil
		tToUnselect:Unselect()
	end

	local tOldRecipient = self.tSelection
	self.tSelection = tRecipient

	G.Utils:FireEvent(self.tCallback, "OnRecipientSelectionChanged", tOldRecipient, tRecipient)
end

function RecipientCollection:OnRecipientUnselected(tRecipient)
	local tOldRecipient = self.tSelection
	self.tSelection = nil

	G.Utils:FireEvent(self.tCallback, "OnRecipientSelectionChanged", tOldRecipient, nil)
end

function RecipientCollection:OnFilterChange(wndHandler, wndControl, eMouseButton)
	if wndControl:IsChecked() and self.bShouldFilter then
		local strButtonName = wndControl:GetName()
		local _, strButtonId = G.Utils:Split(strButtonName, "_")
		local strFilter;

		if strButtonId == "AllButton" then
			strFilter = "A"
		elseif strButtonId == "HButton" then
			strFilter = "H"
		elseif strButtonId == "MButton" then
			strFilter = "M"
		elseif strButtonId == "LButton" then
			strFilter = "L"
		else
			strFilter = "A"
		end

		if strFilter ~= self.strCurrentFilter then
			self.strCurrentFilter = strFilter
			G.Utils:Log("Filter settings changed to: " .. strFilter)
			self:ApplyFilter()
			self:UnselectCurrent()
		end
	end
end

function RecipientCollection:OnFilterRecipients(wndHandler, wndControl, eMouseButton)
	self.bShouldFilter = wndControl:IsChecked()
	self:ApplyFilter()
end
