--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Roller
-- Represents a raider which have rolled for an item.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local Roller = {}
G.Roller = Roller
function Roller:New(tRaidLooter, wndParent, bIsSelectable)
	local o = {
		bIsSelectable  = bIsSelectable,
		wndParent      = wndParent,
		strUnitName    = "",
		strDisplayName = "",
		nScore         = nil,
		nRank          = nil,
		bDisplayRank   = true,
		bDisplayBase   = false,
		bDisplayOrder  = false,
		eRollType      = 0, --0=mainspec, 1=offspec, 2=costume
		wndHandle      = nil,
		wndRank        = nil,
		wndRaider      = nil,
		wndScore       = nil,
		wndAnimation   = nil,
		tRaidLooter    = tRaidLooter,
		tLinker        = G.Linker:New(),
		tCallback      = nil,
	}

    setmetatable(o, self)
    self.__index = self
    return o
end

function Roller:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	self.wndHandle    = self.tRaidLooter:CreateWindow("Roller_Row", self.wndParent, self)
	self.wndRank      = self.wndHandle:FindChild("Rank")
	self.wndRaider    = self.wndHandle:FindChild("Name")
	self.wndScore     = self.wndHandle:FindChild("Roll")
	self.wndAnimation = self.wndHandle:FindChild("LeaderAnimation")
	self.wndButton    = self.wndHandle:FindChild("Button")
end

function Roller:Destroy()
	self.tLinker:Destroy()
	self.wndHandle:Destroy()
	self.wndHandle = nil
	self.wndParent = nil
	self.tRaidLooter = nil
	self.tLinker = nil
	self.tCallback = nil
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function Roller:Set(nRank, strUnitName, nScore, eRollType)
	self.strUnitName = strUnitName
	self.nScore      = nScore
	self.nRank       = nRank

	self:SetRollType(eRollType)

	self.wndScore:SetText(self.nScore)
	self.wndAnimation:Show(self.nRank == 1, true)
end

function Roller:SetCallback(tCallback)
	self.tCallback = tCallback
	if tCallback then
		self.wndButton:AddEventHandler("ButtonSignal", "OnButtonPressed", self)
	else
		self.wndButton:RemoveEventHandler("ButtonSignal", self)
	end
end

function Roller:SetRollType(eRollType)
	self.eRollType = 0 --0 = mainspec
	self.eRollType = eRollType or self.eRollType
	self:UpdateDisplayText()
end

function Roller:SetAnchor(tAnchor)
	local tP = tAnchor.tPoints
	local tO = tAnchor.tOffsets

	self.wndHandle:SetAnchorPoints(tP[1], tP[2], tP[3], tP[4])
	self.wndHandle:SetAnchorOffsets(tO[1], tO[2], tO[3], tO[4])
end

function Roller:GetHeight()
	return self.wndHandle:GetHeight()
end

function Roller:SetRank(nRank)
	self.nRank = nRank
	self:UpdateDisplayText()
	self.wndAnimation:Show(self.nRank == 1, true)
end

function Roller:DisplayRank()
	self.bDisplayRank   = true
	self.bDisplayBase   = false
	self.bDisplayOrder  = false

	self:UpdateDisplayText()
end

function Roller:DisplayBase()
	self.bDisplayRank   = false
	self.bDisplayBase   = true
	self.bDisplayOrder  = false

	self:UpdateDisplayText()
end

function Roller:DisplayOrder()
	self.bDisplayRank   = false
	self.bDisplayBase   = false
	self.bDisplayOrder  = true

	self:UpdateDisplayText()
end

function Roller:GetRank()
	return self.nRank
end

function Roller:IsNamed(strUnitName)
	return self.strUnitName == strUnitName
end

function Roller:GetScore()
	return self.nScore
end

function Roller:GetName()
	return self.strUnitName
end

function Roller:GetOrderValue() --tRanks is optional
	local nRank;
	if self:IsRollingMainspec() then
		nRank = NLLC.Ranker:GetLootRank(self.strUnitName, true)
	elseif self:IsRollingOffspec() then
		nRank = NLLC.Ranker:GetLootRank(self.strUnitName, false)
	end

	if not nRank then
		return -10
	else
		return nRank
	end
end

function Roller:GetBaseOrderValue()
	local nRank = NLLC.Ranker:GetBaseRank(self.strUnitName)

	if not nRank then
		return -10
	else
		return nRank
	end
end

function Roller:IsRollingMainspec()
	return self.eRollType == 0 or not self.eRollType --nil or 0
end

function Roller:IsRollingOffspec()
	return self.eRollType == 1
end

function Roller:IsRollingCostume()
	return self.eRollType == 2
end

function Roller:UpdateDisplayText()
	self.strDisplayName = self.strUnitName:gmatch("[^ ]+")()
	local strRollType = ""
	if self.eRollType == 1 then
		strRollType = " (os)"
	elseif self.eRollType == 2 then
		strRollType = " (c)"
	end
	self.wndRaider:SetText(self.strDisplayName .. strRollType)

	if self.bDisplayRank then
		local rank;
		if self:IsRollingMainspec() then
			rank = NLLC.Ranker:GetLootRank(self.strUnitName, true) or "/"
		elseif self:IsRollingOffspec() then
			rank = NLLC.Ranker:GetLootRank(self.strUnitName, false) --returns nRank or nil/false for noData/noOffspec
			if rank == false then rank = "-" end
			if rank == nil then rank = "/" end
		else
			rank = "c"
		end

		local tColors = NLLC.Colors:GetColors(NLLC.Colors.EnumColorType.Rank)
		self.wndRank:SetTextColor(tColors[rank] or "ffffffff")
		self.wndRank:SetText(rank)
	elseif self.bDisplayBase then
		local rank = NLLC.Ranker:GetBaseRank(self.strUnitName) or "/"

		local tColors = NLLC.Colors:GetColors(NLLC.Colors.EnumColorType.Rank)
		self.wndRank:SetTextColor(tColors[rank] or "ffffffff")
		self.wndRank:SetText(rank)
	elseif self.bDisplayOrder then
		local rank = self:GetRank()

		self.wndRank:SetTextColor("ffffffff")
		self.wndRank:SetText(rank)
	end
end


function Roller:OnButtonPressed(...) G.Utils:InvokeEventHandler(self, "Roller", "OnButtonPressed", ...) end
function Roller:HandleOnButtonPressed(wndHandler, wndControl, eMouseButton)
	if self.tCallback then
		G.Utils:FireEvent(self.tCallback, "OnRollerClicked", self)
	end
end
