--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local GameLib = require "GameLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

local EnumRollerSorting = {
	Rank = 1,
	Base = 2,
	Roll = 3,
}
local tSorters = {}
local bNeedRestore = true

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- RollerCollection
-- Represents a unique collection of rollers that are eligible recipients for a
-- specific undistributed item.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local RollerCollection = {}
G.RollerCollection = RollerCollection
function RollerCollection:New(tRaidLooter, bIsSelectable, wndTarget, wndSelector)
	local o = {
		wndHandle     = wndTarget,
		nLastRank     = 0,
		tCallback     = nil,
		tRollers      = {},
		tRollHints    = {},
		tLootItem     = nil,
		bIsParsing    = false,
		bIsSelectable = bIsSelectable,
		tRaidLooter   = tRaidLooter,
		tLinker       = G.Linker:New(),
		bAscending    = false,

		wndSortSel    = wndSelector,
		optSortSel    = NLLC.Options:GetOption("RollerCollection", "SortEnum"),
		tSorter       = nil,
		eSorter       = nil,

		nRestoreKey   = nil,
		nSaveKey      = nil,
	}

  setmetatable(o, self)
  self.__index = self
  return o
end

function RollerCollection:Init(tTarget)
	self.tLinker:Link(NLLC, self)

	self:InitSorters()

	if self.wndSortSel then
		local tSelection = {
			EnumRollerSorting.Rank,
			EnumRollerSorting.Base,
			EnumRollerSorting.Roll,
		}

		local tSelectionTranslation = {
			[EnumRollerSorting.Rank] = "Sort: Calc. Rank",
			[EnumRollerSorting.Base] = "Sort: Base Rank",
			[EnumRollerSorting.Roll] = "Sort: Default",
		}

		self.optSortSel:OnUpdate(self, "OnSortOptionUpdate")
		NLLC.UIComponents:applyDropdown(self.wndSortSel, tSelection, self.optSortSel, tSelectionTranslation)
		self.optSortSel:ForceUpdate()
	end

	if bNeedRestore then
		self.nRestoreKey = NLLC:OnceRestored("OnRestore", self)
	end
	self.nSaveKey = NLLC:OnceSaving("OnSave", self)
end

function RollerCollection:Destroy()
	if self.nRestoreKey then
		NLLC:RemoveOnceRestored(self.nRestoreKey)
	end
	if self.nSaveKey then
		NLLC:RemoveOnceSaving(self.nSaveKey)
	end

	self.tLinker:Destroy()
	for _, tRoller in pairs(self.tRollers) do
		tRoller:Destroy()
	end
end

function RollerCollection:OnRestore(eLevel, tSaveData)
	if GameLib.CodeEnumAddonSaveLevel.Character == eLevel then
		if not tSaveData or not tSaveData.RollerCollection then
			self.optSortSel:Set(EnumRollerSorting.Rank)
		else
			self.optSortSel:Set(tSaveData.RollerCollection.eSortSel or EnumRollerSorting.Rank)
		end
		bNeedRestore = false
	end
end

function RollerCollection:OnSave(eLevel, tSaveData)
	if GameLib.CodeEnumAddonSaveLevel.Character == eLevel then
		if not tSaveData.RollerCollection then
			tSaveData.RollerCollection = {
				eSortSel = self.optSortSel:Get(),
			}
		end
	end
end

----------------------------------------------------------------------------------
-- Functions
----------------------------------------------------------------------------------
function RollerCollection:BeginRollingForItem(tLootItem, bAscending)
	if not self.bIsParsing and tLootItem ~= nil then
		self.tLootItem  = tLootItem
		self.bIsParsing = true
		self.tRollHints = {} -- Reset hints.
		self.bAscending = bAscending
	end
end

function RollerCollection:EndRolling()
	if self.bIsParsing then
		self.bIsParsing = false
	end
end

function RollerCollection:IsEmpty()
	return self.tRollers == nil or #self.tRollers == 0
end

function RollerCollection:GetItem()
	return self.tLootItem
end

function RollerCollection:SetCallback(tCallback)
	self.tCallback = tCallback
end

function RollerCollection:Clear()
	self.tLootItem = nil
	for _, tRoller in pairs(self.tRollers) do
		tRoller:Destroy()
	end

	self.tRollers  = {}
	self.nLastRank = 0
end

function RollerCollection:Contains(strUnitName)
	return self:GetRoller(strUnitName) ~= nil
end

function RollerCollection:GetRoller(strUnitName)
	for _, tRoller in pairs(self.tRollers) do
		if tRoller:IsNamed(strUnitName) then
			return tRoller
		end
	end

	return nil
end

function RollerCollection:UpdateRollHint(strUnitName, tHint)
	G.Utils:Log("UpdateRollHint {" .. strUnitName .. "}")
	if strUnitName ~= nil and tHint ~= nil then
		local eRollType   = 0
		local tRoller     = self:GetRoller(strUnitName)

		if tHint.bIsOffspec then
			eRollType = 1
		elseif tHint.bIsCostume then
			eRollType = 2
		end

		self.tRollHints[strUnitName] = eRollType

		if tRoller ~= nil then
			local nRank = tRoller:GetRank()

			table.remove(self.tRollers, nRank)

			local bWasTied = self:GetTie()
			local nNewRank, bWillTie = self:GetPredicatedRank(tRoller:GetName(), tRoller:GetScore())

			table.insert(self.tRollers, nNewRank, tRoller)

			if nRank ~= nNewRank then
				if nNewRank == 1 then
					G.Utils:FireEvent(self.tCallback, "OnRollerEvent_NewLeader", nNewRank, strUnitName, tRoller:GetScore())
				end
				tRoller:SetRank(nNewRank)
				self:ArrangeFromRank(math.min(nRank, nNewRank))
			end


			if bWasTied ~= bWillTie then
				G.Utils:FireEvent(self.tCallback, "OnRollerEvent_Tied", bWillTie)
			end

			tRoller:SetRollType(eRollType)
		end
	end
end

function RollerCollection:AddRoller(strUnitName, nScore)
	if G.Debug.bIsTest or not self:Contains(strUnitName) then
		self.nLastRank = self.nLastRank + 1

		local nRank, bTied = self:GetPredicatedRank(strUnitName, nScore)

		if nRank == 1 then
			G.Utils:FireEvent(self.tCallback, "OnRollerEvent_NewLeader", nRank, strUnitName, nScore)
			if bTied then
				G.Utils:FireEvent(self.tCallback, "OnRollerEvent_Tied", bTied)
			end
		end

		local eRollType = self.tRollHints[strUnitName]
		local tRoller = self:CreateNewRoller(nRank, strUnitName, nScore, eRollType)
		table.insert(self.tRollers, nRank, tRoller)

		if nRank == self.nLastRank then
			self.nRollerHeight = tRoller:GetHeight()
		end

		self:ArrangeFromRank(nRank)
		self.wndHandle:RecalculateContentExtents()
	end
end

----------------------------------------------------------------------------------
-- Private functions
----------------------------------------------------------------------------------
function RollerCollection:CreateNewRoller(nRank, strUnitName, nScore, nOrderValue)
	local newRoller = G.Roller:New(self.tRaidLooter, self.wndHandle, self.bIsSelectable)
	newRoller:Init(self)
	newRoller:Set(nRank, strUnitName, nScore, nOrderValue)
	newRoller:SetCallback(self.tCallback)
	if self.eSorter == EnumRollerSorting.Rank then
		newRoller:DisplayRank()
	elseif self.eSorter == EnumRollerSorting.Base then
		newRoller:DisplayBase()
	elseif self.eSorter == EnumRollerSorting.Roll then
		newRoller:DisplayOrder()
	end
	return newRoller
end

function RollerCollection:OrderPredicate(nLeft, nRight)
	if self.bAscending then
		return nLeft < nRight
	else
		return nLeft > nRight
	end
end

--is fine
-- This function also updates the ranks ...
function RollerCollection:ArrangeFromRank(nStartRank)
	local nLeft   =  2
	local nTop;
	local nRight  = -2
	local nBottom;
	local nPad    =  1

	for nRank = nStartRank, self.nLastRank do
		local tRoller = self.tRollers[nRank]
		tRoller:SetRank(nRank)

		nTop    = (nRank - 1) * (self.nRollerHeight + nPad)
		nBottom = nTop + self.nRollerHeight

		tRoller:SetAnchor({
			tPoints  = {0, 0, 1, 0},
		    tOffsets = {nLeft, nTop, nRight, nBottom}
		})
	end
end

--is fine
function RollerCollection:GetSummaryString()
	local tTopRoller, bIsTie = self:GetDistictTop()

	if tTopRoller ~= nil then
		return "Best roll: " .. tTopRoller:GetName() .. " (" .. tTopRoller:GetScore() .. ")."
	elseif bIsTie then
		return "Tie between: " .. self:FormatTopTie()
	end

	return "No rolls."
end

--is fine
function RollerCollection:FormatTopTie()
	local nRollers = #self.tRollers
	local strResult   = ""

	if nRollers == 0 then
		return strResult
	end

	local _, tTieBetween = self:GetTie()

	nRollers = #tTieBetween
	for nIndex, tRoller in ipairs(tTieBetween) do
		strResult = strResult .. tRoller:GetName()

		if nIndex ~= nRollers then
			if nIndex == nRollers - 1 then
				strResult = strResult .. " and "
			else
				strResult = strResult .. ", "
			end
		end
	end

	local nTieScore = self.tRollers[1]:GetScore()
	strResult = strResult .. " (" .. nTieScore .. ")."
	return strResult
end


function RollerCollection:GetPredicatedRank(...)
	return self.tSorter.GetPredicatedRank(self, ...)
end


function RollerCollection:GetTie(...)
	return self.tSorter.GetTie(self, ...)
end

--need to split
-- Note --
-- Returns two variables
--     1. The top roller if any
--     2. If the lead is tied.
--
-- The top roller is the highest main-spec roll followed by
-- off-spec and last costume rolls.
function RollerCollection:GetDistictTop(...)
	return self.tSorter.GetDistictTop(self, ...)
end


function RollerCollection:SelectDistinctTop(tRollers)
	local nRollers = #tRollers

	if nRollers == 0 then
		return nil, false
	elseif nRollers == 1 then
		return tRollers[1], false
	elseif tRollers[1]:GetScore() > tRollers[2]:GetScore() then
		return tRollers[1], false
	else
		-- Tied because the sequence is ordered.
		return nil, true
	end
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function RollerCollection:OnClearButtonPressed(wndHandler, wndControl, eMouseButton)
	self:Clear()
end

function RollerCollection:OnSortOptionUpdate(eSort)
	eSort = eSort or EnumRollerSorting.Rank
	self:ApplySorter(eSort)
	--change the soring!
end

-----------------------------------------------------------------------------------
-- Everything unique for all the sorters
-----------------------------------------------------------------------------------
function RollerCollection:ApplySorter(eSort)
	self.tSorter = tSorters[eSort]
	self.eSorter = eSort
	if eSort == EnumRollerSorting.Rank then
		for i, tRoller in ipairs(self.tRollers) do
			tRoller:DisplayRank()
		end
	elseif eSort == EnumRollerSorting.Base then
		for i, tRoller in ipairs(self.tRollers) do
			tRoller:DisplayBase()
		end
	elseif eSort == EnumRollerSorting.Roll then
		for i, tRoller in ipairs(self.tRollers) do
			tRoller:DisplayOrder()
		end
	end
	self:ResortAll()
end

function RollerCollection:ResortAll()
	local tSorter = self.tSorter

	local tRollers = self.tRollers
	self.tRollers = {table.remove(tRollers, 1)} --start up the table with one from tRollers.

	while #tRollers > 0 do
		local tRoller = table.remove(tRollers, 1)
		local strName, nScore = tRoller:GetName(), tRoller:GetScore()
		local nRank = tSorter.GetPredicatedRank(self, strName, nScore)
		table.insert(self.tRollers, nRank, tRoller)
	end

	self:ArrangeFromRank(1)
end

do
	local initializedSorters = false
	function RollerCollection:InitSorters()
		if not initializedSorters then
			tSorters[EnumRollerSorting.Rank] = self:InitRankSorter()
			tSorters[EnumRollerSorting.Base] = self:InitBaseSorter()
			tSorters[EnumRollerSorting.Roll] = self:InitRollSorter()
			initializedSorters = true
		end
		self.tSorter = tSorters[EnumRollerSorting.Rank]
		self.eSorter = EnumRollerSorting.Rank
	end
end

function RollerCollection:InitRankSorter()
	local RankSorter = {}

	function RankSorter:GetPredicatedRank(strUnitName, nScore)
		-- Insert at the correct position
		local nRank       = 1
		local bTied				= false
		local nCmpScore;

		local eRollType = self.tRollHints[strUnitName]
		local nLootRank = -10
		if not eRollType or eRollType < 2 then
			local bMainspec = not eRollType or eRollType==0
			nLootRank = NLLC.Ranker:GetLootRank(strUnitName, bMainspec) or -10
		end
		local nCmpLootRank;

		for _, tCmpRoller in ipairs(self.tRollers) do
			nCmpScore = tCmpRoller:GetScore()
			nCmpLootRank = tCmpRoller:GetOrderValue()
			if (self:OrderPredicate(nLootRank, nCmpLootRank) or
					nLootRank == nCmpLootRank and self:OrderPredicate(nScore, nCmpScore)) then
				break --correct Rank found
			elseif nLootRank == nCmpLootRank and nScore == nCmpScore and nRank == 1 then
				-- It is a tie
				bTied = true
			end

			nRank = nRank + 1
		end

		return nRank, bTied
	end

	function RankSorter:GetDistictTop()
		local tRollers = {}

		local nMaxOrder = -10
		local nMaxRoll = 0

		for _, tRoller in ipairs(self.tRollers) do
			local nOrder = tRoller:GetOrderValue()
			local nRoll = tRoller:GetScore()
			if nOrder > nMaxOrder then
				tRollers = {tRoller}
				nMaxOrder = nOrder
				nMaxRoll = nRoll
			elseif nOrder == nMaxOrder and nRoll > nMaxRoll then
				tRollers = {tRoller}
				nMaxRoll = nRoll
			elseif nOrder == nMaxOrder and nRoll == nMaxRoll then
				table.insert(tRollers, tRoller)
			end
		end

		return self:SelectDistinctTop(tRollers)
	end

	function RankSorter:GetTie()
		local nTiedOrder;
		local nTiedScore;
		local tTieBetween = {}

		if #self.tRollers == 0 then
			return false, tTieBetween
		end

		for nIndex, tRoller in ipairs(self.tRollers) do
			if nIndex == 1 then
				nTiedOrder = tRoller:GetOrderValue()
				nTiedScore = tRoller:GetScore()
			elseif nTiedOrder ~= tRoller:GetOrderValue() or nTiedScore ~= tRoller:GetScore() then
				break
			end

			table.insert(tTieBetween, tRoller)
		end

		return #tTieBetween > 1, tTieBetween
	end

	return RankSorter
end

function RollerCollection:InitBaseSorter()
	local BaseSorter = {}

	function BaseSorter:GetPredicatedRank(strUnitName, nScore)
		-- Insert at the correct position
		local nRank       = 1
		local bTied				= false
		local nCmpScore;

		local nLootRank = NLLC.Ranker:GetBaseRank(strUnitName) or -10
		local nCmpLootRank;

		for _, tCmpRoller in ipairs(self.tRollers) do
			nCmpScore = tCmpRoller:GetScore()
			nCmpLootRank = tCmpRoller:GetBaseOrderValue()
			if (self:OrderPredicate(nLootRank, nCmpLootRank) or
					nLootRank == nCmpLootRank and self:OrderPredicate(nScore, nCmpScore)) then
				break --correct Rank found
			elseif nLootRank == nCmpLootRank and nScore == nCmpScore and nRank == 1 then
				-- It is a tie
				bTied = true
			end

			nRank = nRank + 1
		end

		return nRank, bTied
	end

	function BaseSorter:GetDistictTop()
		local tRollers = {}

		local nMaxOrder = -10
		local nMaxRoll = 0

		for _, tRoller in ipairs(self.tRollers) do
			local nOrder = tRoller:GetBaseOrderValue()
			local nRoll = tRoller:GetScore()
			if nOrder > nMaxOrder then
				tRollers = {tRoller}
				nMaxOrder = nOrder
				nMaxRoll = nRoll
			elseif nOrder == nMaxOrder and nRoll > nMaxRoll then
				tRollers = {tRoller}
				nMaxRoll = nRoll
			elseif nOrder == nMaxOrder and nRoll == nMaxRoll then
				table.insert(tRollers, tRoller)
			end
		end

		return self:SelectDistinctTop(tRollers)
	end

	function BaseSorter:GetTie()
		local nTiedOrder;
		local nTiedScore;
		local tTieBetween = {}

		if #self.tRollers == 0 then
			return false, tTieBetween
		end

		for nIndex, tRoller in ipairs(self.tRollers) do
			if nIndex == 1 then
				nTiedOrder = tRoller:GetBaseOrderValue()
				nTiedScore = tRoller:GetScore()
			elseif nTiedOrder ~= tRoller:GetBaseOrderValue() or nTiedScore ~= tRoller:GetScore() then
				break
			end

			table.insert(tTieBetween, tRoller)
		end

		return #tTieBetween > 1, tTieBetween
	end

	return BaseSorter
end

function RollerCollection:InitRollSorter()
	local RollSorter = {}

	function RollSorter:GetPredicatedRank(strUnitName, nScore)
		-- Insert at the correct position
		local nRank       = 1
		local bTied				= false

		local nCmpScore;

		local eRollType = self.tRollHints[strUnitName]

		local bMainspec = not eRollType or eRollType==0
		local bCmpMainspec;

		local bOffspec = eRollType==1
		local bCmpOffspec;

		for _, tCmpRoller in ipairs(self.tRollers) do
			nCmpScore = tCmpRoller:GetScore()
			bCmpMainspec = tCmpRoller:IsRollingMainspec()
			bCmpOffspec = tCmpRoller:IsRollingOffspec()

			if not (bMainspec == bCmpMainspec and bOffspec == bCmpOffspec or
					bMainspec == bCmpMainspec and not bOffspec and bCmpOffspec or
					not bMainspec and bCmpMainspec) then --we succeed simply by 'level'
				break;
			elseif bMainspec == bCmpMainspec and bOffspec == bCmpOffspec then --we are on the same 'level'
				if self:OrderPredicate(nScore, nCmpScore) then
					break;
				elseif nScore == nCmpScore and nRank == 1 then
					-- It is a tie
					bTied = true
				end
			end

			nRank = nRank + 1
		end

		return nRank, bTied
	end

	function RollSorter:GetDistictTop()
		local tRollers = {}

		local bMaxMainspec = false
		local bMaxOffspec = false
		local nMaxRoll = 0

		for _, tRoller in ipairs(self.tRollers) do
			local bMainspec = tRoller:IsRollingMainspec()
			local bOffspec = tRoller:IsRollingOffspec()

			local nRoll = tRoller:GetScore()
			if bMainspec and not bMaxMainspec or bOffspec and not bMaxOffspec and not bMaxMainspec then
				tRollers = {tRoller}
				bMaxMainspec = bMainspec
				bMaxOffspec = bOffspec
				nMaxRoll = nRoll
			elseif bMainspec == bMaxMainspec and bOffspec == bMaxOffspec and nRoll > nMaxRoll then
				tRollers = {tRoller}
				nMaxRoll = nRoll
			elseif bMainspec == bMaxMainspec and bOffspec == bMaxOffspec and nRoll == nMaxRoll then
				table.insert(tRollers, tRoller)
			end
		end

		return self:SelectDistinctTop(tRollers)
	end

	function RollSorter:GetTie()
		local bMainspec;
		local bOffspec;
		local nTiedScore;
		local tTieBetween = {}

		if #self.tRollers == 0 then
			return false, tTieBetween
		end

		for nIndex, tRoller in ipairs(self.tRollers) do
			if nIndex == 1 then
				bMainspec = tRoller:IsRollingMainspec()
				bOffspec = tRoller:IsRollingOffspec()
				nTiedScore = tRoller:GetScore()
			elseif (bMainspec ~= tRoller:IsRollingMainspec() or bOffspec ~= tRoller:IsRollingOffspec()
					or nTiedScore ~= tRoller:GetScore()) then
				break
			end

			table.insert(tTieBetween, tRoller)
		end

		return #tTieBetween > 1, tTieBetween
	end

	return RollSorter
end
