--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local GroupLib = require "GroupLib"
local WindowLocation = require "WindowLocation"
local GameLib = require "GameLib"
local Sound = require "Sound"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Utils
-- Contains general utilities such as logging and copy operations. It also
-- contain some convenience functions and event operations.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local Utils = {}
G.Utils = Utils
function Utils:Init(tRaidLooter)
	self.strClassName = "Utils"
	self.bInternal_IsRandomNumbersInitilized = false
	self.tRaidLooter = tRaidLooter
	self.tLinker = G.Linker:New()
	self.tLinker:Link(NLLC, self)
end

function Utils:Destroy()
	self.tLinker:Destroy()

	self.tLinker     = nil
	self.tRaidLooter = nil
end
----------------------------------------------------------------------------------
-- Debug logging
----------------------------------------------------------------------------------
function Utils:Log(var)
	if G.Debug.bIsDebug then
		Print(tostring(var))

		-- If the developer tool "rover" is installed the variable is forwaded
		-- there as well as printed to the chat.
		if SendVarToRover then
			SendVarToRover("RaidLooter", var, 0)
		end
	end
end

function Utils:LogInfo(var)
	if G.Debug.bIsDebug then
		Print(tostring(var))
	end
end

function Utils:LogBacktrace(strErr)
	if G.Debug.bIsDebug then
		Print(debug.traceback(strErr))
	end
end

function Utils:LogError(strWhere, var)
	if var ~= nil then
		Apollo.AddAddonErrorText(strWhere, "Error: " .. var)
	else
		Apollo.AddAddonErrorText(strWhere, "Unknown error")
	end

	self:LogInfo(var)
end

----------------------------------------------------------------------------------
-- Copy operations
----------------------------------------------------------------------------------
function Utils:Shallowcopy(orig)
	if orig == nil then return nil end

    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in pairs(orig) do
            copy[orig_key] = orig_value
        end
    else -- plain types.
        copy = orig
    end
    return copy
end

function Utils:Deepcopy(orig)
	if orig == nil then return nil end

    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[self:Deepcopy(orig_key)] = self:Deepcopy(orig_value)
        end
        setmetatable(copy, self:Deepcopy(getmetatable(orig)))
    else -- plain types.
        copy = orig
    end
    return copy
end

----------------------------------------------------------------------------------
-- Events
----------------------------------------------------------------------------------
function Utils:InvokeEventHandler(tInstance, strInstanceName, strEventHandler, ...)
	-- This will incur a small overhead on invoking an event
	-- But it will add debug tracing to all event handlers so we can see if, when
	-- and the order they are invoked.
	--
	-- WON'T FIX:
	---  This **should** be changed. Prefixing with "Handle" is not
	--   good as it hides the real function name. What is passed in is
	--   what we should call.
	local strFuncName      = 'Handle' .. strEventHandler
	local strQualifiedName = strInstanceName .. ':' .. strFuncName

	if G.Debug.bIsTracingEvents then
		self:Log(strQualifiedName .. ' {triggered}')
	end

	self:FireEvent(tInstance, strFuncName, ...)
end

function Utils:FireEvent(tTarget, strEventName, ...)
	if tTarget ~= nil and strEventName ~= nil then
		local fTarget = tTarget[strEventName]

		if fTarget ~= nil then
			return fTarget(tTarget, ...)
		else
			self:LogError('Utils:FireEvent', 'Target is not responding to :' .. strEventName)
		end
	end
end

function Utils:Call(tTarget, strEventName, ...)
	return self:FireEvent(...)
end

----------------------------------------------------------------------------------
-- Other
----------------------------------------------------------------------------------
function Utils:GetColoredText(strText, strColor, strFont)
	strColor = strColor or "ffffffff"
	strFont = strFont or "CRB_Interface9_BO"
	return "<T Font=\""..strFont.."\" TextColor=\""..strColor.."\">"..strText.."</T>"
end


function Utils:GetMemberInfoMatchingPredicate(fPredacte)
	local nMembers = GroupLib.GetMemberCount()
	for nIndex = 1, nMembers do
		local tUnitMember = GroupLib.GetUnitForGroupMember(nIndex)

		-- WARNING --
		--     1. This will be nil if the player is out-of-range.
		--     2. This can be non-nil if the player is on the edge i.e.
		--        about to be destroyed so we must check the validity of
		--        the unit before proceeding.
		if tUnitMember ~= nil and tUnitMember:IsValid() then
			if fPredacte(tUnitMember) then
				local tMember = GroupLib.GetGroupMember(nIndex)
				return  {
					tUnit   = tUnitMember,
					tMember = tMember,
					nIndex  = nIndex
				}
			end
		end
	end

	return nil
end

function Utils:GetScreenPosition(wndHandle)
	local x, y = wndHandle:GetPos()
	local wndParent = wndHandle:GetParent()
	if wndParent ~= nil then
		local px, py = self:GetScreenPosition(wndParent)
		return px + x, py + y
	end

	return x, y
end

function Utils:GetIntersection(tA, tB, fEquals)
    local tR = {}
	for _, valA in ipairs(tA) do
		for _, valB in ipairs(tB) do
			if fEquals(valA, valB) then
				table.insert(tR, valA)
				break
			end
		end
	end
	return tR
end

function Utils:MoveWindow(wnd, tLoc)
	if wnd == nil or tLoc == nil then
		return
	end

	local tWndLoc = WindowLocation.new(tLoc)
	local nLeft, nTop, nRight, nBottom = tWndLoc:GetOffsets()
	local nWidth = nRight - nLeft
	local nHeight = nBottom - nTop

	-- Only move the window if the width and height are the same
	if nWidth == wnd:GetWidth() and nHeight == wnd:GetHeight() then
		wnd:MoveToLocation(tWndLoc)
	end
end

function Utils:IsAssistant()
	if G.Debug.bIsTest then
		return G.Debug.bIsAssistant
	end

	local tMemberInfo = self:GetMemberInfoMatchingPredicate(function(tUnitMember)
		return tUnitMember:IsThePlayer()
	end)

	if tMemberInfo ~= nil and tMemberInfo.tMember ~= nil then
		return tMemberInfo.tMember.bRaidAssistant
	end

	return false
end

function Utils:IsLeader() return GroupLib.AmILeader() or G.Debug.bIsLeader end
function Utils:IsInRaid() return GroupLib.InRaid() or G.Debug.bIsInRaid end

function Utils:GetPlayerName()
	local tPlayerUnit = GameLib.GetPlayerUnit()

	if tPlayerUnit ~= nil then
		return tPlayerUnit:GetName()
	end

	return ""
end

function Utils:MergeLootersForDrop(tDrop)
	local tEligibleRecipients = {}

	if tDrop.tLooters ~= nil then
		for _, tUnit in pairs(tDrop.tLooters) do
			table.insert(tEligibleRecipients, {
				strUnitName = tUnit:GetName(),
				tUnit       = tUnit
			})
		end
	end

	if tDrop.tLootersOutOfRange ~= nil then
		for _, strOORLooter in pairs(tDrop.tLootersOutOfRange) do
			table.insert(tEligibleRecipients, {
				strUnitName = strOORLooter,
				tUnit       = nil
			})
		end
	end

	return tEligibleRecipients
end

function Utils:FormatSecondsToReadable(nSeconds)
	if nSeconds == 1 then
		return tostring(nSeconds).." second"
	elseif nSeconds < 60 then
		return tostring(nSeconds).." seconds"
	else
		local nMinutes = math.floor(nSeconds/60)
		nSeconds = nSeconds % 60

		local strMinutes = " minutes"
		if nMinutes == 1 then
			strMinutes = " minute"
		end

		local strSeconds = " seconds"
		if strSeconds == 1 then
			strSeconds = " second"
		end

		if strSeconds ~= 0 then
			return tostring(nMinutes) .. strMinutes .. " " .. tostring(nSeconds) .. strSeconds
		else
			return tostring(nMinutes) .. strMinutes
		end
	end
end

function Utils:FormatDataToReadableApproximateTime(nDate)
	local nNow = os.time()
	local nDiff = nNow - nDate

	if nDiff < 0 then
		return "In Future!"
		-- TODO future
	else
		if nDiff < 30 then
			return "A moment ago."
		elseif nDiff >= 30 and nDiff < 90 then
			return "About a minute ago."
		else
			local nMin = math.floor((nDiff/60)+0.5)
			return tostring(nMin) .. " minutes ago."
		end
	end
end

function Utils:IsMasterLootingOn()
	if G.Debug.bIsTest then
		return true
	end

	local tLootRules = GroupLib.GetLootRules()
	return tLootRules ~= nil and (tLootRules.eNormalRule == GroupLib.LootRule.Master
		or tLootRules.eThresholdRule == GroupLib.LootRule.Master)
end

function Utils:HaveArmorType(nClassId, strType)
	return strType == "A" or G.ClassInfo[nClassId].strArmorType == strType
end

function Utils:IsTableEmpty(t)
	return next(t) == nil
end

function Utils:GetItemQualityString(tItem)
	if tItem ~= nil then
		return G.ItemInfo[tItem:GetQuality()].strName
	end
	return ""
end

function Utils:GetItemRarity(tItem)
	if tItem ~= nil then
		return G.ItemInfo[tItem:GetQuality()].strRarityBorder
	end
	return ""
end

function Utils:GetItemExtendedRarity(tItem)
	if tItem ~= nil then
		return G.ItemInfo[tItem:GetQuality()].strExtendedRarityBorder
	end
	return ""
end

function Utils:GetClassIcon(tUnit)
	if tUnit ~= nil and tUnit:IsValid() then
		return G.ClassInfo[tUnit:GetClassId()].strIcon
	end
	return ""
end

function Utils:TableLength(tTable)
	local nCount = 0
	for _ in pairs(tTable) do nCount = nCount + 1 end
	return nCount
end

-- Compatibility: Lua-5.1
function Utils:Split(str, pat)
   local t = {}  -- NOTE: use {n = 0} in Lua-5.0
   local fpat = "(.-)" .. pat
   local last_end = 1
   local s, e, cap = str:find(fpat, 1)
   while s do
      if s ~= 1 or cap ~= "" then
	 table.insert(t,cap)
      end
      last_end = e+1
      s, e, cap = str:find(fpat, last_end)
   end
   if last_end <= #str then
      cap = str:sub(last_end)
      table.insert(t, cap)
   end
   return unpack(t)
end

-- Source: http://lua-users.org/wiki/MakingLuaLikePhp
-- Credit: http://richard.warburton.it/
function Utils:Explode(div, str)
    if (div=='') then return false end
    local pos,arr = 0,{}
    for st,sp in function() return string.find(str,div,pos,true) end do
        table.insert(arr,string.sub(str,pos,st-1))
        pos = sp + 1
    end
    table.insert(arr,string.sub(str,pos))
    return arr
end

function Utils:PrintRandomNumbers(nN)
	local tResult = ""
	for nIndex=1,nN do
		tResult = tResult .. self:Random(100)

		if nIndex ~= nN then
			tResult = tResult .. ", "
		end
	end

	self:Log("Random number: " .. tResult)
end

-- Improved version of the built-in pseudorandom generator
--     See: http://lua-users.org/wiki/MathLibraryTutorial
--
-- Wrapped up in a custom function so the global definition isn't changed.
function Utils:Random(S)
	if self.bInternal_IsRandomNumbersInitilized == nil then
		self.bInternal_IsRandomNumbersInitilized = true

		-- Set a seed and discard first 3.
		math.randomseed(os.time())
		math.random(); math.random(); math.random()
	end

	-- This will indeed call math.random 97+1 times
	-- but the result appears a lot more "random".
	if _G.randomtable == nil then
		_G.randomtable = {}
		for i = 1, 97 do
			_G.randomtable[i] = math.random()
		end
	end

	-- x acts as an index weight.
	local x = math.random()
	local i = 1 + math.floor(97*x)

	-- Perform a swap, x will become the actual random number
	-- at index i.
	x, _G.randomtable[i] = _G.randomtable[i], x

	-- Adjust the range if S is provided.
	if S ~= nil then
		local r = (x * (S - 1))
		local result = math.floor(r + 1.5) -- [1, S]
		return result
	end

	return x -- [0,1]
end

function Utils:PlaySound(nSoundFx)
	local bUseSound = self.tRaidLooter:GetSettings():Read("tGeneral.bUseSound")
	if bUseSound then
		Sound.Play(nSoundFx)
	end
end

function Utils:PlayNewLootSound()
	local bUseLootSound = self.tRaidLooter:GetSettings():Read("tGeneral.bUseLootSound")
	if bUseLootSound then
		self:PlaySound(197)
	end
end

function Utils:PlayNewRollSound()
	local bUseRollSound = self.tRaidLooter:GetSettings():Read("tGeneral.bUseRollSound")
	if bUseRollSound then
		self:PlaySound(231)
	end
end
