--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- Raid Looter
-- The MIT License (MIT)
--
-- Copyright (c) 2015 - 2016
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
-- Debugging
-- /rl dev-mode [on|off]
--
-- Events
-- Incoming events passes throw the 'Utils:InvokeEventHandler' for debug purposes.
-- Sometimes events are forwarded directly by using 'Utils:FireEvent'
--
-- EXTRA NOTE
-- Since all Wildstar addons share the same global space it is important not to;
--     * relay on functions to exist,
--     * Redefine/change or remove standard functions.
--
-- ORIGINAL AUTHOR
-- Johan Lindström (Jabbit-EU, Joxye Nadrax / Wildstar)
-- MODIFIED BY
-- Mischh X (Jabbit-EU / Wildstar)
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--

local Apollo = require "Apollo"
local XmlDoc = require "XmlDoc"

local NLLC = Apollo.GetAddon("NextLabLootClient")
NLLC.G = NLLC.G or {}
local G = NLLC.G

--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
-- XMLFactory
-- Helper for loading XML data.
--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
local XMLFactory = {}
G.XMLFactory = XMLFactory
function XMLFactory:New()
	local o = {
		tBindings = {},
		tCallbackObject = nil,
		tLinker = G.Linker:New(),
		strCallbackName = nil,
		nLoaded = 0,
		nTotal = 0,
	}
	setmetatable(o, self)
	self.__index = self
	return o
end

function XMLFactory:Init(tTarget)
	self.tLinker:Link(NLLC, self)
end

function XMLFactory:Destroy()
	self.tLinker:Destroy()

	self.tLinker         = nil
	self.tBindings       = nil
	self.tCallbackObject = nil
end

function XMLFactory:LoadAll(strEventHandler, tObject)
	G.Utils:Log("All XML files will be loaded.")

	self.tAvailable = {}
	self.strCallbackName = strEventHandler
	self.tCallbackObject = tObject
	self.nLoaded = 0
	self.nTotal = 0
	local tManagementWindowXML = self:LoadFile("RaidLooter/ManagementWindow.xml")
	local tLootWindowXML       = self:LoadFile("RaidLooter/LootWindow.xml")
	local tSharedXML           = self:LoadFile("RaidLooter/Shared.xml")

	self.tBindings = {
		ManagementWindow         = { tDocument = tManagementWindowXML, strXmlRootName = "Window",
			bIsMultiUse = true  },
		LootWindow               = { tDocument = tLootWindowXML,       strXmlRootName = "Window",
			bIsMultiUse = true  },
		Roller_Row               = { tDocument = tSharedXML,           strXmlRootName = "Roller_Row",
			bIsMultiUse = true  },
		Raider_HalfRow           = { tDocument = tSharedXML,           strXmlRootName = "Raider_HalfRow",
			bIsMultiUse = true  },
		LootItem_Square          = { tDocument = tSharedXML,           strXmlRootName = "LootItem_Square",
			bIsMultiUse = true  },
		LootItem_RollDialog      = { tDocument = tSharedXML,           strXmlRootName = "LootItem_RollDialog",
			bIsMultiUse = true  },
		LootItem_ContextMenu     = { tDocument = tSharedXML,           strXmlRootName = "LootItem_ContextMenu",
			bIsMultiUse = true  },
		LootItem_Details         = { tDocument = tSharedXML,           strXmlRootName = "LootItem_Details",
			bIsMultiUse = true  },
		Autoloot_Row             = { tDocument = tSharedXML,           strXmlRootName = "Autoloot_Row",
			bIsMultiUse = true  },
		Autoloot_NewRowPopup     = { tDocument = tSharedXML,           strXmlRootName = "Autoloot_NewRowPopup",
			bIsMultiUse = true  },
		Autoloot_ProfileDropdown = { tDocument = tSharedXML,           strXmlRootName = "Autoloot_ProfileDropdown",
			bIsMultiUse = true  },
		Autoloot_RuleDropdown    = { tDocument = tSharedXML,           strXmlRootName = "Autoloot_RuleDropdown",
			bIsMultiUse = true  },
		Autoloot_DropdownItem    = { tDocument = tSharedXML,           strXmlRootName = "Autoloot_DropdownItem",
			bIsMultiUse = true  },
		Overlay_Square           = { tDocument = tSharedXML,           strXmlRootName = "Overlay_Square",
			bIsMultiUse = true  },
	}
end

function XMLFactory:LoadFile(strFilename)
	local tXmlDoc = XmlDoc.CreateFromFile(strFilename)
	tXmlDoc:RegisterCallback("OnXMLLoaded", self)
	self.nTotal = self.nTotal + 1
	return tXmlDoc
end

function XMLFactory:Unload(strAlias)
	G.Utils:Log("XML file with alias " .. strAlias .. " unloaded.")
	self.tBindings[strAlias] = nil
end

function XMLFactory:UnloadAll()
	G.Utils:Log("All XML files will be unloaded.")
	self.tBindings = {}
	self.tCallback = nil
	self.nLoaded = 0
end

function XMLFactory:CreateWindow(strAlias, wndParent, tEventHandler)
	local tInfo = self.tBindings[strAlias]

	if tInfo ~= nil then
		if tInfo.tDocument:IsLoaded() then
			local wnd = Apollo.LoadForm(tInfo.tDocument, tInfo.strXmlRootName, wndParent, tEventHandler)

			if wnd == nil then
				G.Utils:LogError("XMLFactory:CreateWindow", "Creating the window failed - make sure all files are up-to-date.")
			end

			if not tInfo.bIsMultiUse then
				self:Unload(strAlias)
			end

			return wnd
		else
			G.Utils:LogError("XMLFactory:CreateWindow", "Unable to create the window because the XML is not loaded.")
		end
	else
		G.Utils:LogError("XMLFactory:CreateWindow", "Unable to find info for alias {" .. strAlias .. "}.")
	end

	return nil
end

function XMLFactory:OnXMLLoaded()
	self.nLoaded = self.nLoaded + 1
	G.Utils:Log("XML-file loaded (" .. self.nLoaded .. "/" .. self.nTotal .. ")")

	if self.nLoaded == self.nTotal then
		G.Utils:FireEvent(self.tCallbackObject, self.strCallbackName)
		self.nTotal = 0
	end
end
