local Apollo = require "Apollo"
local GameLib = require "GameLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")
local CHANNEL_NAME = "ChnlNextLabLoot"
local Lead = NLLC.Lead

local Ranker = {
	loaded = false,
	tRanks = {},
	tHistory = {},
	tKeys = {},
	callbacks = {},
	historyCallbacks = {},
	strLeader = Lead:GetLeader(), --at this point probably nil
	vHandler = nil,
}
NLLC.Ranker = Ranker
Apollo.LinkAddon(NLLC, Ranker)

--Base Structure:
--[[
	{
		n = strName,
		r = { --condensed and the 'accessible' part
			nRank = nRank, --base rank without any offsets.
			--offsets:
			nOffset = nOffset, --offset for offpec rolls (not including loot)
			nLootMain = nOffset, --offset for lately recieved loot.
			nLootScnd = nOffset, --offset for lately recieved loot (offspec).
		}
	}
]]

local VersionHandler = NLLC.VersionHandler
local Condenser = NLLC.Condenser.new({
	"nRank",
	"nOffset",
	"nLootMain",
	"nLootScnd",
},{
	["nRank"] = 1,
	["nOffset"] = nil,
	["nLootMain"] = 0,
	["nLootScnd"] = 0,
})

-----------------------
-- Acessible Methods --
-----------------------

function Ranker:RegisterOnRanksChanged(func, handler, ...)
	local callback = self:CreateCallback(func, handler, ...)
	table.insert(self.callbacks, callback)

	if self.loaded then
		local strLead = Lead:GetLeader()
		local strVersion = (strLead or GameLib.GetPlayerCharacterName())..":---"
		if self.vHandler:HasOwner(strLead) then
			strVersion = self.vHandler:GetVersion(strLead)
		end
		callback(self.tRanks, strVersion)
	end
end

function Ranker:RegisterOnHistoryChanged(func, handler, ...)
	local callback = self:CreateCallback(func, handler, ...)
	table.insert(self.historyCallbacks, callback)
end

function Ranker:GetRanks(strName)
	if not self.loaded then error("Ranker not initialized yet") end

	if strName then
		return self.tRanks[strName]
	end

	local strLead = Lead:GetLeader()
	local strVersion = (strLead or GameLib.GetPlayerCharacterName())..":---"
	if self.vHandler:HasOwner(strLead) then
		strVersion = self.vHandler:GetVersion(strLead)
	end

	return self.tRanks, strVersion
end

function Ranker:GetHistory(strName)
	return self.tHistory[strName] or {}
end

function Ranker:AddHistory(strName, strHistory)
	if not strHistory then return end
	if not self.tHistory[strName] then
		self.tHistory[strName] = {}
	end
	table.insert(self.tHistory[strName], 1, strHistory)

	for _, callback in ipairs(self.historyCallbacks) do
		callback(strName, self.tHistory[strName])
	end
end

function Ranker:ApplyLootOffset(strName, bMain, nOffset)
	if not self.loaded then error("Ranker not initialized yet") end
	--assuming self.tRanks is in a valid state:
	local ranks = self.tRanks[strName]
	if not ranks then return false end -- this guy is not part of our rankings, ignore this request

	local key = bMain and 'nLootMain' or 'nLootScnd'
	local strHistory = bMain and 'Loot-M [%s] to [%s]' or 'Loot-O [%s] to [%s]'

	if ranks[key] <= nOffset then
		self:AddHistory(strName, strHistory:format(ranks[key], ranks[key]))
		return true
	end --he has a worse (or equal) offset already

	local nLootMain = bMain and nOffset or ranks.nLootMain
	local nLootScnd = bMain and ranks.nLootScnd or nOffset

	if bMain then
		self:AddHistory(strName, strHistory:format(ranks.nLootMain, nLootMain))
	else
		self:AddHistory(strName, strHistory:format(ranks.nLootScnd, nLootScnd))
	end
	self:SetRank(strName, ranks.nRank, ranks.nOffset, nLootMain, nLootScnd)
	return true
end

--returns nil/false/nRank on noData/noOffspec/nActualRank
function Ranker:GetLootRank(strName, bMain)
	if not self.loaded then error("Ranker not initialized yet") end

	--assuming self.tRanks is in a valid state:
	local ranks = self.tRanks[strName]
	if not ranks then return nil end

	if bMain then
		return ranks.nRank + ranks.nLootMain
	elseif ranks.nOffset then
		return ranks.nRank + ranks.nLootScnd + ranks.nOffset
	else
		return false
	end
end

function Ranker:GetBaseRank(strName)
	if not self.loaded then error("Ranker not initialized yet") end

	--assuming self.tRanks is in a valid state:
	local ranks = self.tRanks[strName]
	if not ranks then return nil end

	return ranks.nRank
end

function Ranker:SetRank(strName, nRank, nOffset, nLootMain, nLootScnd)
	if not self.loaded then error("Ranker not initialized yet") end

	local replace = {
		n = strName,
		r = Condenser:To({
			nRank = nRank,
			nOffset = nOffset,
			nLootMain = nLootMain,
			nLootScnd = nLootScnd,
		}),
	}

	local oldRanks = self:GetRanks(strName)
	if oldRanks then
		local strHistory = '%s [%s] to [%s]'
		if nRank ~= oldRanks.nRank then
			self:AddHistory(strName, strHistory:format("Rank", oldRanks.nRank, nRank))
		end
		if nOffset ~= oldRanks.nOffset then
			self:AddHistory(strName, strHistory:format("Mod", oldRanks.nOffset, nOffset))
		end
		if nLootMain ~= oldRanks.nLootMain then
			self:AddHistory(strName, strHistory:format("Loot-M", oldRanks.nLootMain, nLootMain))
		end
		if nLootScnd ~= oldRanks.nLootScnd then
			self:AddHistory(strName, strHistory:format("Loot-O", oldRanks.nLootScnd, nLootScnd))
		end
	end

	for key, data in self.vHandler:Iterate() do
		if data.n == strName then
			if not Condenser:IsEqual(data.r, replace) then
					self.vHandler:Save(key, replace)
			end
			return true
		end
	end

	return self.vHandler:Save(nil, replace)
end

function Ranker:RemoveRank(strName)
	if not self.vHandler then error("Ranker not initialized yet") end

	for key, data in self.vHandler:Iterate() do
		if data.n == strName then
			return self.vHandler:Remove(key) --return bSuccess
		end
	end

	return false
end

function Ranker:SetAllRanks(tRanks)
	if not self.vHandler then error("Ranker not initialized yet") end
	self.vHandler:BigChange(function(VH) --self.vHandler == VH
		local tCondensed = {}

		for name, rankTbl in pairs(tRanks) do
			tCondensed[name] = {n = name, r = Condenser:To(rankTbl)}
		end

		--first: remove all current settings rank (if they are not coherent with the new ranks)
		for key, data in VH:Iterate() do
			if not tCondensed[data.n] or not Condenser:IsEqual(tCondensed[data.n].r, data.r) then
				VH:Remove(key)
			else
				tCondensed[data.n] = nil
			end
		end

		--then: add all those, which were wrong, or missing
		for name, condensed in pairs(tCondensed) do
			VH:Save(nil, condensed)
		end

	end)
end

-----------------------
-- Internal Stuff --
-----------------------

function Ranker:CreateCallback(func, handler, ...)
	local t = {...}
	if type(func) == "function" then
		return function(arg1,arg2) return func(arg1,arg2, unpack(t)) end
	elseif type(handler) == "table" then
		return function(arg1,arg2) return handler[func](handler, arg1,arg2, unpack(t)) end
	else
		error("Could not create callback, need either a function, or a key-table pair.")
	end
end

function Ranker:OnSave(eType, tSavedData)
	if eType == GameLib.CodeEnumAddonSaveLevel.General then
		tSavedData.Ranker = self.vHandler and self.vHandler:GetData()
	end
end
NLLC:OnceSaving("OnSave", Ranker)

function Ranker:OnRestore(eType, tSavedData)
	if eType == GameLib.CodeEnumAddonSaveLevel.General then
		local data = tSavedData and tSavedData.Ranker or nil
		self.vHandler = VersionHandler.new(CHANNEL_NAME, data, "R")
		self.vHandler:OnAdded("VHRanksAdded", self)
		self.vHandler:OnRemoved("VHRanksRemoved", self)
		self.vHandler:OnChanged("VHRanksChanged", self)
		self.vHandler:SetActive(true)

		Lead:RegisterOnLeadChanged("LeaderChanged", self)
	end
end
NLLC:OnceRestored("OnRestore", Ranker)

function Ranker:PerformCallbacks()
	local tRanks, strVersion = self:GetRanks()

	for _, callback in ipairs(self.callbacks) do
		callback(tRanks, strVersion)
	end
end

function Ranker:LeaderChanged(strLead, bIsSelf)
	self.loaded = true
	self.tRanks = {}
	self.tKeys = {}
	if self.vHandler:HasOwner(strLead) then
		for key, data in self.vHandler:Iterate(strLead) do
			self.tRanks[data.n] = Condenser:From(data.r)
			self.tKeys[data.n] = key
		end
	end
	self:PerformCallbacks()

	self.vHandler:SetSharing(bIsSelf)
end

function Ranker:VHRanksAdded(prefix, type, owner, key, data)
	local strLeader, bSelf = Lead:GetLeader()
	if owner ~= strLeader and not (bSelf and self.vHandler:OwnerIsSelf(owner)) then return end

	self.tRanks[data.n] = Condenser:From(data.r)
	self.tKeys[data.n] = key

	self:PerformCallbacks()
end

function Ranker:VHRanksRemoved(prefix, type, owner, key, data)
	local strLeader, bSelf = Lead:GetLeader()
	if owner ~= strLeader and not (bSelf and self.vHandler:OwnerIsSelf(owner)) then return end

	if self.tKeys[data.n] ~= key then return end

	self.tKeys[data.n] = nil
	self.tRanks[data.n] = nil

	self:PerformCallbacks()
end

function Ranker:VHRanksChanged(prefix, type, owner, key, data)
	local strLeader, bSelf = Lead:GetLeader()
	if owner ~= strLeader and not (bSelf and self.vHandler:OwnerIsSelf(owner)) then return end

	if self.tKeys[data.n] ~= key then return end

	self.tRanks[data.n] = Condenser:From(data.r)

	self:PerformCallbacks()
end
