local Apollo = require "Apollo"
local GameLib = require "GameLib"

local NLLC = Apollo.GetAddon("NextLabLootClient")

local RankUI = {
	wndMain = nil,
	nSortColumn = 2,
}
NLLC.RankUI = RankUI
Apollo.LinkAddon(NLLC, RankUI)

local Ranker = NLLC.Ranker
local Lead = NLLC.Lead
local Colors = NLLC.Colors
local Options = NLLC.Options
local UIComponents = NLLC.UIComponents

local tRankColors = Colors:RegisterOnColorsChanged(Colors.EnumColorType.Rank, "ColorsChanged", RankUI)
local tNameColors = Colors:RegisterOnColorsChanged(Colors.EnumColorType.UnitName, "ColorsChanged", RankUI)

local IMPORT_PATTERN = "(%a[^\t]*)\t([+-]?%d+)\t([+-]?%d*)\t([+-]?%d+)\t([+-]?%d+)"
function RankUI:ImportClipboard()
	self.wndClipboard:SetText("")
	self.wndClipboard:PasteTextFromClipboard()
	local text = self.wndClipboard:GetText()

	local tRanks = {}

	for name, rank, offMod, lootMain, lootOff in text:gmatch(IMPORT_PATTERN) do
		tRanks[name] = {
			nRank = tonumber(rank),
			nOffset = tonumber(offMod),
			nLootMain = tonumber(lootMain),
			nLootScnd = tonumber(lootOff),
		}
	end

	Ranker:SetAllRanks(tRanks)
end

function RankUI:ColorsChanged(newColors, eColorType)
	if eColorType == Colors.EnumColorType.Rank then
		tRankColors = newColors
	elseif eColorType == Colors.EnumColorType.UnitName then
		tNameColors = newColors
	end

	if self.wndMain and self.wndMain:IsShown() then
		self:RanksChanged(Ranker:GetRanks())
	end
end

function RankUI:RanksChanged(tRanks, strVersion)
	if not self.wndMain or not self.wndMain:IsShown() then return end

	local wndGrid = self.wndGrid
	wndGrid:DeleteAll()
	for name, ranks in pairs(tRanks) do
		local base = ranks.nRank
		local main = base + ranks.nLootMain

		local scnd = ranks.nOffset and base + ranks.nOffset + ranks.nLootScnd or "-"

		local iCurrRow = wndGrid:AddRow("")
		wndGrid:SetCellLuaData(iCurrRow, 1, {n=name, r=ranks})
		wndGrid:SetCellDoc(iCurrRow, 1, self:HelperColorizeText(name, tNameColors[name]))
		wndGrid:SetCellSortText(iCurrRow, 1, name)
		wndGrid:SetCellDoc(iCurrRow, 2, self:HelperColorizeText(base, tRankColors[base]))
		wndGrid:SetCellSortText(iCurrRow, 2, base)
		wndGrid:SetCellDoc(iCurrRow, 3, self:HelperColorizeText(main, tRankColors[main]))
		wndGrid:SetCellSortText(iCurrRow, 3, main)
		wndGrid:SetCellDoc(iCurrRow, 4, self:HelperColorizeText(scnd, tRankColors[scnd]))
		wndGrid:SetCellSortText(iCurrRow, 4, scnd)
	end

	self.wndSource:SetText(strVersion:gsub(":", "\n", 1)) --replace the first ':' with '\n'

	local bSelf = select(2,Lead:GetLeader())
	self.wndClipboardBtn:Show(bSelf, true) --only show, if the player himself is the leader
	self.wndNewPlayerFrameBtn:Show(bSelf, true) --only show, if the player himself is the leader

	self:RedoSorting()
	self:RedoLootButtons()

	--the grid repositions itself everytime you add a row - kinda annoying...
	local h,w = wndGrid:GetHeight(), wndGrid:GetWidth()
	local l,t = wndGrid:GetOriginalLocation():GetOffsets()
	wndGrid:SetAnchorOffsets(l,t,l+w,t+h)

	wndGrid:GetParent():RecalculateContentExtents()
end

function RankUI:HistoryChanged(...) --we call this sometimes without agruements.
	local iCurrRow = self.wndGrid:GetCurrentRow()
	local _, bSelf = Lead:GetLeader()
	if iCurrRow and bSelf then
		local tData = self.wndGrid:GetCellData(iCurrRow, 1)
		local strName = tData.n
		local tHistory = Ranker:GetHistory(strName)

		local tListItems = self.wndHistoryList:GetChildren()

		for i, strHistory in ipairs(tHistory) do
			local tItem = tListItems[i] or NLLC:LoadForm("DropdownItem", self.wndHistoryList, nil)
			tItem:SetText(strHistory)
		end

		for i = #tHistory+1, #tListItems, 1 do
			tListItems[i]:Destroy()
		end

		self.wndHistoryList:ArrangeChildrenVert()
		self.wndHistoryList:RecalculateContentExtents()
		self.wndHistoryList:SetVScrollPos(0)
	end
end

function RankUI:HelperColorizeText(strText, strColor)
	return "<T Font=\"CRB_Interface9_BO\" TextColor=\""..(strColor or "ffffffff").."\">"..strText.."</T>"
end

function RankUI:SortButtonChecked(wndHandler, wndControl)
	if wndHandler ~= wndControl then return end
	self.nSortColumn = wndHandler:GetData()
	self:RedoSorting()
end

function RankUI:SortButtonUnchecked(wndHandler, wndControl)
	if wndHandler ~= wndControl then return end
	self:RedoSorting()
end

function RankUI:OnCloseBtn(wndHandler, wndControl)
	if wndHandler ~= wndControl or not self.wndMain then return end
	self.wndMain:Show(false)
end

function RankUI:RedoSorting()
	if not self.wndMain then return end

	self.tSortButtons[1]:SetCheck(self.nSortColumn == 1)
	self.tSortButtons[2]:SetCheck(self.nSortColumn == 2)
	self.tSortButtons[3]:SetCheck(self.nSortColumn == 3)
	self.tSortButtons[4]:SetCheck(self.nSortColumn == 4)
	self.wndGrid:SetSortColumn(self.nSortColumn)
end

function RankUI:MainLootButtonChecked(wndHandler, wndControl)
	if wndHandler ~= wndControl or not self.wndMain then return end
	local nTarget = wndHandler:GetData() --0, -1 or -2

	local iCurrRow = self.wndGrid:GetCurrentRow()
	if iCurrRow then
		local tData = self.wndGrid:GetCellData(iCurrRow, 1)
		Ranker:SetRank(tData.n, tData.r.nRank, tData.r.nOffset, nTarget, tData.r.nLootScnd)
	end
	self:RedoLootButtons()
end
function RankUI:MainLootButtonUnchecked(wndHandler, wndControl)
	if wndHandler ~= wndControl or not self.wndMain then return end
	self:RedoLootButtons()
end
function RankUI:ScndLootButtonChecked(wndHandler, wndControl)
	if wndHandler ~= wndControl or not self.wndMain then return end
	local nTarget = wndHandler:GetData() --0, -1 or -2

	local iCurrRow = self.wndGrid:GetCurrentRow()
	if iCurrRow then
		local tData = self.wndGrid:GetCellData(iCurrRow, 1)
		Ranker:SetRank(tData.n, tData.r.nRank, tData.r.nOffset, tData.r.nLootMain, nTarget)
	end
	self:RedoLootButtons()
end
function RankUI:ScndLootButtonUnchecked(wndHandler, wndControl)
	if wndHandler ~= wndControl or not self.wndMain then return end
	self:RedoLootButtons()
end

function RankUI:GridSelectionChanged()
	self:RedoLootButtons()
end

function RankUI:RedoLootButtons()
	if not self.wndGrid then return end

	local iCurrRow = self.wndGrid:GetCurrentRow()
	local _, bSelf = Lead:GetLeader()
	if iCurrRow and bSelf then
		local tData = self.wndGrid:GetCellData(iCurrRow, 1)
		self.wndLootTitle:SetText(tData.n)

		local main = tData.r.nLootMain
		local scnd = tData.r.nLootScnd

		self.tMainLootButtons[0]:SetCheck(main == 0)
		self.tMainLootButtons[-1]:SetCheck(main == -1)
		self.tMainLootButtons[-2]:SetCheck(main == -2)
		self.tScndLootButtons[0]:SetCheck(scnd == 0)
		self.tScndLootButtons[-1]:SetCheck(scnd == -1)
		self.tScndLootButtons[-2]:SetCheck(scnd == -2)
	end

	self.wndLootOffsets:Show(iCurrRow~=nil and bSelf, true)
	self:HistoryChanged()
end

function RankUI:OnHideClipboardButton()
	self.wndClipboardBtn:SetCheck(false)
end

local EXPORT_PATTERN = "%s\t%d\t%s\t%d\t%d"
function RankUI:OnEnteredExportButton()
	local tExportRows = {}
	for name, ranks in pairs((Ranker:GetRanks())) do
		local nOffset = ranks.nOffset or "-"
		table.insert(tExportRows, EXPORT_PATTERN:format(name, ranks.nRank, nOffset, ranks.nLootMain, ranks.nLootScnd))
	end

	local strExport = table.concat(tExportRows, "\n")
	if strExport == "" then strExport = " " end
	self.wndExport:SetActionData(GameLib.CodeEnumConfirmButtonType.CopyToClipboard, strExport)
end

function RankUI:Show()
	if self.wndMain then
		self.wndMain:Show(true)
		self:RanksChanged(Ranker:GetRanks())
		self:HistoryChanged()
		return
	end
	self.wndMain = NLLC:LoadForm("RankList", nil, self)
	self.wndClipboard = self.wndMain:FindChild("Clipboard")
	self.wndGrid = self.wndMain:FindChild("TabList:Grid")
	self.wndSource = self.wndMain:FindChild("LabelSource")
	self.wndImport = self.wndMain:FindChild("ClipboardWindow:ImportButton")
	self.wndExport = self.wndMain:FindChild("ClipboardWindow:ExportButton")
	self.wndLootOffsets = self.wndMain:FindChild("LootOffsets")
	self.wndLootTitle = self.wndLootOffsets:FindChild("Title")
	self.wndClipboardBtn = self.wndMain:FindChild("ClipboardButton")
	self.wndNewPlayerFrameBtn = self.wndMain:FindChild("NewPlayerButton")
	self.wndHistoryList = self.wndLootOffsets:FindChild("HistoryList")

	self.wndClipboardBtn:AttachWindow(self.wndMain:FindChild("ClipboardWindow"))
	self.wndNewPlayerFrameBtn:AttachWindow(self.wndMain:FindChild("NewPlayerWindow"))
	self.wndExport:SetActionData(GameLib.CodeEnumConfirmButtonType.CopyToClipboard, " ")

	self.tMainLootButtons = {}
	self.tScndLootButtons = {}
	self.tMainLootButtons[0]  = self.wndLootOffsets:FindChild("SetLootMainX"); self.tMainLootButtons[0]:SetData(0)
	self.tMainLootButtons[-1] = self.wndLootOffsets:FindChild("SetLootMain5"); self.tMainLootButtons[-1]:SetData(-1)
	self.tMainLootButtons[-2] = self.wndLootOffsets:FindChild("SetLootMain");  self.tMainLootButtons[-2]:SetData(-2)
	self.tScndLootButtons[0]  = self.wndLootOffsets:FindChild("SetLootScndX"); self.tScndLootButtons[0]:SetData(0)
	self.tScndLootButtons[-1] = self.wndLootOffsets:FindChild("SetLootScnd5"); self.tScndLootButtons[-1]:SetData(-1)
	self.tScndLootButtons[-2] = self.wndLootOffsets:FindChild("SetLootScnd");  self.tScndLootButtons[-2]:SetData(-2)

	self.tSortButtons = {}
	self.tSortButtons[1] = self.wndMain:FindChild("Headers:ButtonSortName"); self.tSortButtons[1]:SetData(1)
	self.tSortButtons[2] = self.wndMain:FindChild("Headers:ButtonSortRank"); self.tSortButtons[2]:SetData(2)
	self.tSortButtons[3] = self.wndMain:FindChild("Headers:ButtonSortMain"); self.tSortButtons[3]:SetData(3)
	self.tSortButtons[4] = self.wndMain:FindChild("Headers:ButtonSortScnd"); self.tSortButtons[4]:SetData(4)


	local wndNewPlayer = self.wndMain:FindChild("NewPlayerWindow")
	local newPlayerOpt = Options:GetOption("RankingUI", "NewPlayer")
	local nameOpt  = Options:GetOption(newPlayerOpt, "Name" )
	local rankOpt  = Options:GetOption(newPlayerOpt, "Rank" )
	local modOpt   = Options:GetOption(newPlayerOpt, "Mod"  )
	local lootMOpt = Options:GetOption(newPlayerOpt, "LootM")
	local lootOOpt = Options:GetOption(newPlayerOpt, "LootO")
	UIComponents:applyTextbox(wndNewPlayer:FindChild("Window_Name"),  nameOpt,  "ffffffff"); nameOpt:Set("")
	UIComponents:applyTextbox(wndNewPlayer:FindChild("Window_Rank"),  rankOpt,  "ffffffff"); rankOpt:Set("1")
	UIComponents:applyTextbox(wndNewPlayer:FindChild("Window_Mod"),   modOpt,   "ffffffff"); modOpt:Set("-")
	UIComponents:applyTextbox(wndNewPlayer:FindChild("Window_LootM"), lootMOpt, "ffffffff"); lootMOpt:Set("0")
	UIComponents:applyTextbox(wndNewPlayer:FindChild("Window_LootO"), lootOOpt, "ffffffff"); lootOOpt:Set("0")

	Ranker:RegisterOnRanksChanged("RanksChanged", self) --this also triggers a :RanksChanged immediatly.
	Ranker:RegisterOnHistoryChanged("HistoryChanged", self)
	self:HistoryChanged()
end

function RankUI:OnSave(eType, tSavedData)
	if eType == GameLib.CodeEnumAddonSaveLevel.General then
		tSavedData.RankingUI = {
			nSortColumn = self.nSortColumn
		}
	end
end
NLLC:OnceSaving("OnSave", RankUI)

function RankUI:OnRestore(eType, tSavedData)
	if eType == GameLib.CodeEnumAddonSaveLevel.General then
		if tSavedData and tSavedData.RankingUI then
			self.nSortColumn = tSavedData.RankingUI.nSortColumn or self.nSortColumn
			self:RedoSorting()
		end
	end
end
NLLC:OnceRestored("OnRestore", RankUI)

NLLC:OnceDocLoaded(function()
	Apollo.RegisterSlashCommand("ranks", "OnSlashCommand", RankUI)
end)

function RankUI:OnSlashCommand()
	RankUI:Show()
end

-------------------------
--- New Player Window ---
-------------------------
do
	local newPlayerOpt = Options:GetOption("RankingUI", "NewPlayer")
	local nameOpt  = Options:GetOption(newPlayerOpt, "Name" )
	local rankOpt  = Options:GetOption(newPlayerOpt, "Rank" )
	local modOpt   = Options:GetOption(newPlayerOpt, "Mod"  )
	local lootMOpt = Options:GetOption(newPlayerOpt, "LootM")
	local lootOOpt = Options:GetOption(newPlayerOpt, "LootO")

	--apply 'Only Numbers!' to options:
	for _, opt in ipairs({rankOpt, lootMOpt, lootOOpt}) do
		opt:OnUpdate(function(var)
			if not tonumber(var) then
				opt:Set("0")
			end
		end)
	end
	--appy 'Only Number / "-"' to modOpt
	modOpt:OnUpdate(function(var)
		if var ~= "-" and not tonumber(var) then
			modOpt:Set("0")
		end
	end)

	function RankUI:OnNewPlayerSelectTargetName()
		local unit = GameLib.GetTargetUnit()
		if not unit then return end
		local name = unit:GetName()
		if not name then return end
		nameOpt:Set(name)
	end

	function RankUI:OnNewPlayerAddPlayer()
		local name = nameOpt:Get()
		if not name or name:len() < 3 then name = "Default Name" end
		Ranker:SetRank(name,
			tonumber(rankOpt:Get()) or 1,
			tonumber(modOpt:Get()) or nil,
			tonumber(lootMOpt:Get()) or 0,
			tonumber(lootOOpt:Get()) or 0)
	end

	function RankUI:OnHideNewPlayerButton()
		self.wndNewPlayerFrameBtn:SetCheck(false)
	end
end
