local Apollo = require "Apollo"
local ApolloTimer = require "ApolloTimer"
local GameLib = require "GameLib"
local ICCommLib = require "ICCommLib"
require "ICComm"

local NLLC = Apollo.GetAddon("NextLabLootClient")


local VersionHandler = {arHandlers = {}}
NLLC.VersionHandler = VersionHandler
VersionHandler.__index = VersionHandler

local Coder = NLLC.Coder --see Coder.lua --encoding/decoding
local Log = NLLC.Log --see Log.lua --logging via VersionHandler

local Joiner = {} --used to join ICComm Channels
VersionHandler.Joiner = Joiner

local ChannelManager = { log = {[0] = 1, [1] = {}}} --handles I/O of ICComm Messages
VersionHandler.ChannelManager = ChannelManager

Apollo.LinkAddon(NLLC, VersionHandler)
Apollo.LinkAddon(NLLC, Joiner)
Apollo.LinkAddon(NLLC, ChannelManager)

local typeTbl = {} --gets replaced!

--[[
	VersionHandler has multiple Methods/Functions:

	VersionHandler.new(chName, initData, prefix) - returns VersionHandler
		- Creates a new VersionHandler.
		- All Methods below should be called from such.

	VersionHandler:Create(type [, data]) - returns new_key, data
		- Creates a new version key to put data into

	VersionHandler:Remove(key)
		- Removes the data at this key.

	VersionHandler:Save(key [, data]) - returns new_key
		- Marks this Version as finished, moving it to another Versionnumber.
		- Dont do this too often!

	VersionHandler:Changed(key)
		- announces a Change of this key. Sharing changes through ICComm
		- this is built to be called quite often on the same key.
		- this will do nothing, if nothing changed.

	VersionHandler:BigChange(func)
		- prevents the VersionHandler from annoucing every little step inside the passed function, but
		  rather annouces them in a big batch in the end of the functions excecution

	VersionHandler:GetOwners() - returns table[i] = owner

	VersionHandler:HasOwner([owner]) - returns true/false whether this owner exists in data

	VersionHandler:OwnerIsSelf(owner) - returns true/false whether this owner is the  player himself

	VersionHandler:Iterate([sOwner]) - returns iterator
		- example: 'for key, data in VH:Iterate("My Name") do ... end'
		- Those data-structures, which have a temporary replacement-key, will only be returned this one.

	VersionHandler:SetSharing( bSharing [, sOwner])
		- Sets this VersioHandler up for sharing one of its versions.
			(or disables it)
		- Default: Share your own versions
		- sOwner: Share the versions of this owner

	VersionHandler:SetActive( bActive )
		- An active VersionHandler will: (An inactive one wont:)
			- Activly share changes (rather than only on request)
			- Activly ask for versions of this prefix

	VersionHandler:SetLogging(bLog)
		- Enables/Disables logging for this VersionHandler. By default, every VersionHandler is logging. via Log:Add(...)

	VersionHandler:OnAdded(func [, tHandler, ...])
		- Registers a function (replaces the previous) to be called, whenever a version was added
		- 'func' is either a function, or a key inside tHandler, to find a Method at. (functions as keys do not work)
		- this also gets called whenever the data gets changed by :Create() and :Save() (not only from outside changes)
		- callback parameters: tHandler:Func(prefix, type, owner, key, data, ...)

	VersionHandler:OnRemoved(func [, tHandler, ...])
		- Registers a function (replaces the previous) to be called, whenever a version was removed
		- 'func' is either a function, or a key inside tHandler, to find a Method at. (functions as keys do not work)
		- this also gets called whenever the data gets changed by :Save() (not only from outside changes)
		- callback parameters: tHandler:Func(prefix, type, owner, key, data, ...)

	VersionHandler:OnChanged(func [, tHandler, ...])
		- Registers a function (replaces the previous) to be called, whenever a version was changed,
			this is not called on removal or adding of a version.
		- 'func' is either a function, or a key inside tHandler, to find a Method at. (functions as keys do not work)
		- this DOES NOT get called whenever OnRemoved or OnAdded gets called.
		- callback parameters: tHandler:Func(prefix, type, owner, key, data, ...)

	VersionHandler:OnNewVersion(func [, tHandler, ...])
		- Registers a function (replaces the previous) to be called, whenever a new version was recieved from another player
		- 'func' is either a function, or a key inside tHandler, to find a Method at. (functions as keys do not work)
		- this DOES NOT get called for your own versions!
		- callback parameters: tHandler:Func(prefix, owner, tVersions, ...)
		- tVersions = {major = number, minor = number, tempMajor = number/nil, tempMajor = number/nil}

	VersionHandler:OnActive(func [, tHandler, ...])
		- callback parameters: tHandler:Func(active, ...)

	VersionHandler:Get(key [, sOwner]) returns data
		- Returns the data of this VersionHandler for this key and owner.
		- if no owner was given, return your own.

	VersionHandler:GetData() - returns data
		- used to initialize a VersionHandler with this VersionHandlers versions.

	VersionHandler:SetTypeTable(tbl)
		- used inside Finisher.lua - sets the table to access Methods of the TypeHandlers (aka. Modules)

]]


--[[
 The Versions look like this: (all examples are tied together)
	data = {
		["*"] = "Mischh X"			--this is the one WE own!
		["_"] = "D105",					--the assigned prefix; in our case mostly used for zone-associations
		["Mischh X"] = {
			["*"] = "Mischh X",		--owner of this version!
			[true] = 7,						--the current numerical version - the latest added version-slot (even if removed)
			[false] = 4,					--sum of all 'removed' version-slots
			[1] = {["*"] = "Indentificator for this type of shared Data", ...},
			[2] = false,					-- a removed, or :Save() Data-Tbl. (this is a 'removed' version-slot)
			[3] = false,
			[4] = false,
			[5] = {["*"] = "P", ...}, 	--try to use short keys and values! (preferrably boolean) for short encoding
			[6] = {["*"] = "C", ...},
			[7] = false,
		}
	}
There are some ways the Addon shares the Structure of versions:
	The version itself: "V:D105:Mischh X:7:4"
		(Both numbers will be encoded in some way - this is the shortest way of telling your version - its usually announced)
		-> this is the only thing needed to verify your addon is up to date!
		-> from this the addon knows, to ask for the structure, if outdated;
		-> It could also request EVERYTHING if never heard of this owner
	The structure: "S:D105:Mischh X:1D2B"
		(translation: one data; three empty; two data; one empty;
		as you see, data is 'encoded' in decimal numbers, empty slots in B64)
		-> from this the addon knows the to-be-removed slots, aswell as the to-be-requested slots
Doing :Save() on [5] would do: (only mentioning the changing keys!)
	[true] = 8,
	[false] = 5,
	[5] = false,
	[8] = {["*"] = "P", ...},
]]--

local MSG_RCV_PATTERN = "(.):(.-):(.+)" -- MessageType : Prefix : MESSAGE
local MSG_RCV_PATTERN_SHORT = "(.):(.+)" -- MessageType : Prefix
local MSG_SND_PATTERN = "%s:%s:%s"
local MSG_SND_PATTERN_SHORT = "%s:%s"

local TYPE_REQ_VERSION = "v" -- small-case-letters are requests, others are announcements
local TYPE_SND_VERSION = "V"
local TYPE_REQ_STRUCTURE = "s"
local TYPE_SND_STRUCTURE = "S"
local TYPE_REQ_STRUCTURE_TEMP = "t"
local TYPE_SND_STRUCTURE_TEMP = "T"
local TYPE_REQ_DATA = "d"
local TYPE_SND_DATA = "D"
local TYPE_REQ_DATA_TEMP = "i" --information! ha! i found another word for data! (not really, but ...)
local TYPE_SND_DATA_TEMP = "I"
local TYPE_SND_CHANGE = "C"
local TYPE_SND_CHANGE_TEMP = "U" --Its like a C, but a little different! (or see it as 'Update')

local tMessageTypes = {
	-- example: "V:D105:Mischh X:7:4" should be parsed & routed to VersionHandler:OnVersion(...)
	[TYPE_REQ_VERSION] = "OnRequestVersion",				-- "v:D105:Mischh X" 	or 	"v:D105"
	[TYPE_SND_VERSION] = "OnVersion",						-- "V:D105:Mischh X:7:4:3:2" (with temp-versions)
	[TYPE_REQ_STRUCTURE] = "OnRequestStructure",			-- "s:D105:Mischh X" 	or 	"s:D105"
	[TYPE_SND_STRUCTURE] = "OnStructure",					-- "S:D105:Mischh X:1D2B
	[TYPE_REQ_STRUCTURE_TEMP] = "OnRequestTempStructure",	-- "t:D105:Mischh X" 	or 	"t:D105"
	[TYPE_SND_STRUCTURE_TEMP] = "OnTempStructure",			-- "T:D105:Mischh X:C3"
	[TYPE_REQ_DATA] = "OnRequestData",						-- "d:D105:Mischh X:B:F>Z:z" (this requests multiple versions)
	[TYPE_SND_DATA] = "OnData",								-- "D:D105:Mischh X:B:{"t""Hello World!""h"iy|"w"iBk|"*""P"}"
	[TYPE_REQ_DATA_TEMP] = "OnRequestTempData",				-- "i:D105:Mischh X:C>E"
	[TYPE_SND_DATA_TEMP] = "OnTempData",					-- "I:D105:Mischh X:C:{"t""TEMPORARY""h"iy|"w"iBk|"*""P"}"
	[TYPE_SND_CHANGE] = "OnChange",							-- "C:D105:Mischh X:B:{"t""Bye World!"}:[]"
	[TYPE_SND_CHANGE_TEMP] = "OnTempChange",				-- "U:D105:Mischh X:C:{"t""_T_"}:[]"
}

local function TEMP_FUNCTION_EMPTY() end

local function concat(a,b,...) --returns "a:b:..."
	local s = tostring(a)..":"..tostring(b)

	if ... then
		s = concat(s, ...)
	end

	return s
end

local function seperate(str, num) --returns a, b, ... (max num-amounts of returns)
	local a, tail = str:match("(..-):(.+)")

	if not a or num<2 then
		return str
	end

	if num > 2 then
		return a, seperate(tail, num-1)
	else
		return a, tail
	end
end

local function callback(func, handler, prefix, ...)
	local t = {...}
	if type(func) == "function" then
		return function(owner, key, data) return func(prefix, data["*"], owner, key, data, unpack(t)) end
	elseif type(handler) == "table" then
		return function(owner, key, data) return handler[func](handler, prefix, data["*"], owner, key, data, unpack(t)) end
	else
		error("Could not create callback, need either a function, or a key-table pair.")
	end
end

local function callback_twoarg(func, handler, prefix, ...)
	local t = {...}
	if type(func) == "function" then
		return function(owner, tVer) return func(prefix, owner, tVer, unpack(t)) end
	elseif type(handler) == "table" then
		return function(owner, tVer) return handler[func](handler, prefix, owner, tVer, unpack(t)) end
	else
		error("Could not create callback, need either a function, or a key-table pair.")
	end
end

local function callback_onearg(func, handler, ...)
	local t = {...}
	if type(func) == "function" then
		return function(arg) return func(arg, unpack(t)) end
	elseif type(handler) == "table" then
		return function(arg) return handler[func](handler, arg, unpack(t)) end
	else
		error("Could not create callback, need either a function, or a key-table pair.")
	end
end

--[[			====== Pubic Functions ======			]]
-- .new, Create, Save, Changed, GetOwners, Iterate, SetSharing, SetActive, GetData

function VersionHandler.new(chName, initData, prefix)
	local data, temp, missData, missTemp = VersionHandler:CreateData(prefix, initData)

	local o = setmetatable({
		prefix = prefix,
		data = data,
		temp = temp,
		missData = missData,
		missTemp = missTemp,
		announced = {},
		annoucTmp = {},
		sharing = {},
		active = false,
		bLog = true,
		onAdd = TEMP_FUNCTION_EMPTY,
		onRmv = TEMP_FUNCTION_EMPTY,
		onCng = TEMP_FUNCTION_EMPTY,
		onAct = TEMP_FUNCTION_EMPTY,
		onVer = TEMP_FUNCTION_EMPTY,
	}, VersionHandler)

	Joiner:Join(chName, o)

	table.insert(VersionHandler.arHandlers, o)

	return o
end

function VersionHandler:Create(type, tbl)
	tbl = tbl or {}
	tbl["*"] = type or tbl["*"]
	local owner = self.data["*"]

	local key = self.temp[owner][true]+1

	self.temp[owner][key] = tbl
	self.temp[owner][true] = key

	self.onAdd(owner, -key, tbl)

	if self.active then --only share a version change, if we actively announce it
		self:OnRequestVersion(self.channel, owner, nil)
	end

	return -key, tbl
end

function VersionHandler:Remove(key)
	local owner = self.data["*"]
	local version, bTemp = math.abs(key), key<0
	local src = self:GetOwnerData(owner, bTemp)

	if src[true] < version or not src[version] then
		return error("Tried to remove non-existant Data.")
	end

	local tmp = src[version]
	src[version] = false
	src[false] = src[false]+1

	self.onRmv(owner, key, tmp)

	if self.active then --only share a version change, if we actively announce it
		self:OnRequestVersion(self.channel, owner, nil)
	end

	return true
end

function VersionHandler:RemoveAll()
	local owner = self.data["*"]
	local data = self:GetOwnerData(owner, false)
	local temp = self:GetOwnerData(owner, true)

	for i = data[true], 1, -1 do
		if data[i] then
			local tmp = data[i]
			data[i] = false
			data[false] = data[false]+1

			self.onRmv(owner, i, tmp)
		end
	end

	for i = temp[true], 1, -1 do
		if temp[i] then
			local tmp = temp[i]
			temp[i] = false
			temp[false] = temp[false]+1

			self.onRmv(owner, -i, tmp)
		end
	end

	if self.active then --only share a version change, if we actively announce it
		self:OnRequestVersion(self.channel, owner, nil)
	end

	return true
end

function VersionHandler:Save(key, tbl)
	if not key and not tbl then
		return -- we need at least one to do anything.
	end
	local owner = self.data["*"]
	local data = self:GetOwnerData(owner)

	if key then -- remove the old version at key, write 'tbl' if not given.
		local bTemp, version = key<0, math.abs(key)
		local src = bTemp and self:GetOwnerData(owner, bTemp) or data

		if not src or not src[version] then
			return
			--the function was called with the 'hope' of removing something
			--we will not be able to - rather report nothing changed.
		end

		local tmp = src[version]
		tbl = tbl or tmp
		src[version] = false
		src[false] = src[false]+1

		self.onRmv(owner, key, tmp)
	end
	--tbl is given here (either was from call, or populated from key~=nil)

	--find the new key; place tbl; return key, tbl;
	key = data[true]+1
	data[true] = key
	data[key] = tbl

	self.onAdd(owner, key, tbl)

	if self.active then --only share a version change, if we actively announce it
		self:OnRequestVersion(self.channel, owner, nil)
	end

	return key, tbl
end

do
	local timers = {} -- [VersionHandler] = timer
	local function OnTimer(tbl)
		local done = true
		for v in pairs(tbl.req) do
			tbl.req[v] = nil
			tbl.self:SendChange(nil, math.abs(v), v<0)
			done = false
		end
		if done then
			tbl.tmr:Stop()
			timers[tbl.self] = nil
		end
	end
	function VersionHandler:Changed(key, data)
		if self.active and self.sharing[self.data["*"] or ""] then --only share a version change, if we actively announce it
			if not timers[self] then
				timers[self] = {
					self = self,
					req = {},
					OnTimer = OnTimer,
				}
				timers[self].tmr = ApolloTimer.Create(1, true, "OnTimer", timers[self])
			end
			timers[self].req[key] = true
		end

		if data then
			local tar = self:GetOwnerData(nil, key<0)
			tar[math.abs(key)] = data
		else
			data = self:Get(key)
		end

		self.onCng(self.data["*"], key, data)
	end
end

function VersionHandler:BigChange(func)
	if type(func) ~= "function" then error("VersionHandler:BigChange(func) expected function, got "..type(func)) end

	local owner = self.data["*"]

	local oldKeys = {}
	for key in self:Iterate() do
		oldKeys[key] = true
	end

	local oldVMaj, oldVMin, oldVMajTemp, oldVMinTemp = self:GetVersion(nil, true)

	local active = self.active
	self.active = false

	func(self)

	self.active = active

	--prevent these from firing a second time:
	local onAdd, onRmv, onCng = self.onAdd, self.onRmv, self.onCng
	self.onAdd, self.onRmv, self.onCng = TEMP_FUNCTION_EMPTY, TEMP_FUNCTION_EMPTY, TEMP_FUNCTION_EMPTY

	for key in pairs(oldKeys) do
		if self:Get(key) then
			self:Changed(key)
		end
	end

	self.onAdd, self.onRmv, self.onCng = onAdd, onRmv, onCng

	local newVMaj, newVMin, newVMajTemp, newVMinTemp = self:GetVersion(nil, true)
	if active and (oldVMaj~=newVMaj or oldVMin~=newVMin or oldVMajTemp~=newVMajTemp or oldVMinTemp~=newVMinTemp) then
		self:OnRequestVersion(self.channel, owner, nil)
	end
end

function VersionHandler:GetOwners()
	local tbl = {}
	for name, _ in pairs(self.data) do
		if name:len() > 1 then --filter "*", "_" out.
			table.insert(tbl, name)
		end
	end
	return tbl
end

function VersionHandler:HasOwner(owner)
	owner = owner or self.data["*"]

	if owner:len() <= 1 then return false end

	return self.data[owner] and true or false
end

function VersionHandler:OwnerIsSelf(owner)
	return self.data["*"] == owner
end

function VersionHandler:Iterate(owner)
	owner = owner or self.data["*"]
	local data, temp = self:GetOwnerData(owner), self:GetOwnerData(owner, true)

	local next, tbl, idx, v = ipairs(data)

	local function nxt()
		idx, v = next(tbl, idx)
		while type(v) == "boolean" do
			idx, v = next(tbl, idx)
		end
		if v == nil then
			if tbl == data then
				next, tbl, idx = ipairs(temp)
				return nxt() --start from top!
			else
				return
			end
		end
		if tbl == data then
			return idx,v
		else
			return -idx,v
		end
	end

	return nxt
end

function VersionHandler:SetSharing(bSharing, owner)
	owner = owner or self.data["*"]

	if bSharing then
		self.sharing[owner] = true
		self:SendVersion(owner, nil)
	else
		self.sharing[owner] = nil
		--do we wanna send a message like 'dactivate-plz' ?
	end
end

function VersionHandler:SetActive( bActive )
	local announce = bActive~=self.active
	self.active = bActive
	if bActive then
		self:RequestVersion()
	end
	if announce then
		self.onAct(bActive)
	end
end

function VersionHandler:SetLogging(bLog)
	self.bLog = bLog or false
end

function VersionHandler:OnAdded(func, tHandler, ...)
	self.onAdd = callback(func, tHandler, self.data["_"], ...)
end

function VersionHandler:OnRemoved(func, tHandler, ...)
	self.onRmv = callback(func, tHandler, self.data["_"], ...)
end

function VersionHandler:OnChanged(func, tHandler, ...)
	self.onCng = callback(func, tHandler, self.data["_"], ...)
end

function VersionHandler:OnActive(func, tHandler, ...)
	self.onAct = callback_onearg(func, tHandler, ...)
end

function VersionHandler:OnNewVersion(func, tHandler, ...)
	self.onVer = callback_twoarg(func, tHandler, self.data["_"], ...)
end

function VersionHandler:Get(key, owner)
	local version, bTemp = math.abs(key), key<0
	owner = owner or self.data["*"]

	return (bTemp and self.temp or self.data)[owner][version]
end

function VersionHandler:GetData( )
	return self.data
end

function VersionHandler:SetTypeTable(tbl)
	typeTbl = tbl
end

--[[			====== Helper Methods ======			]]
-- CreateData, GetVersion, GetStructure, ApplyStructure, OnChannelReady?

function VersionHandler:CreateData(prefix, initData)
	local name = GameLib.GetPlayerCharacterName()

	if initData and initData["_"] == prefix and initData["*"] then
		local oldName = initData["*"]
		if name ~= oldName then
			initData["*"], initData[name], initData[oldName] = name, initData[oldName], nil
		end
	else
		initData = nil
	end

	local data = initData or {
		["*"] = name, --the mark used, to know which one is our table.
		["_"] = prefix,
		[name] = { --this is our own version!
			[true] = 0, --currently empty
			[false] = 0,
		},
	}
	local temp = {--the temporary versions
		[name] = {
			[true] = 0,
			[false] = 0,
		}
	}

	local missData = {}
	for owner, ownerData in pairs(data) do
		if owner:len() > 1 then --filter "*", "_" out.
			missData[owner] = {}
			for idx, d in ipairs(ownerData) do
				if d == true then
					missData[owner][idx] = true
				end
			end
		end
	end

	local missTemp = {}
	for owner, ownerData in pairs(temp) do
		if owner:len() > 1 then --filter "*", "_" out.
			missTemp[owner] = {}
			for idx, d in ipairs(ownerData) do
				if d == true then
					missTemp[owner][idx] = true
				end
			end
		end
	end

	return data, temp, missData, missTemp
end

function VersionHandler:GetMinMissing(owner, bTemp)
	local _, tMiss = self:GetOwnerData(owner, bTemp)
	local nMiss = next(tMiss)
	for idx in pairs(tMiss) do
		if idx < nMiss then
			nMiss = idx
		end
	end
	return nMiss
end

function VersionHandler:GetOwnerData(owner, bTemp)
	owner = owner or self.data["*"]
	if not self.data[owner] then
		self.data[owner] = {[true] = 0, [false] = 0}
	elseif not self.data[owner][true] or not self.data[owner][false] then
		local sum, rem = #self.data[owner],0
		for idx, var in ipairs(self.data[owner]) do
			if var == false then
				rem = rem+1
			end
		end
		self.data[owner][true], self.data[owner][false] = sum, rem
	end
	if not self.temp[owner] then
		self.temp[owner] = {[true] = 0, [false] = 0}
	elseif not self.temp[owner][true] or not self.temp[owner][false] then
		local sum, rem = #self.temp[owner],0
		for idx, var in ipairs(self.temp[owner]) do
			if var == false then
				rem = rem+1
			end
		end
		self.temp[owner][true], self.temp[owner][false] = sum, rem
	end

	self.missData[owner] = self.missData[owner] or {}
	self.missTemp[owner] = self.missTemp[owner] or {}

	return bTemp and self.temp[owner] or self.data[owner], bTemp and self.missTemp[owner] or self.missData[owner]
end

function VersionHandler:GetVersion(owner, bNumeric) --retruns "Mischh X:7:4" or returns 7, 4
	owner = owner or self.data["*"]
	if not self.data[owner] then --we dont know this one!
		return
	end

	local data, temp = self:GetOwnerData(owner, false), self:GetOwnerData(owner, true)

	if bNumeric then
		local maj, min = data[true], data[false]
		if temp and temp[true] > 0 then
			return maj, min, temp[true], temp[false]
		end
		return maj, min
	end

	local major, minor, tempMj, tempMi = self:GetVersion(owner, true)

	major, minor = Coder.toB64(major), Coder.toB64(minor)
	if tempMj and tempMi then
		tempMj, tempMi = Coder.toB64(tempMj), Coder.toB64(tempMi)
		return concat(owner, major, minor, tempMj, tempMi)
	else
		return concat(owner, major, minor)
	end
end

function VersionHandler:GetStructure(owner, bTemp)
	owner = owner or self.data["*"]
	local tbl = self:GetOwnerData(owner, bTemp)
	if not tbl then
		return --we cant get any structure..
	end

	local str = ""

	--check, what to start with:
	local t = type(tbl[1])
	local counter = 0


	if not t then
		return concat(owner, "0") --the 'default' - nothing-here - structure
	end


	for i=1, tbl[true], 1 do
		if type(tbl[i]) == t then
			counter = counter+1
		else
			if t == "boolean" then
				--turn the counter to B64
				--our B64 does not contain numbers, so we can simply stuff them together
				str = str..Coder.toB64(counter)
			else
				str = str..tostring(counter)
			end

			t = type(tbl[i])
			counter = 1
		end
	end

	if t == "boolean" then
		str = str..Coder.toB64(counter)
	else
		str = str..tostring(counter)
	end

	return concat(owner, str)
end

--name is a bit misleading, in early development the structure actually matched afterwards
--now it just removes data, that the structure marks as empty
function VersionHandler:ApplyStructure(owner, struct, bTemp)
	--validate struct and put it into a usable format!
	local bFirstFilled = struct:find("^%d") and 1 or 0
	local structTbl = {}
	if not pcall(function()
		for b64, num in struct:gmatch("(%D*)(%d*)") do
			if b64 then
				structTbl[#structTbl+1] = Coder.fromB64(b64)
			end
			if num then
				local n = tonumber(num)
				if not n then return end
				structTbl[#structTbl+1] = n
			end
		end
	end) then
		return --failed to verify the structure, probably thrown error decoding B64!
	end

	local data, miss = self:GetOwnerData(owner, bTemp)
	local request = {}--the table, we fill as return (fill with missing indexes)
	local fill, empt = 0, 0
--~ 	local bExpand = true --go on with expanding 'data'; once we miss

	local idx = 1
	for i, n in ipairs(structTbl) do
		if i%2 == bFirstFilled then
			fill = fill + n
			--the struct sais, all in this range are supposed to be filled
			for j = idx, idx+n-1, 1 do
				if type(data[j]) ~= "table" then
					data[j] = true -- dont want to do this anymore.
					miss[j] = true
--~ 					if data[i] == nil then
--~ 						bExpand = false
--~ 					end

					request[#request+1] = j
				end
			end
		else
			empt = empt + n
			for j = idx, idx+n-1, 1 do
				local tmp = data[j]
				data[j] = false
				if type(tmp) == "table" then
					local key = (bTemp and -1 or 1) * j
					self.onRmv(owner, key, tmp)
				end
			end
		end
		idx = idx+n
	end

	--this will only go through something, if data[true]>fill
	for i = fill+empt, (data[true] or fill+empt), 1 do
		local tmp = data[i]
		data[i] = nil
		if type(tmp) == "table" then
			local key = (bTemp and -1 or 1) * i
			self.onRmv(owner, key, tmp)
		end
	end

	data[true] = fill+empt
	data[false] = empt

	return request
end

function VersionHandler:OnChannelReady(channel)
	self.channel = channel
	ChannelManager:Filter(channel, self.data["_"], self)
	if self.active then
		--ask for stuff to import?
		self:RequestVersion()
	end

	self:OnRequestVersion(self.channel, nil, nil)--share all the versions, we want to share, as if somebody asked for them.
end

function VersionHandler:OnGroupSwitched() --we changed groups (left/joined one)
	for _, vHandler in ipairs(VersionHandler.arHandlers) do
		if vHandler.active then
			vHandler:RequestVersion()
		end
		for owner in pairs(vHandler.sharing) do
			vHandler:SendVersion(owner, nil)
		end
	end
end

function VersionHandler:OnDocLoaded()
	Apollo.RegisterEventHandler("Group_Join", "OnGroupSwitched", VersionHandler)
	Apollo.RegisterEventHandler("Group_Left", "OnGroupSwitched", VersionHandler)
end
NLLC:OnceDocLoaded("OnDocLoaded", VersionHandler)

--[[			====== Version-Messages ======			]]
-- RequestVersion, OnRequestVersion, SendVersion, OnVersion

function VersionHandler:RequestVersion(owner, target)
	ChannelManager:Send(self.channel, self.prefix, TYPE_REQ_VERSION, owner, target)
end

function VersionHandler:OnRequestVersion( channel, msg, srcName )
	if msg then --the request was targeted to the version of this owner (msg)
		if self.sharing[msg] or msg==srcName then
			self:SendVersion(msg, srcName)
		end
	else --untargeted -> share all sharing versions!
		for name in pairs(self.sharing) do
			self:SendVersion(name, srcName)
		end
	end
end

function VersionHandler:SendVersion(owner, target)
	local strVersion = self:GetVersion(owner)
	if not strVersion then return end
	ChannelManager:Send(self.channel, self.prefix, TYPE_SND_VERSION, strVersion, target)
end

function VersionHandler:OnVersion( channel, msg, srcName )
	local owner, major, minor, tMajor, tMinor = seperate(msg, 5)

	if not owner or not major or not minor then
		return --all data is required, do nothing, if not supported
	end

	if not pcall(function()
		major, minor = Coder.fromB64(major), Coder.fromB64(minor)
		if tMajor and tMinor then
			tMajor, tMinor = Coder.fromB64(tMajor), Coder.fromB64(tMinor)
		end
	end) then
		return -- couldnt decode the numbers -> not supported data
	end

	local sentVersion = false

	local data, temp = self:GetOwnerData(owner, false), self:GetOwnerData(owner, true)
	local nMissData, nMissTemp = self:GetMinMissing(owner, false), self:GetMinMissing(owner, false)

	if (not data or not temp or data[true] < major or data[true] == major and data[false] < minor
		or tMajor and tMinor and (temp[true] < tMajor or temp[true] == tMajor and temp[false] < tMinor)) then
		-- a greater version!
		self.onVer(owner, {
			major = major,
			minor = minor,
			tempMajor = tMajor,
			tempMinor = tMinor,
		})
	end

	--compare the versions to our current versions, decide whether we need to ask for Structures
	if not data or data[true] < major or nMissData and nMissData < major then
		self:RequestStructure(owner, srcName)
	elseif data[true] > major then
		if owner == srcName then self:SendVersion(owner, srcName); sentVersion = true end
	elseif data[false] < minor then
		self:RequestStructure(owner, srcName)
	elseif data[false] > minor then
		if owner == srcName then self:SendVersion(owner, srcName); sentVersion = true end
	end

	if tMajor and tMinor and tMajor>0 then
		if not temp or temp[true] < tMajor or nMissTemp and nMissTemp < tMajor then
			self:RequestTempStructure(owner, srcName)
		elseif temp[true] > tMajor then
			if owner == srcName and sentVersion then self:SendVersion(owner, srcName) end
		elseif temp[false] < tMinor then
			self:RequestTempStructure(owner, srcName)
		elseif temp[false] > tMinor then
			if owner == srcName and sentVersion then self:SendVersion(owner, srcName) end
		end
	end
end

--[[			====== Structure-Messages ======			]]
-- RequestStructure, OnRequestStructure, SendStructure, OnStructure

function VersionHandler:RequestStructure(owner, target, bTemp)
	if bTemp then
		self:RequestTempStructure(owner, target)
	else
		ChannelManager:Send(self.channel, self.prefix, TYPE_REQ_STRUCTURE, owner, target)
	end
end

function VersionHandler:OnRequestStructure(channel, msg, srcName)
	if msg then --the request was targeted to the version of this owner (msg)
		if self.sharing[msg] or msg==srcName then
			self:SendStructure(msg, srcName)
		end
	else --untargeted -> share all sharing versions!
		for name in pairs(self.sharing) do
			self:SendStructure(name, srcName)
		end
	end
end

function VersionHandler:SendStructure(owner, target)
	ChannelManager:Send(self.channel, self.prefix, TYPE_SND_STRUCTURE, self:GetStructure(owner), target)
end

function VersionHandler:OnStructure(channel, msg, srcName)
	--compare the newly found structure to our own version.
	local owner, struct = seperate(msg, 2)

	if not struct then return end

	local request = self:ApplyStructure(owner, struct)

	if not request then return end --something went wrong

	self:RequestData(owner, request, srcName)
end

--[[			====== TempStructure-Messages ======			]]
-- RequestTempStructure, OnRequestTempStructure, SendTempStructure, OnTempStructure

function VersionHandler:RequestTempStructure(owner, target)
	ChannelManager:Send(self.channel, self.prefix, TYPE_REQ_STRUCTURE_TEMP, owner, target)
end

function VersionHandler:OnRequestTempStructure(channel, msg, srcName)
	if msg then --the request was targeted to the version of this owner (msg)
		if self.sharing[msg] or msg==srcName then
			self:SendTempStructure(msg, srcName)
		end
	else --untargeted -> share all sharing versions!
		for name in pairs(self.sharing) do
			self:SendTempStructure(name, srcName)
		end
	end
end

function VersionHandler:SendTempStructure(owner, target)
	ChannelManager:Send(self.channel, self.prefix, TYPE_SND_STRUCTURE_TEMP, self:GetStructure(owner, true), target)
end

function VersionHandler:OnTempStructure(channel, msg, srcName)
	--compare the newly found temporary-structure to our own version.
	local owner, struct = seperate(msg, 2)

	if not struct then return end

	local request = self:ApplyStructure(owner, struct, true)

	if not request then return end --something went wrong

	self:RequestTempData(owner, request, srcName)
end

--[[			====== Data-Messages ======			]]
-- RequestData, OnRequestData, SendData, OnData

function VersionHandler:RequestData(owner, idTbl, target, bTemp)
	if not owner or not idTbl or not target then
		return
	end

	if type(idTbl) == "number" then
		idTbl = {idTbl}
	end

	local str = owner

	local last;
	local miss;
	for i, id in ipairs(idTbl) do
		if not last then
			str = str..":"..Coder.toB64(id)
			miss = nil
		elseif id == last+1 then
			miss = id
		elseif id > last+1 then
			if miss == last then
				str = str..">"..Coder.toB64(last)
			end
			str = str..":"..Coder.toB64(id)
			miss = nil
		end
		last = id
	end

	if last then
		if miss == last then
			str = str..">"..Coder.toB64(last)
		end
		ChannelManager:Send(self.channel, self.prefix, bTemp and TYPE_REQ_DATA_TEMP or TYPE_REQ_DATA, str, target)
	end
end

function VersionHandler:OnRequestData(channel, msg, srcName, bTemp)
	if not msg then return end
	local owner, request = seperate(msg, 2)

	local reqStr = {}

	for str in (":"..request):gmatch(":([^:]*)") do
		reqStr[#reqStr+1] = str
	end

	local requested = {}

	if not pcall(function()
		local n = 1
		for i, str in ipairs(reqStr) do
			local from, to = str:match("^([^>]*)>([^>]*)$")
			if not from or not to then
				requested[n] = Coder.fromB64(str)
				n = n+1
			else
				from, to = Coder.fromB64(from), Coder.fromB64(to)
				for num = from, to, 1 do
					requested[n] = num
					n = n+1
				end
			end
		end
	end) then
		return
	end

	self:SendData(owner, requested, srcName, bTemp)
end

function VersionHandler:SendData(owner, idTbl, target, bTemp)
	if not owner or not idTbl then
		return
	end

	if type(idTbl) == "number" then
		idTbl = {idTbl}
	end

	local data = self:GetOwnerData(owner, bTemp)

	local n, sending = 0, {}
	for _, id in ipairs(idTbl) do
		if data[id] then
			--"Mischh X:B:{"t""Hello World!""h"iy|"w"iBk|"*""P"}"
			n = n+1
			sending[n] = concat( owner, Coder.toB64(id), Coder.toEncTbl(data[id]) )
		end
	end

	if n>0 then
		local function finish()
			--we want to share our version with him, once we are done.
			--Because we might have messed up some changes, which were done in the meantime.
			self:SendVersion(owner, target)
		end

		ChannelManager:SendSeries(self.channel, self.prefix, bTemp and TYPE_SND_DATA_TEMP or TYPE_SND_DATA,
				sending, target, finish)
	end
end

function VersionHandler:OnData(channel, msg, srcName, bTemp)
	if not msg then return end
	local owner, version, data = seperate(msg, 3)
	if not data then return end --else: owner & version are given aswell.

	if not pcall( function()
		version, data = Coder.fromB64(version), Coder.fromEncTbl(data)
	end) then
		return
	end

	local tData, tMiss = self:GetOwnerData(owner, bTemp)

	if tData[true] < version then
		if tData[true]+1 == version then
			tData[true] = version
		else
			--we seem to not be prepared for such an 'advanced' version, maybe rather request a structure?!
			--and scince its this bad, maybe switch this for a version-request??
			self:RequestStructure(owner, srcName, bTemp)
			return
		end
	end

	if tData[version] == false then
		tData[false] = tData[false]-1
	end
	local tmp = tData[version]
	tData[version] = data
	tMiss[version] = nil

	local key = (bTemp and -1 or 1) * version
	if type(tmp) ~= "table" then --if was newly created
		self.onAdd(owner, key, data)
	elseif tmp["_"] == data["_"] then
		self.onCng(owner, key, data)
	else
		self.onRmv(owner, key, tmp)
		self.onAdd(owner, key, data)
	end
end

--[[			====== TempData-Messages ======			]]
-- RequestTempData, OnRequestTempData, SendTempData, OnTempData

function VersionHandler:RequestTempData(owner, idTbl, target)
	return self:RequestData(owner, idTbl, target, true)
end

function VersionHandler:OnRequestTempData(channel, msg, srcName)
	return self:OnRequestData(channel, msg, srcName, true)
end

function VersionHandler:SendTempData(owner, idTbl, target)
	return self:SendData(owner, idTbl, target, true)
end

function VersionHandler:OnTempData(channel, msg, srcName)
	return self:OnData(channel, msg, srcName, true)
end

--[[			====== Change-Messages ======			]]
-- SendChange, OnChange
do
	local function deepCopy(tbl)
		if type(tbl) ~= "table" then
			return tbl
		end

		local copy = {}
		for i,v in pairs(tbl) do
			if type(i) ~= "table" then
				if type(v) == "table" then
					v = deepCopy(v)
				end
				copy[i] = v
			end
		end
		return copy
	end

	local function deepCompare(tbl, newTbl)
		if not tbl then
			local t = deepCopy(newTbl)
			return t, nil, t
		end
		if not newTbl then
			return nil, true, nil
		end
		if type(tbl) == "table" and type(newTbl) == "table" then
			local add, rmv, cop = {}, {}, {}
			local numR = 0
			for i,v in pairs(tbl) do
				if type(i) ~= "table" then
					add[i], rmv[i], cop[i] = deepCompare(tbl[i], newTbl[i])
					if rmv[i] ~= nil then
						numR = numR+1
					end
				end
			end
			for i,v in pairs(newTbl) do
				if type(i) ~= "table" and not cop[i] then
					add[i], rmv[i], cop[i] = deepCompare(tbl[i], newTbl[i])
					if rmv[i] ~= nil then
						numR = numR+1
					end
				end
			end
			return add, (numR>0 and rmv or nil), cop
		elseif tbl == newTbl then
			local t = deepCopy(newTbl)
			return nil, nil, t
		else
			local t = deepCopy(newTbl)
			return t, nil, t
		end
	end

	local function deepSet(tbl, app, rmv)
		for i,v in pairs(app or {}) do
			if type(v) == "table" then
				if type(tbl[i]) ~= "table" then
					tbl[i] = {}
				end
				deepSet(tbl[i], v, nil)
			else
				tbl[i] = v
			end
		end
		for i,v in pairs(rmv or {}) do
			if type(v) == "table" then
				if type(tbl[i]) ~= "table" then
					tbl[i] = {}
				end
				deepSet(tbl[i], nil, v)
			else
				tbl[i] = nil
			end
		end
	end

	--dont accept a target, because changes are always announced.
	function VersionHandler:SendChange(owner, version, bTemp)
		local announced = bTemp and self.annoucTmp or self.accounced
		owner = owner or self.data["*"]
		if not announced[owner] then
			announced[owner] = {}
		end
		announced = announced[owner]

		local data = self:GetOwnerData(owner, bTemp)
		if not data or not data[version] then return end --no data to share?
		data = data[version]

		local comp = typeTbl[data["*"]]:ToComparable(data)
		--compare our last announcement to the new data.
		local add, rmv, cop = deepCompare(announced[version], comp)
		announced[version] = cop --apply the copy.

		add, rmv = typeTbl[data["*"]]:FromComparable(add), typeTbl[data["*"]]:FromComparable(rmv)
		add, rmv = Coder.toEncTbl(add), Coder.toEncTbl(rmv)
		version = Coder.toB64(version)
		ChannelManager:Send(self.channel, self.prefix, bTemp and TYPE_SND_CHANGE_TEMP or TYPE_SND_CHANGE,
				concat(owner, version, add, rmv))
	end

	function VersionHandler:OnChange(channel, msg, srcName, bTemp)
		--tbls contains both, the add and remove table!
		local owner, version, tbls = seperate(msg, 3)
		local add, rmv; --filled from tbls.

		if not tbls then return end

		--check, if we even have this version. If we dont: rather ask for the full version.
		local data = self:GetOwnerData(owner, bTemp)

		if not pcall(function()
			version = Coder.fromB64(version)
			add, tbls = Coder.fromEncTbl(tbls)
			tbls = tbls:match("^:(.*)$")
			rmv, tbls = Coder.fromEncTbl(tbls)

			if tbls ~= "" then
				error("Something went wrong decoding Change-Tables.")
			end
		end) then
			return
		end

		if not data or not data[version] then
			self:RequestData(owner, version, srcName, bTemp)
			return
		end

		local orig = data[version]
		local typ = orig["*"]

		if not typ or not typeTbl[typ] then return end

		orig =  typeTbl[typ]:ToComparable(orig)
		add =  typeTbl[typ]:ToComparable(add)
		rmv =  typeTbl[typ]:ToComparable(rmv)

		deepSet(orig, add, rmv)

		orig = typeTbl[typ]:FromComparable(orig)
		data[version] = orig
		local key = (bTemp and -1 or 1) * version
		self.onCng(owner, key, orig)
	end
end

--[[			====== TempChange-Messages ======			]]
-- SendTempChange, OnTempChange
function VersionHandler:SendTempChange(owner, version)
	self:SendChange(owner, version, true)
end

function VersionHandler:OnTempChange(channel, msg, srcName)
	self:OnChange(channel, msg, srcName, true)
end


do --[[			====== ChannelManager ======			]]
	--[[local tMessageTypes = {
		-- example: "V:D105:Mischh X:7:4" should be routed to VersionHandler:OnVersion(...)
		V = "OnVersion", --there are much more in the real table. this is just an example.
	}]]

	local handlers = {
		--[channel] = { ["prefix"] = handler, ... }
	}

	function ChannelManager:Filter(channel, prefix, handler)
		local chName = channel:GetName()
		if not handlers[chName] then
			handlers[chName] = {}
			channel:SetReceivedMessageFunction("OnMessage", self)
		end

		if handlers[chName][prefix] then
			error("ChannelManager: multiple handlers for single prefix: \""..tostring(prefix).."\"")
		else
			handlers[chName][prefix] = handler
		end
	end

	function ChannelManager:OnMessage(channel, message, srcName)
		local typ, prefix, msg = message:match(MSG_RCV_PATTERN)
		if not typ then
			typ, prefix = message:match(MSG_RCV_PATTERN_SHORT)
		end
		if not typ then --couldnt match the message to the pattern -> throw the message away
			return
		end
		local chName = channel:GetName()
		if not handlers[chName] then --there is not a single handler, that could recieve something?
			return
		end
		local prefixes = handlers[chName]

		if not prefixes[prefix] then -- there is no handler waiting for this prefix
			return
		end
		local handler = prefixes[prefix]
		if handler.bLog then Log:Add("R", srcName, message) end

		if not tMessageTypes[typ] then --dont know this message-type ?!
			return
		end
		-- pass the message on to the handler:
		handler[ tMessageTypes[typ] ](handler, channel, msg, srcName)
	end

	function ChannelManager:Send(channel, prefix, typ, msg, target)
		if not channel then return end

		local out;
		if msg then
			out = MSG_SND_PATTERN:format(typ, prefix, msg)
		else
			out = MSG_SND_PATTERN_SHORT:format(typ, prefix)
		end

		if self:SpamFilter(channel, target, out) then
			local prefixes = handlers[channel:GetName()]
			if prefixes and prefixes[prefix] then
				if prefixes[prefix].bLog then Log:Add("S", target, out) end
			end
		end
	end

	function ChannelManager:SendSeries(channel, prefix, typ, tbl, target, finisher)
		if not channel then return end

		local arMsg = {}
		for i, msg in ipairs(tbl) do
			if msg then
				arMsg[i] = MSG_SND_PATTERN:format(typ, prefix, msg)
			else
				arMsg[i] = MSG_SND_PATTERN_SHORT:format(typ, prefix)
			end
		end

		local tInserted = self:SpamFilterSeries(channel, target, arMsg, finisher)

		local bLog = false;
		local prefixes = handlers[channel:GetName()]
		if prefixes and prefixes[prefix] then
			bLog = prefixes[prefix].bLog
		end

		if bLog and tInserted then
			for i, msg in pairs(tInserted) do
				Log:Add("S", target, msg)
			end
		end
	end

	local tChannels = {
		-- [chName] = channel
	}
	local tMessages = {
		--	[chName] = {
			--	["Unit Target"] = {	"Next Message to be sent", "another message to be sent", function() to be executed end, ...}
			--	[false] = {"First Message to public", ...}
		--	}
	}
	local tSpam = {
		--	[chName] = {
			-- ["Unit Target"] = { ["Next Message to be sent"] = true, ["another message to be sent"] = true, ...}
			-- [false] = {["First Message to public"] = true, ...}
		--	}
	}

	function ChannelManager:SpamFilter(channel, target, msg)
		if not target then target = false end

		local chName = channel:GetName()
		if not tChannels[chName] then
			tChannels[chName] = channel
			tMessages[chName] = {}
			tSpam[chName] = {}
		end

		tMessages[chName][target] = tMessages[chName][target] or {}
		tSpam[chName][target] = tSpam[chName][target] or {}

		if not tSpam[chName][target][msg] then
			table.insert(tMessages[chName][target], msg)
			tSpam[chName][target][msg] = true
			self:AwakeSpamTimer()
			return true
		end
		return false
	end

	function ChannelManager:SpamFilterSeries(channel, target, arMsg, finish)
		if not target then target = false end

		local chName = channel:GetName()
		if not tChannels[chName] then
			tChannels[chName] = channel
			tMessages[chName] = {}
			tSpam[chName] = {}
		end

		tMessages[chName][target] = tMessages[chName][target] or {}
		tSpam[chName][target] = tSpam[chName][target] or {}

		local tInserted = {}

		for i,msg in ipairs(arMsg) do
			if not tSpam[chName][target][msg] then
				table.insert(tMessages[chName][target], msg)
				table.insert(tInserted, msg)
				tSpam[chName][target][msg] = true
			end
		end

		if type(finish) == "function" then
			table.insert(tMessages[chName][target], finish)
		end

		if #tMessages[chName][target] > 0 then
			self:AwakeSpamTimer()
		end
		return tInserted
	end

	function ChannelManager:AwakeSpamTimer()
		if not self.spamTimer then
			self.spamTimer = ApolloTimer.Create(0.5, true, "OnSpamTimer", self)
			self.spamTimer:Start()
		end
	end

	function ChannelManager:OnSpamTimer()
		local continue = false
		for chName, tChnlMessages in pairs(tMessages) do
			local channel = tChannels[chName]
			local tCleanup = {}
			for target, tMsg in pairs(tChnlMessages) do
				local msg = table.remove(tMsg, 1)
				while type(msg) == "function" do
					msg()
					msg = table.remove(tMsg, 1)
				end

				if msg then --Actually send it!
					tSpam[chName][target][msg] = nil
					if target then
						channel:SendPrivateMessage(target, msg)
					else
						channel:SendMessage(msg)
					end
				end
				if #tMsg < 1 then tCleanup[target] = true end
			end
			for target in pairs(tCleanup) do
				tChnlMessages[target] = nil
			end
			if next(tChnlMessages) then
				continue = true
			end
		end
		if not continue then
			self.spamTimer:Stop()
			self.spamTimer = nil
		end
	end
end

do --[[			====== Joiner ======			]]
	local joined = {}
	local requested = {}

	function Joiner:Join(name, handler)
		if joined[name] then
			self:NotifyReady(handler, joined[name])
		else
			requested[name] = requested[name] or {}
			table.insert(requested[name], handler)
			self:NotifyTimer() --make sure the timer is either running, or will be started here.
		end
	end

	function Joiner:NotifyReady(handler, channel)
		handler:OnChannelReady( channel )
	end

	function Joiner:NotifyTimer()
		--our timer should be a 1sec timer, continiously running, until we have nothing to do anymore.
		if not self.timer then
			self.timer = ApolloTimer.Create(1, true, "OnTimer", Joiner)
			self.timer:Start()
		end
	end

	function Joiner:OnTimer()
		local remaining = {} -- all the channels still not joined

		for name, tbl in pairs(requested) do
			local chnl = ICCommLib.JoinChannel(name, ICCommLib.CodeEnumICCommChannelType.Group)
			if chnl:IsReady() then
				for _, handler in ipairs(tbl) do --notify the handlers waiting for the channel!
					self:NotifyReady(handler, chnl)
				end
				joined[name] = chnl
			else
				remaining[name] = tbl
			end
		end

		requested = remaining

		if not next(requested) then --success! yeah!
			self.timer:Stop()
			self.timer = nil
		end
	end
end
